
#include "anatomy.h"

mySet **anatomy_hashTupleWithSenValue(data *aData, int noOfBucket)
{
	mySet **returnValue;

	tuple *tupleList;
	tuple *aTuple;

	int i;

	int noOfDim;
	int aSenValue;

	noOfDim = aData->noOfDim;

	tupleList = aData->tupleList;

	returnValue = (mySet **) malloc(sizeof(mySet *)*noOfBucket);
	for (i = 0; i < noOfBucket; i++)
	{
		returnValue[i] = mySet_new();
	}

	for (i = 0; i < aData->noOfTuple; i++)
	{
		aTuple = &(tupleList[i]);

		// level 0
		aSenValue = aTuple->content[noOfDim-1].value;

		mySet_insert(returnValue[aSenValue], i);
	}

	return returnValue;
}

// NULL: do not have enough l non-empty bucket
int *anatomy_getLLargestBucket(mySet **bucket, int l, int noOfBucket)
{
	int *returnValue;

	int *size;
	int *index;

	int i;

	size = (int *) malloc(sizeof(int)*noOfBucket);
	index = (int *) malloc(sizeof(int)*noOfBucket);
	memset(size, 0, sizeof(int)*noOfBucket);
	memset(index, 0, sizeof(int)*noOfBucket);

	for (i = 0; i < noOfBucket; i++)
	{
		index[i] = i;
		size[i] = bucket[i]->noOfElt;
	}

	quicksort_i(index, size, 0, noOfBucket-1);


	if (size[l-1] == 0)
	{
		returnValue = NULL;
	}
	else
	{
		returnValue = (int *) malloc(sizeof(int)*l);

		for (i = 0; i < l; i++)
		{
			returnValue[i] = index[i];
		}
	}
	
	free(size);
	free(index);

	return returnValue;
}

int anatomy_isContainSenValue(equivalenceTuple *aEqTuple, int aSenValue)
{
	int returnValue;

	int noOfMember;
	int *member;

	dimension *senValue;
	int thisSenValue;

	int i;

	member = aEqTuple->member;
	noOfMember = aEqTuple->noOfMember;

	senValue = aEqTuple->senValue;

	returnValue = 0;
	for (i = 0; i < noOfMember; i++)
	{
		thisSenValue = senValue[i].value;

		if (thisSenValue == aSenValue)
		{
			returnValue = 1;
			break;
		}
	}

	return returnValue;
}


int anatomy_isContainSenIndex_considerNewSen(equivalenceTuple *aEqTuple, int aSenValue, configure *config)
{
	int returnValue;

	int noOfMember;
	int *member;

	dimension *senValue;
	int thisSenValue;

	int i;

	member = aEqTuple->member;
	noOfMember = aEqTuple->noOfMember;

	senValue = aEqTuple->senValue;

	returnValue = 0;
	for (i = 0; i < noOfMember; i++)
	{
		thisSenValue = senValue[i].value;

		if (thisSenValue == aSenValue)
		{
			returnValue = 1;
			break;
		}
	}

	if (returnValue == 0)
	{
		if (aEqTuple->noOfEachNewSen[aSenValue] != 0)
		{
			returnValue = 1;
		}
	}

	return returnValue;
}

// -1: cannot find an eq. tuple
int anatomy_findEqTupleWithoutSenValue(equivalenceClass *eqClass, int aSenValue)
{
	int returnValue;

	int noOfEqTuple;
	equivalenceTuple **eqTupleList;

	equivalenceTuple *aEqTuple;

	int isContain;

	int i;

	noOfEqTuple = eqClass->noOfEqClass;
	eqTupleList = eqClass->eqClass;

	returnValue = -1;
	for (i = 0; i < noOfEqTuple; i++)
	{
		aEqTuple = eqTupleList[i];

		isContain = anatomy_isContainSenValue(aEqTuple, aSenValue);
		if (isContain == 0)
		{
			returnValue = i;
			break;
		}
	}

	return returnValue;
}

int anatomy_findNoOfEqTupleWithSenIndex(equivalenceClass *eqClass, int senIndex)
{
	int returnValue;

	int isContain;

	equivalenceTuple **eqTupleList;
	equivalenceTuple *aEqTuple;
	int noOfEqTuple;

	int i;
	
	eqTupleList = eqClass->eqClass;
	noOfEqTuple = eqClass->noOfEqClass;

	returnValue = 0;
	for (i = 0; i < noOfEqTuple; i++)
	{
		aEqTuple = eqTupleList[i];

		isContain = ass_isContainSenIndex(aEqTuple, senIndex);
		if (isContain == 1)
		{
			returnValue++;
		}
	}

	return returnValue;
}


int anatomy_findNoOfEqTupleWithSenIndex_considerNewSen(equivalenceClass *eqClass, int senIndex, configure *config)
{
	int returnValue;

	int isContain;

	equivalenceTuple **eqTupleList;
	equivalenceTuple *aEqTuple;
	int noOfEqTuple;

	int i;
	
	eqTupleList = eqClass->eqClass;
	noOfEqTuple = eqClass->noOfEqClass;

	returnValue = 0;
	for (i = 0; i < noOfEqTuple; i++)
	{
		aEqTuple = eqTupleList[i];

		isContain = ass_isContainSenIndex_considerNewSen(aEqTuple, senIndex, config);

		if (isContain == 1)
		{
			returnValue++;
		}
	}

	return returnValue;
}


int anatomy_findNoOfViolateEqTuple(equivalenceClass *eqClass, int senIndex, signatureGroupGroupGroup *allSig, int ***allSigMapping, int m, int noOfTuple)
{
	int returnValue;

	int isViolate;
	int isContain;

	equivalenceTuple **eqTupleList;
	equivalenceTuple *aEqTuple;
	int noOfEqTuple;

	int i;
	
	eqTupleList = eqClass->eqClass;
	noOfEqTuple = eqClass->noOfEqClass;

	returnValue = 0;
	for (i = 0; i < noOfEqTuple; i++)
	{
		aEqTuple = eqTupleList[i];

		isContain = ass_isContainSenIndex(aEqTuple, senIndex);
		if (isContain == 1)
		{
			isViolate = ass_isViolateGlobalDistBoundConditionInOneEqTuple(aEqTuple, senIndex, allSig, allSigMapping, m, noOfTuple);
			if (isViolate == 1)
			{
				returnValue++;
			}
		}
	}

	return returnValue;
}



void anatomy_main(data *aData, configure *config, generalizedAllTable *genAllTable)
{
	mySet **bucket;

	int noOfSenValue;
	int noOfDim;

	int l;

	int m;

	int noOfViolate;

	int *bucketIndexList;
	int bucketIndex;

	tuple *tupleList;

	mySet *aBucket;
	signatureGroupGroupGroup *allSig;

	int noOfElt;
	int anElt;

	int tupleID;

	int ***allSigMapping;

	FILE *avgRelError_filePtr;
	

	allCondition *aAllCondition;
	int noOfQID;

	twoTable *finalTable;

	double avgRelError;

	double avgEqTupleSize, avgSenEqTupleSize;

	mechAllGroup *finalMechAllGroup;

	dimension *newSenValue;

	int aSenValue;
	int *newEltList;

	FILE *debugFilePtr;
	char *debugFilename = "debug.txt";

	int i, j;

	int eqID;
	int senIndex;

	//int dummy;
	double infoLoss, distortion;
	int noOfSenChange;

	double globalDistDiff;
	double globalDistDiffSen;

	equivalenceTuple *anEqTuple;

	int threshold;
	int noOfSenIndex;

	int noOfEqTupleSenValue;

	equivalenceClass *proEqClass;

	char *avgRelErrorFilename = "error.txt";
	char *timeFilename = "time.txt";

	int queryNoOfDim = config->queryNoOfDim;
	double s = config->s;
	int noOfCondition = config->noOfCondition;

#ifndef _WIN32
	FILE *timeFilePtr;

	float  userTime, sysTime;
	float userTime_dist, sysTime_dist;
	struct rusage myTime1, myTime2, myTime_distEnd;

	getrusage(RUSAGE_SELF,&myTime1);
#endif

	noOfDim = aData->noOfDim;
	noOfSenValue = aData->dimSize[noOfDim-1];
	noOfQID = noOfDim-1;
	m = config->d;

	tupleList = aData->tupleList;

	l = config->d;


	// signature
	threshold = calViolation_calculateSampleSize(config->error, config->sigma);
	//threshold = 1;
	printf(" Sample threshold = %d\n", threshold);
	

	// assume only one sensitive index
	noOfSenIndex = 2;



	// generate all signature
	printf("Generating signatures...\n");
	allSig = sig_findAllSignature(aData, threshold);

	sig_printStatistics(stdout, allSig);

	// find distribution
	printf("Finding the global distributions...\n");
	allSigMapping = sig_findGlobalDistributionFromOrgTable(config, allSig, aData, noOfSenIndex);

#ifndef _WIN32

	getrusage(RUSAGE_SELF, &myTime_distEnd);	

	userTime =
     		((float) (myTime_distEnd.ru_utime.tv_sec  - myTime1.ru_utime.tv_sec)) +
       		((float) (myTime_distEnd.ru_utime.tv_usec - myTime1.ru_utime.tv_usec)) * 1e-6;
	sysTime =
       		((float) (myTime_distEnd.ru_stime.tv_sec  - myTime1.ru_stime.tv_sec)) +
       		((float) (myTime_distEnd.ru_stime.tv_usec - myTime1.ru_stime.tv_usec)) * 1e-6;


	userTime_dist = userTime;
	sysTime_dist = sysTime;
#endif

	bucket = anatomy_hashTupleWithSenValue(aData, noOfSenValue);

	proEqClass = (equivalenceClass *) malloc(sizeof(equivalenceClass));
	memset(proEqClass, 0, sizeof(equivalenceClass));

	bucketIndexList = anatomy_getLLargestBucket(bucket, l, noOfSenValue);
	while (bucketIndexList != NULL)
	{
		anEqTuple = (equivalenceTuple *) malloc(sizeof(equivalenceTuple));
		memset(anEqTuple, 0, sizeof(equivalenceTuple));

		newEltList = (int *) malloc(sizeof(int)*l);
		for (i = 0; i < l; i++)
		{
			bucketIndex = bucketIndexList[i];

			aBucket = bucket[bucketIndex];
			
			noOfElt = aBucket->noOfElt;
			newEltList[i] = aBucket->elt[noOfElt-1];

			// remove the last elt (virtually)
			aBucket->noOfElt--;
		}

		anEqTuple->member = newEltList;
		anEqTuple->noOfMember = l;

		newSenValue = (dimension *) malloc(sizeof(dimension)*l);
		for (i = 0; i < l; i++)
		{
			anElt = newEltList[i];
			aSenValue = tupleList[anElt].content[noOfDim-1].value;

			newSenValue[i].level = 0;
			newSenValue[i].value = aSenValue;
		}
		anEqTuple->senValue = newSenValue;
		
		insertEqTupleIntoEqClass(proEqClass, anEqTuple);

		free(bucketIndexList);

		bucketIndexList = anatomy_getLLargestBucket(bucket, l, noOfSenValue);
	}

	if (bucketIndexList == NULL)
	{
		for (i = 0; i < noOfSenValue; i++)
		{
			aBucket = bucket[i];
			if (aBucket->noOfElt != 0)
			{
				//for (j = 0; j < aBucket->noOfElt; j++)
				if (aBucket->noOfElt == 1)
				{
					//
					
					tupleID = aBucket->elt[0];

					eqID = anatomy_findEqTupleWithoutSenValue(proEqClass, aSenValue);

					if (eqID == -1)
					{
						// randomly assign
						eqID = 0;
					}

					// insert it to eqID
					anEqTuple = proEqClass->eqClass[eqID];

					anEqTuple->member = (int *) realloc(anEqTuple->member, sizeof(int)*(anEqTuple->noOfMember+1));
					anEqTuple->senValue = (dimension *) realloc(anEqTuple->senValue, sizeof(dimension)*(anEqTuple->noOfMember+1));

					anEqTuple->member[anEqTuple->noOfMember] = tupleID;
					anEqTuple->senValue[anEqTuple->noOfMember].level = 0;
					anEqTuple->senValue[anEqTuple->noOfMember].value = aSenValue;
					anEqTuple->noOfMember++;
				}
				else
				{
					

					anEqTuple = (equivalenceTuple *) malloc(sizeof(equivalenceTuple));
					memset(anEqTuple, 0, sizeof(equivalenceTuple));

					anEqTuple->member = (int *) malloc(sizeof(int)*aBucket->noOfElt);
					anEqTuple->senValue = (dimension *) malloc(sizeof(dimension)*aBucket->noOfElt);

					for (j = 0; j < aBucket->noOfElt; j++)
					{
						tupleID = aBucket->elt[j];

						anEqTuple->member[j] = tupleID;
						anEqTuple->senValue[j].level = 0;
						anEqTuple->senValue[j].value = aSenValue;
					}

					anEqTuple->noOfMember = aBucket->noOfElt;

					insertEqTupleIntoEqClass(proEqClass, anEqTuple);
					
					
				}
			
			}
		}
	}

	//
	debugFilePtr = (FILE *) fopen(debugFilename, "w");
	if (debugFilePtr == NULL)
	{
		printf("File %s cannot be opened!\n", debugFilename);
		exit(0);
	}

	//
	//printf("Final equivalence class...\n");
	//printEquivalenceClass(debugFilePtr, proEqClass, noOfDim);

	//
	equivalenceClass_generateGeneralizedValue(proEqClass, aData, genAllTable, noOfDim);
	createSenValueInEqClass(proEqClass, aData, noOfSenValue);
	eq_createSenIndexForAllEqTuple(proEqClass, config, noOfSenValue, noOfSenIndex);


	// transform
	printf("Transform an equivalence class to mechGroup...\n");
	createSenValueInEqClass(proEqClass, aData, noOfSenValue);
	copySenValueToNewSenValueInEqClass(proEqClass, noOfSenValue);
	finalMechAllGroup = transformFromEqClassToMechAllGroup(proEqClass, noOfSenValue);
#ifndef _WIN32
	getrusage(RUSAGE_SELF, &myTime2);	
#endif
	//
	//allCondition *predicate_generateAllCondition(int noOfCondition, int queryNoOfDim, int noOfQID, double s, int *dimSize)
	printf("Generating predicates...\n");
	aAllCondition = predicate_generateAllCondition(noOfCondition, queryNoOfDim, noOfQID, s, aData->dimSize);

	// generate two tables
	printf("Generating two tables...\n");
	finalTable = generateTwoTableFromMechGroup(finalMechAllGroup, aData);

	// printing
	//printTwoTable(debugFilePtr, finalTable);

	// calculate the estimated count and actual count
	avgRelError = calculateAvgRelativeError(finalMechAllGroup, aAllCondition, aData, noOfQID);

	printf("Average relative error : %.2f\n", avgRelError);

	

	// calculate the coverage error (information loss and distortion)
	//void calculateQIDInfoLossAndSenDistortion(int noOfQID, int totalNoOfTuple, equivalenceClass *eqClass, generalizedAllTable *genAllTable, double *infoLoss, double *distortion)
	calculateQIDInfoLossAndSenDistortion(noOfQID, config->noOfTuple, proEqClass, genAllTable, &infoLoss, &distortion, &noOfSenChange);


	printf("Information Loss : %.2f\n", infoLoss);
	printf("Distortion       : %.2f\n", distortion);
	printf("No. of Sen Change: %d\n", noOfSenChange);


	avgRelError_filePtr = (FILE *) fopen(avgRelErrorFilename, "w");
	fprintf(avgRelError_filePtr, "%.6f\n", avgRelError);
	fprintf(avgRelError_filePtr, "%.6f\n", infoLoss);
	fprintf(avgRelError_filePtr, "%.6f\n", distortion);
	fprintf(avgRelError_filePtr, "%d\n", noOfSenChange);
	

	// assume one senIndex
	senIndex = 1;
	avgEqTupleSize = ass_findAvgEqTupleSize(proEqClass);
	avgSenEqTupleSize = ass_findAvgSenEqTupleSize(proEqClass, senIndex);
	fprintf(avgRelError_filePtr, "%.6f\n", avgEqTupleSize);
	fprintf(avgRelError_filePtr, "%.6f\n", avgSenEqTupleSize);

	globalDistDiff = ass_findAvgGlobalDistDiff(proEqClass, senIndex, allSig, allSigMapping, m);
	globalDistDiffSen = ass_findAvgGlobalDistDiffSen(proEqClass, senIndex, allSig, allSigMapping, m);
	fprintf(avgRelError_filePtr, "%.10f\n", globalDistDiff);
	fprintf(avgRelError_filePtr, "%.10f\n", globalDistDiffSen);

	//
	noOfEqTupleSenValue = anatomy_findNoOfEqTupleWithSenIndex(proEqClass, senIndex);
	noOfViolate = anatomy_findNoOfViolateEqTuple(proEqClass, senIndex, allSig, allSigMapping, m, aData->noOfTuple);
	fprintf(avgRelError_filePtr, "\n");
	fprintf(avgRelError_filePtr, "%d\n", proEqClass->noOfEqClass);	// total. no. of eq tuples
	fprintf(avgRelError_filePtr, "%d\n", noOfEqTupleSenValue);			// total no. of eq tuples with sensitive values
	fprintf(avgRelError_filePtr, "%d\n", noOfViolate);				// total no. of eq. tuples with privacy breach


	printf("\n");
	printf("Avg. eq. tuple size = %.6f\n", avgEqTupleSize);
	printf("Avg. sen. eq. tuple size = %.6f\n", avgSenEqTupleSize);
	printf("Global dist. diff = %.10f\n", globalDistDiff);
	printf("Global dist. diff (sen) = %.10f\n", globalDistDiffSen);
	printf("Total no. of eq. classes = %d\n", proEqClass->noOfEqClass);
	printf("Total no. of eq. classes with sen. values = %d\n", noOfEqTupleSenValue);
	printf("Total no. of eq. classes with privacy breach = %d\n", noOfViolate);
	printf("\n");


	fclose(avgRelError_filePtr);


#ifndef _WIN32
	userTime =
     		((float) (myTime2.ru_utime.tv_sec  - myTime1.ru_utime.tv_sec)) +
       		((float) (myTime2.ru_utime.tv_usec - myTime1.ru_utime.tv_usec)) * 1e-6;
	sysTime =
       		((float) (myTime2.ru_stime.tv_sec  - myTime1.ru_stime.tv_sec)) +
       		((float) (myTime2.ru_stime.tv_usec - myTime1.ru_stime.tv_usec)) * 1e-6;

	printf("User time : %f seconds\n",userTime);
	printf("System time : %f seconds\n\n",sysTime);

	fprintf(debugFilePtr, "Execution Time\n");
	fprintf(debugFilePtr, "--------------\n");
	fprintf(debugFilePtr, "User time : %f seconds\n",userTime);
    fprintf(debugFilePtr, "System time : %f seconds\n\n",sysTime);


	timeFilePtr = (FILE *) fopen(timeFilename, "w");
	fprintf(timeFilePtr, "%.6f\n",userTime);
    fprintf(timeFilePtr, "%.6f\n\n",sysTime);
	fprintf(timeFilePtr, "%.6f\n",userTime_dist);
    fprintf(timeFilePtr, "%.6f\n\n",sysTime_dist);
	fprintf(timeFilePtr, "%.6f\n",userTime-userTime_dist);
    fprintf(timeFilePtr, "%.6f\n\n",sysTime-sysTime_dist);
	fclose(timeFilePtr);
#endif

	fclose(debugFilePtr);
}



equivalenceClass *anatomy_main_output(data *aData, configure *config, generalizedAllTable *genAllTable)
{
	mySet **bucket;

	int noOfSenValue;
	int noOfDim;

	int l;

	int m;

	//int noOfViolate;

	int *bucketIndexList;
	int bucketIndex;

	tuple *tupleList;

	mySet *aBucket;
	//signatureGroupGroupGroup *allSig;

	int noOfElt;
	int anElt;

	int tupleID;

	//int ***allSigMapping;

	//FILE *avgRelError_filePtr;
	

	//allCondition *aAllCondition;
	int noOfQID;

	//twoTable *finalTable;

	//double avgRelError;

	//double avgEqTupleSize, avgSenEqTupleSize;

	//mechAllGroup *finalMechAllGroup;

	dimension *newSenValue;

	int aSenValue;
	int *newEltList;

	//FILE *debugFilePtr;
	char *debugFilename = "debug.txt";

	int i, j;

	int eqID;
	//int senIndex;

	//int dummy;
	//double infoLoss, distortion;
	//int noOfSenChange;

	//double globalDistDiff;
	//double globalDistDiffSen;

	equivalenceTuple *anEqTuple;

	int threshold;
	int noOfSenIndex;

	//int noOfEqTupleSenValue;

	equivalenceClass *proEqClass;

	char *avgRelErrorFilename = "error.txt";
	char *timeFilename = "time.txt";

	int queryNoOfDim = config->queryNoOfDim;
	double s = config->s;
	int noOfCondition = config->noOfCondition;

	int noOfBucket;
	int avgSize;
	int eltNo;

#ifndef _WIN32
	FILE *timeFilePtr;

	float  userTime, sysTime;
	float userTime_dist, sysTime_dist;
	struct rusage myTime1, myTime2, myTime_distEnd;

	getrusage(RUSAGE_SELF,&myTime1);
#endif

	noOfDim = aData->noOfDim;
	noOfSenValue = aData->dimSize[noOfDim-1];
	noOfQID = noOfDim-1;
	m = config->d;

	tupleList = aData->tupleList;

	l = config->d;


	// signature
	threshold = calViolation_calculateSampleSize(config->error, config->sigma);
	//threshold = 1;
	printf(" Sample threshold = %d\n", threshold);
	

	// assume only one sensitive index
	noOfSenIndex = 2;



	// generate all signature
	//printf("Generating signatures...\n");
	//allSig = sig_findAllSignature(aData, threshold);

	//sig_printStatistics(stdout, allSig);

	// find distribution
	//printf("Finding the global distributions...\n");
	//allSigMapping = sig_findGlobalDistributionFromOrgTable(config, allSig, aData, noOfSenIndex);

	// special case
	if (noOfSenValue < config->d)
	{
		noOfBucket = config->d;
	}
	else
	{
		noOfBucket = noOfSenValue;
	}


	bucket = anatomy_hashTupleWithSenValue(aData, noOfBucket);

	// special
	if (noOfSenValue < config->d)
	{
		// special case
		// consider only two possible sen. values

		avgSize = bucket[0]->noOfElt/(noOfBucket-1);

		for (i = 2; i < noOfBucket; i++)
		{
			 for (j = 0; j < avgSize; j++)
			 {
				 noOfElt = bucket[0]->noOfElt;
				 eltNo = bucket[0]->elt[noOfElt-1];
				 bucket[0]->noOfElt--;

				 mySet_insert(bucket[i], eltNo);
			 }
		}
	}

	proEqClass = (equivalenceClass *) malloc(sizeof(equivalenceClass));
	memset(proEqClass, 0, sizeof(equivalenceClass));

	bucketIndexList = anatomy_getLLargestBucket(bucket, l, noOfBucket);
	while (bucketIndexList != NULL)
	{
		anEqTuple = (equivalenceTuple *) malloc(sizeof(equivalenceTuple));
		memset(anEqTuple, 0, sizeof(equivalenceTuple));

		newEltList = (int *) malloc(sizeof(int)*l);
		for (i = 0; i < l; i++)
		{
			bucketIndex = bucketIndexList[i];

			aBucket = bucket[bucketIndex];
			
			noOfElt = aBucket->noOfElt;
			newEltList[i] = aBucket->elt[noOfElt-1];

			// remove the last elt (virtually)
			aBucket->noOfElt--;
		}

		anEqTuple->member = newEltList;
		anEqTuple->noOfMember = l;

		newSenValue = (dimension *) malloc(sizeof(dimension)*l);
		for (i = 0; i < l; i++)
		{
			anElt = newEltList[i];
			aSenValue = tupleList[anElt].content[noOfDim-1].value;

			newSenValue[i].level = 0;
			newSenValue[i].value = aSenValue;
		}
		anEqTuple->senValue = newSenValue;
		
		insertEqTupleIntoEqClass(proEqClass, anEqTuple);

		free(bucketIndexList);

		bucketIndexList = anatomy_getLLargestBucket(bucket, l, noOfBucket);
	}

	if (bucketIndexList == NULL)
	{
		for (i = 0; i < noOfSenValue; i++)
		{
			aSenValue = i;

			aBucket = bucket[i];
			if (aBucket->noOfElt != 0)
			{
				//for (j = 0; j < aBucket->noOfElt; j++)
				if (aBucket->noOfElt == 1)
				{
					//
					
					tupleID = aBucket->elt[0];

					eqID = anatomy_findEqTupleWithoutSenValue(proEqClass, aSenValue);

					if (eqID == -1)
					{
						// randomly assign
						eqID = 0;
					}

					// insert it to eqID
					anEqTuple = proEqClass->eqClass[eqID];

					anEqTuple->member = (int *) realloc(anEqTuple->member, sizeof(int)*(anEqTuple->noOfMember+1));
					anEqTuple->senValue = (dimension *) realloc(anEqTuple->senValue, sizeof(dimension)*(anEqTuple->noOfMember+1));

					anEqTuple->member[anEqTuple->noOfMember] = tupleID;
					anEqTuple->senValue[anEqTuple->noOfMember].level = 0;
					anEqTuple->senValue[anEqTuple->noOfMember].value = aSenValue;
					anEqTuple->noOfMember++;
				}
				else
				{
					

					anEqTuple = (equivalenceTuple *) malloc(sizeof(equivalenceTuple));
					memset(anEqTuple, 0, sizeof(equivalenceTuple));

					anEqTuple->member = (int *) malloc(sizeof(int)*aBucket->noOfElt);
					anEqTuple->senValue = (dimension *) malloc(sizeof(dimension)*aBucket->noOfElt);

					for (j = 0; j < aBucket->noOfElt; j++)
					{
						tupleID = aBucket->elt[j];

						anEqTuple->member[j] = tupleID;
						anEqTuple->senValue[j].level = 0;
						anEqTuple->senValue[j].value = aSenValue;
					}

					anEqTuple->noOfMember = aBucket->noOfElt;

					insertEqTupleIntoEqClass(proEqClass, anEqTuple);
					
					
				}
			
			}
		}
	}

	//
	//debugFilePtr = (FILE *) fopen(debugFilename, "w");
	//if (debugFilePtr == NULL)
	//{
	//	printf("File %s cannot be opened!\n", debugFilename);
	//	exit(0);
	//}

	//
	//printf("Final equivalence class...\n");
	//printEquivalenceClass(debugFilePtr, proEqClass, noOfDim);

	//
	equivalenceClass_generateGeneralizedValue(proEqClass, aData, genAllTable, noOfDim);
	createSenValueInEqClass(proEqClass, aData, noOfSenValue);
	eq_createSenIndexForAllEqTuple(proEqClass, config, noOfSenValue, noOfSenIndex);


	return proEqClass;

}


