
#include "ass.h"

extern int ceiling(double no);

int ass_calculateMinimumEqSizeForOneAttrSet(float triangle, int lo, float f)
{
	int returnValue;

	double temp;

	if (f == triangle)
	{
		returnValue = -1;
	}
	else
	{
		temp = (triangle*(lo - 1.0)*f/(1.0 - f) - triangle + lo*f)/(f - triangle);
		returnValue = ceiling(temp);
	}

	return returnValue;
}

void ass_calculateFDistDiffForOneAttrSetInEqTuple(equivalenceTuple *aEqTuple, int sizeIndex, int typeIndex, int senIndex, signatureGroup *aSigGroup, int ***allSigMapping, float *triangle, float *f_max)
{

	signature **sigList;

	float f;
	float f_min;

	int memberID;
	int sigID;

	int noOfMember;
	int k;

	signature *aSig;

	float aProb;
	int *member;

	noOfMember = aEqTuple->noOfMember;
	member = aEqTuple->member;

	//noOfSig = aSigGroup->noOfSig;
	sigList = aSigGroup->sigList;
	
	// find the greatest distribution
	// find the smallest distribution
	//f = (float) (1.0/noOfMember);
	//f_min = (float) (1.0/noOfMember);
	f = 0;
	f_min = 1;
	for (k = 0; k < noOfMember; k++)
	{
		memberID = member[k];
		sigID = allSigMapping[memberID][sizeIndex][typeIndex];
		
		if (sigID == -1)
		{
			// do nothing
			aProb = (float) (1.0/noOfMember); 
		}
		else
		{
			aSig = sigList[sigID];
			
			aProb = aSig->prob[senIndex];
		
		}

		if (aProb > f)
		{
			f = aProb;
		}
		if (aProb < f_min)
		{
			f_min = aProb;
		}
	}
	(*triangle) = f - f_min;
	(*f_max) = f;
}


void ass_calculateFDistDiffForOneAttrSetWithTwoEqTuple(equivalenceTuple *aEqTuple1, equivalenceTuple *aEqTuple2, int sizeIndex, int typeIndex, int senIndex, signatureGroup *aSigGroup, int ***allSigMapping, float *triangle, float *f_max)
{
	signature **sigList;

	float f;
	float f_min;

	int memberID;
	int sigID;

	int noOfMember1;
	int noOfMember2;
	int k;

	signature *aSig;

	float aProb;
	int *member1;
	int *member2;

	int noOfFinalMember;

	noOfMember1 = aEqTuple1->noOfMember;
	member1 = aEqTuple1->member;

	noOfMember2 = aEqTuple2->noOfMember;
	member2 = aEqTuple2->member;

	noOfFinalMember = noOfMember1 + noOfMember2;

	//noOfSig = aSigGroup->noOfSig;
	sigList = aSigGroup->sigList;
	
	// find the greatest distribution
	// find the smallest distribution
	f = (float) (1.0/noOfFinalMember);
	f_min = (float) (1.0/noOfFinalMember);

	// eqTuple1
	for (k = 0; k < noOfMember1; k++)
	{
		memberID = member1[k];
		sigID = allSigMapping[memberID][sizeIndex][typeIndex];
		
		if (sigID == -1)
		{
			// do nothing
		}
		else
		{
			aSig = sigList[sigID];
			
			aProb = aSig->prob[senIndex];
			
			if (aProb > f)
			{
				f = aProb;
			}
			if (aProb < f_min)
			{
				f_min = aProb;
			}
		}
	}

	// eqTuple2
	for (k = 0; k < noOfMember2; k++)
	{
		memberID = member2[k];
		sigID = allSigMapping[memberID][sizeIndex][typeIndex];
		
		if (sigID == -1)
		{
			// do nothing
		}
		else
		{
			aSig = sigList[sigID];
			
			aProb = aSig->prob[senIndex];
			
			if (aProb > f)
			{
				f = aProb;
			}
			if (aProb < f_min)
			{
				f_min = aProb;
			}
		}
	}

	(*triangle) = f - f_min;
	(*f_max) = f;
}


// assume only the eq. tuple which has at most one sensitive value for each possible sen. value
// also assume it has at least one sensitive value
float ass_calculateDistDiffForAllAttrSetInEqTuple(equivalenceTuple *aEqTuple, int senIndex, signatureGroupGroupGroup *allSig, int ***allSigMapping, int m)
{
	float returnValue;

	float triangle;
	float f;

	int i, j;

	int noOfSize;
	signatureGroupGroup **sizeList;

	signatureGroup *aSigGroup;

	signatureGroupGroup *aSigGroupGroup;
	signatureGroup **groupList;
	int noOfGroup;

	//
	noOfSize = allSig->noOfSize;
	sizeList = allSig->sizeList;

	returnValue = 0;
	for (i = 0; i < noOfSize; i++)
	{
		aSigGroupGroup = sizeList[i];

		noOfGroup = aSigGroupGroup->noOfGroup;
		groupList = aSigGroupGroup->groupList;

		for (j = 0; j < noOfGroup; j++)
		{
			aSigGroup = groupList[j];

			ass_calculateFDistDiffForOneAttrSetInEqTuple(aEqTuple, i, j, senIndex, aSigGroup, allSigMapping, &triangle, &f);

			returnValue += triangle;
		}
	}

	return returnValue;
}

// assume only the eq. tuple which has at most one sensitive value for each possible sen. value
// also assume it has at least one sensitive value
float ass_calculateDistDiffForAllAttrSetWithTwoEqTuple(equivalenceTuple *aEqTuple1, equivalenceTuple *aEqTuple2, int senIndex, signatureGroupGroupGroup *allSig, int ***allSigMapping, int m)
{
	float returnValue;

	float triangleCombine;
	//float triangleSingle;
	float fCombine;
	//float fSingle;

	int i, j;

	int noOfSize;
	signatureGroupGroup **sizeList;

	signatureGroup *aSigGroup;

	signatureGroupGroup *aSigGroupGroup;
	signatureGroup **groupList;
	int noOfGroup;

	//
	noOfSize = allSig->noOfSize;
	sizeList = allSig->sizeList;

	returnValue = 0;
	for (i = 0; i < noOfSize; i++)
	{
		aSigGroupGroup = sizeList[i];

		noOfGroup = aSigGroupGroup->noOfGroup;
		groupList = aSigGroupGroup->groupList;

		for (j = 0; j < noOfGroup; j++)
		{
			aSigGroup = groupList[j];

			ass_calculateFDistDiffForOneAttrSetWithTwoEqTuple(aEqTuple1, aEqTuple2, i, j, senIndex, aSigGroup, allSigMapping, &triangleCombine, &fCombine);
			//ass_calculateFDistDiffForOneAttrSetInEqTuple(aEqTuple1, i, j, senIndex, aSigGroup, allSigMapping, &triangleSingle, &fSingle);

			//returnValue += (triangleCombine - triangleSingle);
			returnValue += triangleCombine;
		}
	}

	return returnValue;
}




// assume only the eq. tuple which has at most one sensitive value for each possible sen. value
// also assume it has at least one sensitive value
int ass_calculateMinimumEqSizeForAllAttrSetInEqTuple(equivalenceTuple *aEqTuple, int senIndex, signatureGroupGroupGroup *allSig, int ***allSigMapping, int m, int noOfTuple)
{
	int returnValue;

	float triangle;
	int lo;
	float f;

	int expectedSize;

	int i, j;

	int noOfMember;
	int *member;

	int noOfSize;
	signatureGroupGroup **sizeList;

	signatureGroup *aSigGroup;

	signatureGroupGroup *aSigGroupGroup;
	signatureGroup **groupList;
	int noOfGroup;

	noOfMember = aEqTuple->noOfMember;
	member = aEqTuple->member;

	// do the random distribution first
	triangle = 0;
	lo = m;
	f = (float) (1.0/noOfMember);	

	returnValue = ass_calculateMinimumEqSizeForOneAttrSet(triangle, lo, f);

	//
	noOfSize = allSig->noOfSize;
	sizeList = allSig->sizeList;

	for (i = 0; i < noOfSize; i++)
	{
		aSigGroupGroup = sizeList[i];

		noOfGroup = aSigGroupGroup->noOfGroup;
		groupList = aSigGroupGroup->groupList;

		for (j = 0; j < noOfGroup; j++)
		{
			aSigGroup = groupList[j];

			ass_calculateFDistDiffForOneAttrSetInEqTuple(aEqTuple, i, j, senIndex, aSigGroup, allSigMapping, &triangle, &f);

			// calculate expected size
			expectedSize = ass_calculateMinimumEqSizeForOneAttrSet(triangle, lo, f);

			if (expectedSize == -1)
			{
				// infinity
				expectedSize = noOfTuple;
			}


			if (expectedSize > returnValue)
			{
				returnValue = expectedSize;
			}
		}
	}

	return returnValue;
}


equivalenceTuple *ass_formOneIndividualEqTuple(tuple *aTuple, int id, int noOfDim)
{
	equivalenceTuple *returnValue;

	int i;

	returnValue = (equivalenceTuple *) malloc(sizeof(equivalenceTuple));
	memset(returnValue, 0, sizeof(equivalenceTuple));

	returnValue->content = (dimension *) malloc(sizeof(dimension)*noOfDim);
	for (i = 0; i < noOfDim; i++)
	{
		memcpy(returnValue->content, aTuple->content, sizeof(dimension)*noOfDim);
	}

	returnValue->noOfMember = 1;
	returnValue->member = (int *) malloc(sizeof(int));
	returnValue->member[0] = id;

	return returnValue;
}

equivalenceClass *ass_formAllIndividualEqTuple(data *aData)
{
	equivalenceClass *returnValue;
	equivalenceTuple *eqTuple;

	tuple *aTuple;

	tuple *tupleList;

	int noOfDim;

	int i;

	returnValue = new_equivalenceClass();

	noOfDim = aData->noOfDim;
	tupleList = aData->tupleList;
	for (i = 0; i < aData->noOfTuple; i++)
	{
		aTuple = &(tupleList[i]);

		eqTuple = ass_formOneIndividualEqTuple(aTuple, i, noOfDim);
		insertEqTupleIntoEqClassWithoutClone(returnValue, eqTuple);
	}

	return returnValue;
}



int ass_isEqTupleContainSomeSenIndexSet(equivalenceTuple *aEqTuple, int *senIndexSet, int noOfElt)
{
	int i;

	int aSenIndex;
	int *noOfEachSenIndex;

	int aNo;

	int returnValue;

	noOfEachSenIndex = aEqTuple->noOfEachSenIndex;

	returnValue = 0;
	for (i = 0; i < noOfElt; i++)
	{
		aSenIndex = senIndexSet[i];

		aNo = noOfEachSenIndex[aSenIndex];
		if (aNo != 0)
		{
			returnValue = 1;
			break;
		}
	}

	return returnValue;
}

// return a list of eq. tuple ID
int *ass_findAllEqTupleNotContainSenIndexSet(equivalenceClass *eqClass, int *senIndexSet, int noOfElt, int *noOfFinalEqTuple)
{
	int *returnValue;

	int i;

	int *tempList;
	int count;

	int noOfEqTuple;
	equivalenceTuple **eqTupleList;

	int isContainSome;

	equivalenceTuple *aEqTuple;

	noOfEqTuple = eqClass->noOfEqClass;
	eqTupleList = eqClass->eqClass;

	tempList = (int *) malloc(sizeof(int)*noOfEqTuple);

	count = 0;
	for (i = 0; i < noOfEqTuple; i++)
	{
		aEqTuple = eqTupleList[i];

		isContainSome = ass_isEqTupleContainSomeSenIndexSet(aEqTuple, senIndexSet, noOfElt);
		if (isContainSome == 0)
		{
			// insert
			tempList[count] = i;
			count++;
		}
	}

	// finalize the list
	returnValue = (int *) malloc(sizeof(int)*count);
	memcpy(returnValue, tempList, sizeof(int)*count);
	(*noOfFinalEqTuple) = count;

	free(tempList);

	return returnValue;
}

// -1: when we cannot find any valid closest eq. tuple
// assume one sen. index
int ass_findClosestValidEqTuple(equivalenceClass *eqClass, equivalenceTuple *oneEqTuple, int aSenIndex, signatureGroupGroupGroup *allSig, int ***allSigMapping, int m)
{
	int returnValue;

	float LARGEVALUE = (float) 99999999;

	int eqID_min;
	int noOfEqTupleNoSenIndexSet;
	int noOfElt;
	int *senIndexSet;

	int *eqTupleListNoSenIndexSet;

	int eqID;

	equivalenceTuple *aPossEqTuple;

	int i;
	float distDiff;
	float distDiffMin;

	equivalenceTuple **eqTupleList;

	eqTupleList = eqClass->eqClass;

	// generate senIndexSet
	noOfElt = 1;
	senIndexSet = (int *) malloc(sizeof(int));
	senIndexSet[0] = 1;

	eqTupleListNoSenIndexSet = ass_findAllEqTupleNotContainSenIndexSet(eqClass, senIndexSet, noOfElt, &noOfEqTupleNoSenIndexSet);

	distDiffMin = LARGEVALUE;
	eqID_min = -1;
	for (i = 0; i < noOfEqTupleNoSenIndexSet; i++)
	{
		eqID = eqTupleListNoSenIndexSet[i];

		aPossEqTuple = eqTupleList[eqID];

		// calculate distance
		distDiff = ass_calculateDistDiffForAllAttrSetWithTwoEqTuple(oneEqTuple, aPossEqTuple, aSenIndex, allSig, allSigMapping, m);

		if (distDiff < distDiffMin)
		{
			distDiffMin = distDiff;
			eqID_min = eqID;
		}
	}

	
	returnValue = eqID_min;

	if (noOfEqTupleNoSenIndexSet > 0)
	{
		free(eqTupleListNoSenIndexSet);
	}
	free(senIndexSet);

	return returnValue;
}


// -1: when we cannot find any valid closest eq. tuple
// assume one sen. index
int *ass_findClosestValidEqTuple_additional(equivalenceClass *eqClass, equivalenceTuple *oneEqTuple, int aSenIndex, signatureGroupGroupGroup *allSig, int ***allSigMapping, int m, int addNoOfTuple)
{
	int *returnValue;

	float LARGEVALUE = (float) 99999999;

	int eqID_min;
	int noOfEqTupleNoSenIndexSet;
	int noOfElt;
	int *senIndexSet;

	int *eqTupleListNoSenIndexSet;

	int eqID;

	equivalenceTuple *aPossEqTuple;

	int i;
	double *distDiff;
	double *rev_distDiff;

	float distDiffMin;

	int *index;
	int *rev_index;

	int anIndex;
	int oneEqID;

	equivalenceTuple **eqTupleList;

	eqTupleList = eqClass->eqClass;

	// generate senIndexSet
	noOfElt = 1;
	senIndexSet = (int *) malloc(sizeof(int));
	senIndexSet[0] = 1;

	eqTupleListNoSenIndexSet = ass_findAllEqTupleNotContainSenIndexSet(eqClass, senIndexSet, noOfElt, &noOfEqTupleNoSenIndexSet);

	if (noOfEqTupleNoSenIndexSet > 0)
	{
		if (noOfEqTupleNoSenIndexSet >= addNoOfTuple)
		{
			distDiff = (double *) malloc(sizeof(double)*noOfEqTupleNoSenIndexSet);
			index = (int *) malloc(sizeof(int)*noOfEqTupleNoSenIndexSet);
			memset(distDiff, 0, sizeof(double)*noOfEqTupleNoSenIndexSet);
			memset(index, 0, sizeof(int)*noOfEqTupleNoSenIndexSet);

			rev_distDiff = (double *) malloc(sizeof(double)*noOfEqTupleNoSenIndexSet);
			rev_index = (int *) malloc(sizeof(int)*noOfEqTupleNoSenIndexSet);
			memset(rev_distDiff, 0, sizeof(double)*noOfEqTupleNoSenIndexSet);
			memset(rev_index, 0, sizeof(int)*noOfEqTupleNoSenIndexSet);

			distDiffMin = LARGEVALUE;
			eqID_min = -1;
			for (i = 0; i < noOfEqTupleNoSenIndexSet; i++)
			{
				index[i] = i;

				eqID = eqTupleListNoSenIndexSet[i];

				aPossEqTuple = eqTupleList[eqID];

				// calculate distance
				distDiff[i] = ass_calculateDistDiffForAllAttrSetWithTwoEqTuple(oneEqTuple, aPossEqTuple, aSenIndex, allSig, allSigMapping, m);

				//if (distDiff < distDiffMin)
				//{
				//	distDiffMin = distDiff;
				//	eqID_min = eqID;
				//}
			}

			// quick sort
			quicksort(index, distDiff, 0, noOfEqTupleNoSenIndexSet-1);

			// reverse
			for (i = 0; i < noOfEqTupleNoSenIndexSet; i++)
			{
				rev_index[i] = index[noOfEqTupleNoSenIndexSet-i-1];
				rev_distDiff[i] = distDiff[noOfEqTupleNoSenIndexSet-i-1];
			}

			returnValue = (int *) malloc(sizeof(int)*addNoOfTuple);
			for (i = 0; i < addNoOfTuple; i++)
			{
				anIndex = rev_index[i];
				oneEqID = eqTupleListNoSenIndexSet[anIndex];
				returnValue[i] = oneEqID;
			}

			free(distDiff);
			free(index);
			free(rev_distDiff);
			free(rev_index);
		}
		else
		{
			returnValue = NULL;
		}
		
	}
	else
	{
		returnValue = NULL;
	}


	
	//returnValue = eqID_min;

	if (noOfEqTupleNoSenIndexSet > 0)
	{
		free(eqTupleListNoSenIndexSet);
	}
	free(senIndexSet);

	return returnValue;
}

// 1 : violate
// 0 : not violate
int ass_isViolateGlobalDistBoundConditionInOneEqTuple(equivalenceTuple *aEqTuple, int senIndex, signatureGroupGroupGroup *allSig, int ***allSigMapping, int m, int noOfTuple)
{
	int returnValue;

	int minSize;

	minSize = ass_calculateMinimumEqSizeForAllAttrSetInEqTuple(aEqTuple, senIndex, allSig, allSigMapping, m, noOfTuple);	

	if (minSize > aEqTuple->noOfMember)
	{
		returnValue = 1;
	}
	else
	{
		returnValue = 0;
	}

	return returnValue;
}


// 1 : violate
// 0 : not violate
// additionalNoOfTuple
int ass_isViolateGlobalDistBoundConditionInOneEqTuple_additional(equivalenceTuple *aEqTuple, int senIndex, signatureGroupGroupGroup *allSig, int ***allSigMapping, int m, int noOfTuple, int *additionalNoOfTuple)
{
	int returnValue;

	int minSize;

	minSize = ass_calculateMinimumEqSizeForAllAttrSetInEqTuple(aEqTuple, senIndex, allSig, allSigMapping, m, noOfTuple);	

	if (minSize > aEqTuple->noOfMember)
	{
		returnValue = 1;
		(*additionalNoOfTuple) = minSize - aEqTuple->noOfMember;
	}
	else
	{
		returnValue = 0;
	}

	return returnValue;
}



// 1: success
// 0: fail
int ass_mergeEqTupleWithOthersForSatisfyPrivacyRequirement(int violateEqID, equivalenceClass *eqClass, int aSenIndex, signatureGroupGroupGroup *allSig, int ***allSigMapping, int m, int noOfDim, generalizedAllTable *genAllTable, int noOfSenValue, int noOfSenIndex, int noOfTuple)
{
	int returnValue;
	int eqID_closest;

	equivalenceTuple *oneEqTuple;
	equivalenceTuple *closestEqTuple;
	equivalenceTuple *newEqTuple;

	int isViolate;

	oneEqTuple = eqClass->eqClass[violateEqID];

	isViolate = ass_isViolateGlobalDistBoundConditionInOneEqTuple(oneEqTuple, aSenIndex, allSig, allSigMapping, m, noOfTuple);

	returnValue = 1;
	while (isViolate == 1)
	{
		// find closest
		eqID_closest = ass_findClosestValidEqTuple(eqClass, oneEqTuple, aSenIndex, allSig, allSigMapping, m);

		if (eqID_closest == -1)
		{
			// do nothing
			returnValue = 0;
			break;
		}
		else
		{
			closestEqTuple = eqClass->eqClass[eqID_closest];

			// merget two
			newEqTuple = mergeTwoEquivalenceTupleWithSenAndSenIndex(noOfDim, genAllTable, oneEqTuple, closestEqTuple, noOfSenValue, noOfSenIndex);

			// remove two and insert new merged eq. tuple
			// assume insert at the end of list
			eq_removeTwoEqTupleAndAddOneEqTupleToEqClass(eqClass, violateEqID, eqID_closest, newEqTuple);

			returnValue = 1;

			// free oneEqTuple and free closestEqTuple
			freeEquivalenceTupleWithSenValueAndSenIndex(oneEqTuple);
			freeEquivalenceTupleWithSenValueAndSenIndex(closestEqTuple);
		}

		oneEqTuple = newEqTuple;
		violateEqID = eqClass->noOfEqClass-1;

		isViolate = ass_isViolateGlobalDistBoundConditionInOneEqTuple(oneEqTuple, aSenIndex, allSig, allSigMapping, m, noOfTuple);
	}

	return returnValue;
	
}


// 1: success
// 0: fail
int ass_mergeEqTupleWithOthersForSatisfyPrivacyRequirement_additional(int violateEqID, equivalenceClass *eqClass, int aSenIndex, signatureGroupGroupGroup *allSig, int ***allSigMapping, int m, int noOfDim, generalizedAllTable *genAllTable, int noOfSenValue, int noOfSenIndex, int noOfTuple)
{
	int returnValue;
	int *eqID_closestSet;

	equivalenceTuple *oneEqTuple;
	equivalenceTuple *closestEqTuple;
	equivalenceTuple *newEqTuple;

	int isViolate;

	int addNoOfTuple;

	int i;//, index;

	int oneEqID;
	int *index;
	int *value;
	int aEqIDIndex;

	oneEqTuple = eqClass->eqClass[violateEqID];

	isViolate = ass_isViolateGlobalDistBoundConditionInOneEqTuple_additional(oneEqTuple, aSenIndex, allSig, allSigMapping, m, noOfTuple, &addNoOfTuple);

	returnValue = 1;
	while (isViolate == 1)
	{
		// find closest
		eqID_closestSet = ass_findClosestValidEqTuple_additional(eqClass, oneEqTuple, aSenIndex, allSig, allSigMapping, m, addNoOfTuple);

		if (eqID_closestSet == NULL)
		{
			// do nothing
			returnValue = 0;
			break;
		}
		else
		{
			for (i = 0; i < addNoOfTuple; i++)
			{
				oneEqID = eqID_closestSet[i];
				closestEqTuple = eqClass->eqClass[oneEqID];

				// merget two
				newEqTuple = mergeTwoEquivalenceTupleWithSenAndSenIndex(noOfDim, genAllTable, oneEqTuple, closestEqTuple, noOfSenValue, noOfSenIndex);

				// remove two and insert new merged eq. tuple
				// assume insert at the end of list
				//eq_removeTwoEqTupleAndAddOneEqTupleToEqClass(eqClass, violateEqID, eqID_closest, newEqTuple);

				

				if (i != 0)
				{
					// free oneEqTuple and free closestEqTuple
					freeEquivalenceTupleWithSenValueAndSenIndex(oneEqTuple);
					//freeEquivalenceTupleWithSenValueAndSenIndex(closestEqTuple);
				}

				oneEqTuple = newEqTuple;
			}

			// quicksort
			index = (int *) malloc(sizeof(int)*(addNoOfTuple+1));
			value = (int *) malloc(sizeof(int)*(addNoOfTuple+1));
			for (i = 0; i < addNoOfTuple+1; i++)
			{
				index[i] = i;
			}
			for (i = 0; i < addNoOfTuple; i++)
			{
				value[i] = eqID_closestSet[i];
			}
			value[addNoOfTuple] = violateEqID;

			quicksort_i(index, value, 0, addNoOfTuple);

			// remove
			for (i = 0; i < addNoOfTuple+1; i++)
			{
				aEqIDIndex = value[i];
				closestEqTuple = eqClass->eqClass[aEqIDIndex];

				removeEqTupleFromEqClass(eqClass, aEqIDIndex);
				freeEquivalenceTupleWithSenValueAndSenIndex(closestEqTuple);
			}

			// insert new
			insertEqTupleIntoEqClassWithoutClone(eqClass, newEqTuple);

			// free
			free(index);
			free(value);

			returnValue = 1;
		}

		oneEqTuple = newEqTuple;
		violateEqID = eqClass->noOfEqClass-1;

		isViolate = ass_isViolateGlobalDistBoundConditionInOneEqTuple(oneEqTuple, aSenIndex, allSig, allSigMapping, m, noOfTuple);
	}

	return returnValue;
	
}

int ass_isContainSenIndex(equivalenceTuple *aEqTuple, int senIndex)
{
	int returnValue;

	if (aEqTuple->noOfEachSenIndex[senIndex] == 0)
	{
		returnValue = 0;
	}
	else
	{
		returnValue = 1;
	}

	return returnValue;
}

int ass_isContainSenIndex_considerNewSen(equivalenceTuple *aEqTuple, int senIndex, configure *config)
{
	int returnValue;
	int i;
	int total;

	if (aEqTuple->noOfEachSenIndex[senIndex] == 0)
	{
		returnValue = 0;
	}
	else
	{
		returnValue = 1;
	}

	if (returnValue == 0)
	{
		total = 0;
		for (i = config->senStartNo; i <= config->senEndNo; i++)
		{
			total += aEqTuple->noOfEachNewSen[i];
		}

		if (total == 0)
		{
			returnValue = 0;
		}
		else
		{
			returnValue = 1;
		}
	}

	return returnValue;
}

// if violate, also return the violateEqID
// if not, no need to return violateEqID
int ass_isViolateGlobalDistBoundConditionInAllEqTuple(equivalenceClass *eqClass, int senIndex, signatureGroupGroupGroup *allSig, int ***allSigMapping, int m, int *violateEqID, int noOfTuple)
{
	int returnValue;

	int noOfEqTuple;
	equivalenceTuple **eqTupleList;
	equivalenceTuple *aEqTuple;

	int i;

	int isContainSen;

	int isViolate;

	noOfEqTuple = eqClass->noOfEqClass;
	eqTupleList = eqClass->eqClass;

	returnValue = 0;
	for (i = 0; i < noOfEqTuple; i++)
	{
		aEqTuple = eqTupleList[i];

		isContainSen = ass_isContainSenIndex(aEqTuple, senIndex);

		if (isContainSen == 1)
		{
			isViolate = ass_isViolateGlobalDistBoundConditionInOneEqTuple(aEqTuple, senIndex, allSig, allSigMapping, m, noOfTuple);
			if (isViolate == 1)
			{
				returnValue = 1;
				(*violateEqID) = i;
				break;
			}
		}
	}

	return returnValue;
}


double ass_findAvgEqTupleSize(equivalenceClass *eqClass)
{
	double returnValue;

	int i;

	int noOfEqTuple;
	equivalenceTuple **eqTupleList;
	equivalenceTuple *anEqTuple;

	noOfEqTuple = eqClass->noOfEqClass;
	eqTupleList = eqClass->eqClass;

	returnValue = 0;
	for (i = 0; i < noOfEqTuple; i++)
	{
		anEqTuple = eqTupleList[i];

		returnValue += anEqTuple->noOfMember;
	}

	returnValue = returnValue/((double) noOfEqTuple);

	return returnValue;
}

double ass_findAvgSenEqTupleSize(equivalenceClass *eqClass, int senIndex)
{
	double returnValue;

	int i;

	int isContain;

	int noOfEqTuple;
	int noOfSatisfyEqTuple;
	equivalenceTuple **eqTupleList;
	equivalenceTuple *anEqTuple;

	noOfEqTuple = eqClass->noOfEqClass;
	eqTupleList = eqClass->eqClass;

	noOfSatisfyEqTuple = 0;

	returnValue = 0;
	for (i = 0; i < noOfEqTuple; i++)
	{
		anEqTuple = eqTupleList[i];

		isContain = ass_isContainSenIndex(anEqTuple, senIndex);
		if (isContain == 1)
		{
			returnValue += anEqTuple->noOfMember;
			noOfSatisfyEqTuple++;
		}
	}

	if (noOfSatisfyEqTuple == 0)
	{
		returnValue = 0;
	}
	else
	{
		returnValue = returnValue/((double) noOfSatisfyEqTuple);
	}

	return returnValue;
}

double ass_findAvgGlobalDistDiff_forOne(equivalenceTuple *aEqTuple, int senIndex, signatureGroupGroupGroup *allSig, int ***allSigMapping, int m)
{
	double returnValue;

	float triangle;
	float f;

	int i, j;

	int count;

	int noOfSize;
	signatureGroupGroup **sizeList;

	signatureGroup *aSigGroup;

	signatureGroupGroup *aSigGroupGroup;
	signatureGroup **groupList;
	int noOfGroup;

	//
	noOfSize = allSig->noOfSize;
	sizeList = allSig->sizeList;

	count = 0;

	returnValue = 0;
	for (i = 0; i < noOfSize; i++)
	{
		aSigGroupGroup = sizeList[i];

		noOfGroup = aSigGroupGroup->noOfGroup;
		groupList = aSigGroupGroup->groupList;

		for (j = 0; j < noOfGroup; j++)
		{
			aSigGroup = groupList[j];

			ass_calculateFDistDiffForOneAttrSetInEqTuple(aEqTuple, i, j, senIndex, aSigGroup, allSigMapping, &triangle, &f);

			returnValue += triangle;

			count++;
		}
	}

	returnValue = returnValue/((double)count);

	return returnValue;
}

double ass_findAvgGlobalDistDiff(equivalenceClass *eqClass, int senIndex, signatureGroupGroupGroup *allSig, int ***allSigMapping, int m)
{
	double returnValue;

	int noOfEqTuple;
	equivalenceTuple **eqTupleList;
	equivalenceTuple *aEqTuple;

	int i;

	noOfEqTuple = eqClass->noOfEqClass;
	eqTupleList = eqClass->eqClass;

	returnValue = 0;
	for (i = 0; i < noOfEqTuple; i++)
	{
		aEqTuple = eqTupleList[i];

		returnValue += ass_findAvgGlobalDistDiff_forOne(aEqTuple, senIndex, allSig, allSigMapping, m);
	}

	returnValue = returnValue/noOfEqTuple;

	return returnValue;
}


double ass_findAvgGlobalDistDiffSen(equivalenceClass *eqClass, int senIndex, signatureGroupGroupGroup *allSig, int ***allSigMapping, int m)
{
	double returnValue;

	int noOfEqTuple;
	equivalenceTuple **eqTupleList;
	equivalenceTuple *aEqTuple;

	int isContain;

	int noOfSatisfyEqTuple;

	int i;

	noOfEqTuple = eqClass->noOfEqClass;
	eqTupleList = eqClass->eqClass;

	noOfSatisfyEqTuple = 0;

	returnValue = 0;
	for (i = 0; i < noOfEqTuple; i++)
	{
		aEqTuple = eqTupleList[i];

		isContain = ass_isContainSenIndex(aEqTuple, senIndex);
		if (isContain == 1)
		{
			returnValue += ass_findAvgGlobalDistDiff_forOne(aEqTuple, senIndex, allSig, allSigMapping, m);
			noOfSatisfyEqTuple++;
		}
	}

	if (noOfSatisfyEqTuple == 0)
	{
		returnValue = 0;
	}
	else
	{
		returnValue = returnValue/((double) noOfSatisfyEqTuple);
	}

	return returnValue;
}


double ass_findAvgGlobalDistDiffSen_considerNewSen(equivalenceClass *eqClass, int senIndex, signatureGroupGroupGroup *allSig, int ***allSigMapping, int m, configure *config)
{
	double returnValue;

	int noOfEqTuple;
	equivalenceTuple **eqTupleList;
	equivalenceTuple *aEqTuple;

	int isContain;

	int noOfSatisfyEqTuple;

	int i;

	noOfEqTuple = eqClass->noOfEqClass;
	eqTupleList = eqClass->eqClass;

	noOfSatisfyEqTuple = 0;

	returnValue = 0;
	for (i = 0; i < noOfEqTuple; i++)
	{
		aEqTuple = eqTupleList[i];

		isContain = ass_isContainSenIndex_considerNewSen(aEqTuple, senIndex, config);
		if (isContain == 1)
		{
			returnValue += ass_findAvgGlobalDistDiff_forOne(aEqTuple, senIndex, allSig, allSigMapping, m);
			noOfSatisfyEqTuple++;
		}
	}

	if (noOfSatisfyEqTuple == 0)
	{
		returnValue = 0;
	}
	else
	{
		returnValue = returnValue/((double) noOfSatisfyEqTuple);
	}

	return returnValue;
}

void ass_main(configure *config, data *aData, generalizedAllTable *genAllTable)
{
	int ***allSigMapping;
	int senIndex;
	int noOfSenIndex;

	signatureGroupGroupGroup *allSig;

	int m;

	double globalDistDiff;
	double globalDistDiffSen;

	int violateEqID;
	int isViolateAll;

	int noOfSenValue;
	int noOfDim;
	int threshold;
	int noOfTuple;

	int isSuccess;


	FILE *debugFilePtr;
	char *debugFilename = "debug.txt";

	equivalenceClass *proEqClass;


	FILE *avgRelError_filePtr;
	

	allCondition *aAllCondition;
	int noOfQID;

	twoTable *finalTable;

	double avgRelError;

	double avgEqTupleSize, avgSenEqTupleSize;

	mechAllGroup *finalMechAllGroup;

	//int dummy;
	double infoLoss, distortion;
	int noOfSenChange;

	int noOfViolate;

	char *avgRelErrorFilename = "error.txt";
	char *timeFilename = "time.txt";

	int queryNoOfDim = config->queryNoOfDim;
	double s = config->s;
	int noOfCondition = config->noOfCondition;

#ifndef _WIN32
	FILE *timeFilePtr;

	float  userTime, sysTime;
	float userTime_dist, sysTime_dist;
	struct rusage myTime1, myTime2, myTime_distEnd;

	getrusage(RUSAGE_SELF,&myTime1);
#endif

	g_config = config;
	g_data = aData;

	m = config->d;
	noOfDim = config->noOfDim;
	noOfSenValue = aData->dimSize[noOfDim-1];
	noOfTuple = config->noOfTuple;

	noOfQID = noOfDim-1;

	threshold = calViolation_calculateSampleSize(config->error, config->sigma);
	//threshold = 1;
	printf(" Sample threshold = %d\n", threshold);
	

	// assume only one sensitive index
	noOfSenIndex = 2;



	// generate all signature
	printf("Generating signatures...\n");
	allSig = sig_findAllSignature(aData, threshold);

	sig_printStatistics(stdout, allSig);

	// find distribution
	printf("Finding the global distributions...\n");
	allSigMapping = sig_findGlobalDistributionFromOrgTable(config, allSig, aData, noOfSenIndex);

	// print distribution
	{
		FILE *sigFilePtr;
		char *sigFilename = "sig.txt";
		sigFilePtr = (FILE *) fopen(sigFilename, "w");
		if (sigFilePtr == NULL)
		{
			printf("File %s cannot be opened!\n", sigFilename);
			exit(0);
		}
		sig_printSignatureGroupGroupGroup(sigFilePtr, allSig, noOfSenIndex);
		fclose(sigFilePtr);

		exit(0);
	}

#ifndef _WIN32

	getrusage(RUSAGE_SELF, &myTime_distEnd);	

	userTime =
     		((float) (myTime_distEnd.ru_utime.tv_sec  - myTime1.ru_utime.tv_sec)) +
       		((float) (myTime_distEnd.ru_utime.tv_usec - myTime1.ru_utime.tv_usec)) * 1e-6;
	sysTime =
       		((float) (myTime_distEnd.ru_stime.tv_sec  - myTime1.ru_stime.tv_sec)) +
       		((float) (myTime_distEnd.ru_stime.tv_usec - myTime1.ru_stime.tv_usec)) * 1e-6;


	userTime_dist = userTime;
	sysTime_dist = sysTime;
#endif

	// find individual group
	printf("Forming individual groups...\n");
	proEqClass = ass_formAllIndividualEqTuple(aData);

	//
	printf("Creating statistics for eq. tuples...\n");
	createSenValueInEqClass(proEqClass, aData, noOfSenValue);
	eq_createSenIndexForAllEqTuple(proEqClass, config, noOfSenValue, noOfSenIndex);

	// for each sensitive index
	printf("Merging...\n");
	//for (i = 0; i < noOfSenIndex; i++)
	{
		senIndex = 1;

		noOfViolate = 0;

		// is some eq. tuple violate
		isViolateAll = ass_isViolateGlobalDistBoundConditionInAllEqTuple(proEqClass, senIndex, allSig, allSigMapping, m, &violateEqID, noOfTuple);
		while (isViolateAll == 1)
		{
			noOfViolate++;
			if (noOfViolate%50 == 0)
			{
				printf("  Processing %d-th violated eq. tuples...\n", noOfViolate);
			}
			//
			//isSuccess = ass_mergeEqTupleWithOthersForSatisfyPrivacyRequirement(violateEqID, proEqClass, senIndex, allSig, allSigMapping, m, noOfDim, genAllTable, noOfSenValue, noOfSenIndex, noOfTuple);
			isSuccess = ass_mergeEqTupleWithOthersForSatisfyPrivacyRequirement_additional(violateEqID, proEqClass, senIndex, allSig, allSigMapping, m, noOfDim, genAllTable, noOfSenValue, noOfSenIndex, noOfTuple);
			if (isSuccess == 0)
			{
				break;
			}

			isViolateAll = ass_isViolateGlobalDistBoundConditionInAllEqTuple(proEqClass, senIndex, allSig, allSigMapping, m, &violateEqID, noOfTuple);
		}
	}

	if (isSuccess == 0)
	{
		printf(" ********* Cannot find a grouping..\n");
	}

	//
	debugFilePtr = (FILE *) fopen(debugFilename, "w");
	if (debugFilePtr == NULL)
	{
		printf("File %s cannot be opened!\n", debugFilename);
		exit(0);
	}

	//
	//printf("Final equivalence class...\n");
	//printEquivalenceClass(debugFilePtr, proEqClass, noOfDim);

	// transform
	printf("Transform an equivalence class to mechGroup...\n");
	createSenValueInEqClass(proEqClass, aData, noOfSenValue);
	copySenValueToNewSenValueInEqClass(proEqClass, noOfSenValue);
	finalMechAllGroup = transformFromEqClassToMechAllGroup(proEqClass, noOfSenValue);
#ifndef _WIN32
	getrusage(RUSAGE_SELF, &myTime2);	
#endif
	//
	//allCondition *predicate_generateAllCondition(int noOfCondition, int queryNoOfDim, int noOfQID, double s, int *dimSize)
	printf("Generating predicates...\n");
	aAllCondition = predicate_generateAllCondition(noOfCondition, queryNoOfDim, noOfQID, s, aData->dimSize);

	// generate two tables
	printf("Generating two tables...\n");
	finalTable = generateTwoTableFromMechGroup(finalMechAllGroup, aData);

	// printing
	//printTwoTable(debugFilePtr, finalTable);

	// calculate the estimated count and actual count
	avgRelError = calculateAvgRelativeError(finalMechAllGroup, aAllCondition, aData, noOfQID);

	printf("Average relative error : %.2f\n", avgRelError);

	

	// calculate the coverage error (information loss and distortion)
	//void calculateQIDInfoLossAndSenDistortion(int noOfQID, int totalNoOfTuple, equivalenceClass *eqClass, generalizedAllTable *genAllTable, double *infoLoss, double *distortion)
	calculateQIDInfoLossAndSenDistortion(noOfQID, config->noOfTuple, proEqClass, genAllTable, &infoLoss, &distortion, &noOfSenChange);


	printf("Information Loss : %.2f\n", infoLoss);
	printf("Distortion       : %.2f\n", distortion);
	printf("No. of Sen Change: %d\n", noOfSenChange);


	avgRelError_filePtr = (FILE *) fopen(avgRelErrorFilename, "w");
	fprintf(avgRelError_filePtr, "%.6f\n", avgRelError);
	fprintf(avgRelError_filePtr, "%.6f\n", infoLoss);
	fprintf(avgRelError_filePtr, "%.6f\n", distortion);
	fprintf(avgRelError_filePtr, "%d\n", noOfSenChange);
	

	// assume one senIndex
	senIndex = 1;
	avgEqTupleSize = ass_findAvgEqTupleSize(proEqClass);
	avgSenEqTupleSize = ass_findAvgSenEqTupleSize(proEqClass, senIndex);
	fprintf(avgRelError_filePtr, "%.6f\n", avgEqTupleSize);
	fprintf(avgRelError_filePtr, "%.6f\n", avgSenEqTupleSize);

	globalDistDiff = ass_findAvgGlobalDistDiff(proEqClass, senIndex, allSig, allSigMapping, m);
	globalDistDiffSen = ass_findAvgGlobalDistDiffSen(proEqClass, senIndex, allSig, allSigMapping, m);
	fprintf(avgRelError_filePtr, "%.10f\n", globalDistDiff);
	fprintf(avgRelError_filePtr, "%.10f\n", globalDistDiffSen);

	//
	fprintf(avgRelError_filePtr, "\n");
	fprintf(avgRelError_filePtr, "%d\n", proEqClass->noOfEqClass);	// total. no. of eq tuples
	fprintf(avgRelError_filePtr, "%d\n", noOfViolate);			// total no. of eq tuples with sensitive values
	fprintf(avgRelError_filePtr, "0\n");				// total no. of eq. tuples with privacy breach

	printf("\n");
	printf("Avg. eq. tuple size = %.6f\n", avgEqTupleSize);
	printf("Avg. sen. eq. tuple size = %.6f\n", avgSenEqTupleSize);
	printf("Global dist. diff = %.10f\n", globalDistDiff);
	printf("Global dist. diff (sen) = %.10f\n", globalDistDiffSen);
	printf("Total no. of eq. classes = %d\n", proEqClass->noOfEqClass);
	printf("Total no. of eq. classes with sen. values = %d\n", noOfViolate);
	printf("Total no. of eq. classes with privacy breach = 0\n");
	printf("\n");


	fclose(avgRelError_filePtr);


#ifndef _WIN32
	userTime =
     		((float) (myTime2.ru_utime.tv_sec  - myTime1.ru_utime.tv_sec)) +
       		((float) (myTime2.ru_utime.tv_usec - myTime1.ru_utime.tv_usec)) * 1e-6;
	sysTime =
       		((float) (myTime2.ru_stime.tv_sec  - myTime1.ru_stime.tv_sec)) +
       		((float) (myTime2.ru_stime.tv_usec - myTime1.ru_stime.tv_usec)) * 1e-6;

	printf("User time : %f seconds\n",userTime);
	printf("System time : %f seconds\n\n",sysTime);

	fprintf(debugFilePtr, "Execution Time\n");
	fprintf(debugFilePtr, "--------------\n");
	fprintf(debugFilePtr, "User time : %f seconds\n",userTime);
    fprintf(debugFilePtr, "System time : %f seconds\n\n",sysTime);


	timeFilePtr = (FILE *) fopen(timeFilename, "w");
	fprintf(timeFilePtr, "%.6f\n",userTime);
    fprintf(timeFilePtr, "%.6f\n\n",sysTime);
	fprintf(timeFilePtr, "%.6f\n",userTime_dist);
    fprintf(timeFilePtr, "%.6f\n\n",sysTime_dist);
	fprintf(timeFilePtr, "%.6f\n",userTime-userTime_dist);
    fprintf(timeFilePtr, "%.6f\n\n",sysTime-sysTime_dist);
	fclose(timeFilePtr);
#endif

	fclose(debugFilePtr);

	// free
	// free allSigMapping
}

