
#ifndef EQ_H
#define EQ_H


#include "readData.h"
#include "generalization.h"
#include "fullSpace.h"
//#include "topDown.h"
//#include "signature.h"

struct equivalenceTuple_s
{
	dimension *content;
	int noOfMember;
	int noOfGenMember;
	int noOfUnGenMember;
	int noOfAvaMember;
	int noOfSen;
	int *member;
	int *isSen;

	dimension *senValue;
	dimension *newSenValue;

	int *noOfEachSen;
	int *noOfEachNewSen;
	int *noOfEachSenIndex;

	// newly added
	int id;

	int *signatureID;
};
typedef struct equivalenceTuple_s equivalenceTuple;

struct equivalenceClass_s
{
	int noOfEqClass;
	equivalenceTuple **eqClass;
};
typedef struct equivalenceClass_s equivalenceClass;

extern FILE *g_debugFilePtr;

extern configure *g_config;
extern data *g_data;

equivalenceClass *generateEquivalenceClass(data *dataSet, summary *synopsis, configure *conf);

equivalenceClass *new_equivalenceClass();
equivalenceTuple *equivalenceTuple_new();

int ceiling(double no);
void progressiveMain(data *dataSet, configure *conf, generalizedAllTable *genAllTable);
int isAlphaKAnonymity_EquivalenceClass(equivalenceTuple *eqOne, configure *conf);

extern double calculateDistortion(equivalenceClass *eqClass, configure *config, generalizedAllTable *genAllTable);
extern double calculateSenDistortion(equivalenceClass *eqClass, configure *config, generalizedAllTable *genAllTable);
extern int calculateGreatestPossRemovalCount(equivalenceClass *eqClass, configure *conf);
extern int chooseBestAttribute(equivalenceClass *eqClass, data *dataSet, configure *conf);
extern void generalizeEqClass(equivalenceClass *eqClass, int attrNo, configure *conf, generalizedAllTable *genAllTable);

float calculateDistanceBetweenEquivalenceTuple(configure *conf, generalizedAllTable *genAllTable, equivalenceTuple *eqTuple1, equivalenceTuple *eqTuple2);

void insertEqTupleIntoEqClass(equivalenceClass *eqClass, equivalenceTuple *eqTuple);
void removeEqTupleFromEqClass(equivalenceClass *eqClass, int index);

void mergeEqClass(equivalenceClass *eqClass, data *dataSet, configure *conf);

equivalenceTuple *mergeTwoEquivalenceTuple(int noOfDim, generalizedAllTable *genAllTable, equivalenceTuple *eqTuple1, equivalenceTuple *eqTuple2);

void hierachicalMain(data *dataSet, configure *conf, generalizedAllTable *genAllTable);
void randomMain(data *dataSet, configure *conf, generalizedAllTable *genAllTable);
void fullSpaceMain(data *dataSet, configure *conf, generalizedAllTable *genAllTable);
void topDownAlphaMain(data *dataSet, configure *conf, generalizedAllTable *genAllTable);
void mechMain(data *dataSet, configure *conf, generalizedAllTable *genAllTable);

equivalenceClass *cloneEquivalenceClass(equivalenceClass *eqClass, configure *conf);
equivalenceClass *cloneEquivalenceClassWithProjection(equivalenceClass *eqClass, int noOfDim, AttributeList *attrList, data *dataSet, configure *conf);
void freeEquivalenceClass(equivalenceClass *eqClass);

void printEquivalenceClass(FILE *filePtr, equivalenceClass *eqClass, int noOfDim);

int findMin(int no1, int no2);

equivalenceTuple *cloneEquivalenceTuple(equivalenceTuple *eqTuple, configure *conf);

extern double *checkAllGenDomainSet(AllGeneralizationDomainSet *allGenDomainSet, equivalenceClass *eqClass, configure *conf, data *dataSet, generalizedAllTable *genAllTable, int isCalculate);
double *checkAllGenDomainSet_TCloseness(double threshold, int *globalDist, AllGeneralizationDomainSet *allGenDomainSet, equivalenceClass *eqClass, configure *conf, data *dataSet, generalizedAllTable *genAllTable, int isCalculate);
extern equivalenceClass *chooseBestFinalEqClass(AllGeneralizationDomainSet *allGenDomainSet, double *dist, equivalenceClass *eqClass, data *dataSet, configure *conf, generalizedAllTable *genAllTable);

extern equivalenceClass *top_generalize2TopLevel(data *dataSet, generalizedAllTable *genAllTable, configure *conf);
//extern equivalenceTuple ****top_generateAllExpandValidEqTuple(equivalenceClass *eqClass, configure *conf, data *dataSet, generalizedAllTable *genAllTable, int ***noOfChild, int ****isSatisfy);

extern equivalenceTuple ****top_generateAllExpandValidEqTuple(equivalenceTuple ***parentEqTuple, int noOfParent, configure *conf, data *dataSet, generalizedAllTable *genAllTable, int ***noOfChild, int ****isSatisfy);
extern void top_processBestAttribute(equivalenceClass *finalEqClass, equivalenceClass *procEqClass, int noOfParentEqTuple, equivalenceTuple ***parentEqTuple, equivalenceTuple ****childEqTuple, int **noOfChild, int ***isSatisfy, configure *conf);

void freeEquivalenceTuple(equivalenceTuple *eqTuple);

equivalenceClass *topDownAlpha_KAlphaAnonymity(data *dataSet, configure *conf, generalizedAllTable *genAllTable, FILE *g_debugFilePtr);
equivalenceClass *fullSpace_KAlphaAnonymity(data *dataSet, configure *conf, generalizedAllTable *genAllTable, FILE *g_debugFilePtr);
equivalenceClass *fullSpace_TCloseness(data *dataSet, configure *conf, generalizedAllTable *genAllTable, FILE *g_debugFilePtr);

void eq_createSenIndexForAllEqTuple(equivalenceClass *eqClass, configure *config, int noOfSenValue, int noOfSenIndex);

void insertEqTupleIntoEqClassWithoutClone(equivalenceClass *eqClass, equivalenceTuple *eqTuple);
void freeEquivalenceTupleWithSenValueAndSenIndex(equivalenceTuple *eqTuple);
equivalenceTuple *mergeTwoEquivalenceTupleWithSenAndSenIndex(int noOfDim, generalizedAllTable *genAllTable, equivalenceTuple *eqTuple1, equivalenceTuple *eqTuple2, int noOfSenValue, int noOfSenIndex);
void eq_removeTwoEqTupleAndAddOneEqTupleToEqClass(equivalenceClass *eqClass, int removeEqID1, int removeEqID2, equivalenceTuple *eqTuple);
void eq_createSenIndexForAllEqTuple(equivalenceClass *eqClass, configure *config, int noOfSenValue, int noOfSenIndex);
void equivalenceClass_generateGeneralizedValue(equivalenceClass *eqClass, data *aData, generalizedAllTable *genAllTable, int noOfDim);

void equivalenceTuple_insertMember(equivalenceTuple *oneEqTuple, int id, data *aData, configure *config);

void equivalenceClass_mergeTwoEqTuple_withRemove_forNormal(equivalenceClass *eqClass, int index1, int index2);


extern int *tclose_getWholeDist(data *aData);
extern void tclose_singleExecution(data *aData, configure *config, generalizedAllTable *genAllTable);

#endif
