
#include "injector.h"

// assume only leaf values
int injector_tupleIsSatisfy(int noOfQID, dimension *QIDSet, int *dimNo, tuple *aTuple)
{
	int returnValue;

	int i;
	int aDimNo;

	int tupleValue;
	int QIDValue;

	returnValue = 1;
	for (i = 0; i < noOfQID; i++)
	{
		QIDValue = QIDSet[i].value;
		//QIDLevel = QIDSet[i].level;

		aDimNo = dimNo[i];

		tupleValue = aTuple->content[aDimNo].value;
		//tupleLevel = aTuple->content[aDimNo].level;

		if (tupleValue != QIDValue)
		{
			returnValue = 0;
			break;
		}
	}

	return returnValue;
}

mySet *injector_findIVSForTuple(assocRuleTable *aAssocRuleTable, tuple *aTuple)
{
	mySet *returnValue;

	int i;

	int noOfRule;
	assocRule **ruleList;

	assocRule *aRule;

	int isSatisfy;

	int noOfQID;
	dimension *QIDSet;
	int *dimNo;


	returnValue = mySet_new();

	noOfRule = aAssocRuleTable->noOfRule;
	ruleList = aAssocRuleTable->ruleList;

	for (i = 0; i < noOfRule; i++)
	{
		aRule = ruleList[i];

		noOfQID = aRule->noOfX;
		QIDSet = aRule->X;
		dimNo = aRule->dimNo;

		isSatisfy = injector_tupleIsSatisfy(noOfQID, QIDSet, dimNo, aTuple);

		if (isSatisfy == 1)
		{
			mySet_insertNonRepeat(returnValue, aRule->Y);
		}
	}

	return returnValue;
}

mySet **injector_findIVSForAllTuple(assocRuleTable *aAssocRuleTable, data *aData)
{
	mySet **returnValue;

	int noOfTuple;
	int i;

	tuple *tupleList;

	tuple *aTuple;

	noOfTuple = aData->noOfTuple;

	tupleList = aData->tupleList;

	returnValue = (mySet **) malloc(sizeof(mySet *)*noOfTuple);
	for (i = 0; i < noOfTuple; i++)
	{
		aTuple = &(tupleList[i]);
		returnValue[i] = injector_findIVSForTuple(aAssocRuleTable, aTuple);
	}

	return returnValue;
}

mySet *injector_findIVSForEqTuple(equivalenceTuple *aEqTuple, data *aData, mySet **IVSList)
{
	mySet *returnValue;

	int i;

	int noOfMember;
	int *member;

	int dataID;

	mySet *IVSForTuple;

	tuple *tupleList;

	tupleList = aData->tupleList;

	noOfMember = aEqTuple->noOfMember;
	member = aEqTuple->member;

	returnValue = mySet_new();
	for (i = 0; i < noOfMember; i++)
	{
		dataID = member[i];

		//aTuple = &(tupleList[dataID]);

		//IVSForTuple = injector_findIVSForTuple(aAssocRuleTable, aTuple);
		IVSForTuple = IVSList[dataID];

		mySet_unionNonRepeat(returnValue, IVSForTuple);

		//mySet_free(IVSForTuple);
	}

	return returnValue;
}

// assume senValue has content
mySet *injector_findSenSetForEqTuple(equivalenceTuple *eqTuple, data *aData)
{
	mySet *returnValue;

	dimension *senValueList;
	int aSenValue;


	int i;

	returnValue = mySet_new();

	// SEN[g]
	senValueList = eqTuple->senValue;
	for (i = 0; i < eqTuple->noOfMember; i++)
	{
		aSenValue = senValueList[i].value;

		mySet_insertNonRepeat(returnValue, aSenValue);
	}
	
	return returnValue;
}

// assume senValue has content
mySet *injector_findISForEqTuple(equivalenceTuple *eqTuple, data *aData, mySet **IVSList)
{
	mySet *returnValue;

	mySet *IVSSet;


	// SEN[g]
	returnValue = injector_findSenSetForEqTuple(eqTuple, aData);
	
	// IVS[g]
	IVSSet = injector_findIVSForEqTuple(eqTuple, aData, IVSList);

	mySet_unionNonRepeat(returnValue, IVSSet);

	mySet_free(IVSSet);

	return returnValue;
}

int injector_computeNIT_withOverlapBetweenSenSetAndIVS(specialN *aSpecialN, mySet *senSet)
{
	int returnValue;

	int i;

	int noOfIVS;
	int *count;
	mySet **IVS;

	mySet *oneIVS;

	int isOverlap;

	returnValue = 0;

	noOfIVS = aSpecialN->noOfIVS;
	count = aSpecialN->count;
	IVS = aSpecialN->IVS;

	for (i = 0; i < noOfIVS; i++)
	{
		oneIVS = IVS[i];

		isOverlap = mySet_isOverlap(oneIVS, senSet);
		if (isOverlap == 1)
		{
			returnValue += count[i];
		}
	}

	return returnValue;
}

int injector_computeNIT_forEqTuple(equivalenceTuple *eqTuple, data *aData, mySet **IVSList, int noOfSenValue, int *noOfTupleWithSenValueList, specialN **aSpecialNList)
{
	int returnValue;

	mySet *IS;

	int i;
	int senValue;

	int isExist;
	specialN *aSpecialN;

	mySet *senSet;

	IS = injector_findISForEqTuple(eqTuple, aData, IVSList);

	returnValue = 0;
	for (i = 0; i < noOfSenValue; i++)
	{
		senValue = i;

		isExist = mySet_isExist(IS, senValue);

		if (isExist == 1)
		{
			returnValue += noOfTupleWithSenValueList[senValue];
		}
		else
		{
			aSpecialN = aSpecialNList[senValue];

			senSet = injector_findSenSetForEqTuple(eqTuple, aData);

			returnValue += injector_computeNIT_withOverlapBetweenSenSetAndIVS(aSpecialN, senSet);

			mySet_free(senSet);
		}
	}

	mySet_free(IS);


	return returnValue;
}

int injector_computeNIT_forMergeEqTuple(equivalenceTuple *eqTuple, int id, configure *conf, data *aData, mySet **IVSList, int noOfSenValue, int *noOfTupleWithSenValueList, specialN **aSpecialNList)
{
	int returnValue;

	equivalenceTuple *newEqTuple;

	newEqTuple = cloneEquivalenceTuple(eqTuple, conf);
	equivalenceTuple_insertMember(newEqTuple, id, aData, conf);
	returnValue = injector_computeNIT_forEqTuple(newEqTuple, aData, IVSList, noOfSenValue, noOfTupleWithSenValueList, aSpecialNList);

	freeEquivalenceTuple(newEqTuple);
	free(newEqTuple);

	return returnValue;
}

void injector_incrementDataStructure(int senValue, int tupleID, mySet **IVSList, int noOfSenValue, int *noOfTupleWithSenValueList, specialN **aSpecialNList)
{
	mySet *oneIVSList;
	specialN *oneSpecialNList;
	int noOfIVS;

	int i;

	mySet **IVS;
	int *count;

	int isEqual;

	noOfTupleWithSenValueList[senValue]++;

	oneIVSList = IVSList[tupleID];

	oneSpecialNList = aSpecialNList[senValue];

	noOfIVS = oneSpecialNList->noOfIVS;
	IVS = oneSpecialNList->IVS;
	count = oneSpecialNList->count;
	for (i = 0; i < noOfIVS; i++)
	{
		isEqual = mySet_isEqual(IVS[i], oneIVSList);

		if (isEqual == 1)
		{
			count[i]++;
			break;
		}
	}
}

void injector_decrementDataStructure(int senValue, int tupleID, mySet **IVSList, int noOfSenValue, int *noOfTupleWithSenValueList, specialN **aSpecialNList)
{
	mySet *oneIVSList;
	specialN *oneSpecialNList;
	int noOfIVS;

	int i;

	mySet **IVS;
	int *count;

	int isEqual;

	noOfTupleWithSenValueList[senValue]--;

	oneIVSList = IVSList[tupleID];

	oneSpecialNList = aSpecialNList[senValue];

	noOfIVS = oneSpecialNList->noOfIVS;
	IVS = oneSpecialNList->IVS;
	count = oneSpecialNList->count;
	for (i = 0; i < noOfIVS; i++)
	{
		isEqual = mySet_isEqual(IVS[i], oneIVSList);

		if (isEqual == 1)
		{
			count[i]--;
			break;
		}
	}
}

int *injector_computeAllNIT(equivalenceClass *eqClassList, data *aData, mySet **IVSList, int noOfSenValue, int *noOfTupleWithSenValueList, specialN **aSpecialNList)
{
	int *returnValue;

	int i;

	int noOfEqClass;
	equivalenceTuple **eqClass;
	equivalenceTuple *oneEqTuple;

	noOfEqClass = eqClassList->noOfEqClass;
	eqClass = eqClassList->eqClass;

	returnValue = (int *) malloc(sizeof(int)*noOfEqClass);
	memset(returnValue, 0, sizeof(int)*noOfEqClass);
	for (i = 0; i < noOfEqClass; i++)
	{
		oneEqTuple = eqClass[i];
		returnValue[i] = injector_computeNIT_forEqTuple(oneEqTuple, aData, IVSList, noOfSenValue, noOfTupleWithSenValueList, aSpecialNList);
	}

	return returnValue;
}


int injector_findEqTupleIDWithMaxNIT(equivalenceClass *eqClassList, data *aData, mySet **IVSList, int noOfSenValue, int *noOfTupleWithSenValueList, specialN **aSpecialNList)
{
	int returnValue;

	int *allNIT;
	int i;
	int max;

	int noOfEqClass;
	equivalenceTuple **eqClass;

	allNIT = injector_computeAllNIT(eqClassList, aData, IVSList, noOfSenValue, noOfTupleWithSenValueList, aSpecialNList);

	noOfEqClass = eqClassList->noOfEqClass;
	eqClass = eqClassList->eqClass;

	max = -1;
	returnValue = -1;
	for (i = 0; i < noOfEqClass; i++)
	{
		if (allNIT[i] > max)
		{
			max = allNIT[i];
			returnValue = i;
		}
	}

	free(allNIT);

	return returnValue;
}

int injector_isTupleCompatibleWithTuple(configure *config, int tupleID1, int tupleID2, int noOfDim, data *aData, mySet **IVSList)
{
	int returnValue;

	mySet *IVS1;
	mySet *IVS2;

	tuple *tuple1;
	tuple *tuple2;

	int senValue1, senValue2;

	int isExist;

	IVS1 = IVSList[tupleID1];
	IVS2 = IVSList[tupleID2];

	tuple1 = &(aData->tupleList[tupleID1]);
	tuple2 = &(aData->tupleList[tupleID2]);

	senValue1 = tuple1->content[noOfDim-1].value;
	senValue2 = tuple2->content[noOfDim-1].value;

	//if ((config->senStartNo <= senValue1) && (senValue1 <= config->senEndNo) && (config->senStartNo <= senValue2) && (senValue2 <= config->senEndNo) )
	{
		returnValue = 1;
		if (senValue1 == senValue2)
		{
			if ((config->senStartNo <= senValue1) && (senValue1 <= config->senEndNo) )
			{
				returnValue = 0;
			}
			else
			{
				returnValue = 1;
			}
		}
		else
		{
			isExist = mySet_isExist(IVS2, senValue1);
			if (isExist == 1)
			{
				returnValue = 0;
			}
			else
			{
				isExist = mySet_isExist(IVS1, senValue2);
				if (isExist == 1)
				{
					returnValue = 0;
				}
			}
		}
	}
	//else
	//{
	//}

	return returnValue;
}

int injector_isTupleCompatibleWithEqTuple(configure *config, int tupleID, equivalenceTuple *eqTuple, int noOfDim, data *aData, mySet **IVSList)
{
	int returnValue;

	int i;
	int noOfMember;
	int *member;

	int tupleID_eqTuple;

	int isCompatible_tupleTuple;

	noOfMember = eqTuple->noOfMember;
	member = eqTuple->member;

	returnValue = 1;
	for (i = 0; i < noOfMember; i++)
	{
		tupleID_eqTuple = member[i];

		isCompatible_tupleTuple = injector_isTupleCompatibleWithTuple(config, tupleID_eqTuple, tupleID, noOfDim, aData, IVSList);

		if (isCompatible_tupleTuple == 0)
		{
			returnValue = 0;
			break;
		}
	}

	return returnValue;
}

int injector_findIndexCompatibleWithEqTupleAndWithMinNIT(mySet *remainTupleSet, equivalenceTuple *eqTuple, data *aData, mySet **IVSList, configure *conf, int *noOfTupleWithSenValueList, specialN **aSpecialNList)
{
	int returnValue;

	int i;

	int noOfElt;
	int *elt;

	int tupleID;
	int noOfDim;

	int isCompatible;

	int min;
	int value;

	int noOfSenValue;

	noOfElt = remainTupleSet->noOfElt;
	elt = remainTupleSet->elt;

	noOfDim = aData->noOfDim;

	noOfSenValue = aData->dimSize[noOfDim-1];

	if (noOfElt > 0)
	{
		min = -1;
		returnValue = -1;
		for (i = 0; i < noOfElt; i++)
		{
			tupleID = elt[i];

			isCompatible = injector_isTupleCompatibleWithEqTuple(conf, tupleID, eqTuple, noOfDim, aData, IVSList);

			if (isCompatible == 1)
			{
				value = injector_computeNIT_forMergeEqTuple(eqTuple, tupleID, conf, aData, IVSList, noOfSenValue, noOfTupleWithSenValueList, aSpecialNList);

				if (min == -1)
				{
					min = value;
					returnValue = i;
				}
				else
				{
					if (value < min)
					{
						min = value;
						returnValue = i;
					}
				}
			}
		}
	}
	else
	{
		returnValue = -1;
	}

	return returnValue;

}

int injector_findNoOfTupleIncompatibleInEqTupleWithTuple(configure *config, int tupleID, equivalenceTuple *eqTuple, int noOfDim, data *aData, mySet **IVSList)
{
	int returnValue;

	int i;
	int noOfMember;
	int *member;

	int tupleID_eqTuple;

	int isCompatible_tupleTuple;

	noOfMember = eqTuple->noOfMember;
	member = eqTuple->member;

	returnValue = 0;
	for (i = 0; i < noOfMember; i++)
	{
		tupleID_eqTuple = member[i];

		isCompatible_tupleTuple = injector_isTupleCompatibleWithTuple(config, tupleID_eqTuple, tupleID, noOfDim, aData, IVSList);

		if (isCompatible_tupleTuple == 0)
		{
			returnValue++;
		}
	}

	return returnValue;
}

// return eqTuple index
int injector_findEqTupleIndexWithMinNoOfTupleIncompatibleWithTuple(configure *config, int tupleID, equivalenceClass *eqClass, int noOfDim, data *aData, mySet **IVSList)
{
	int returnValue;

	int noOfEqTuple;
	equivalenceTuple **eqTupleList;
	equivalenceTuple *oneEqTuple;

	int noOfTupleIncompatible;

	int i;
	int min;

	eqTupleList = eqClass->eqClass;
	noOfEqTuple = eqClass->noOfEqClass;


	returnValue = -1;
	min = -1;
	for (i = 0; i < noOfEqTuple; i++)
	{
		oneEqTuple = eqTupleList[i];
		noOfTupleIncompatible = injector_findNoOfTupleIncompatibleInEqTupleWithTuple(config, tupleID, oneEqTuple, noOfDim, aData, IVSList);

		if (min == -1)
		{
			min = noOfTupleIncompatible;
			returnValue = i;
		}
		else
		{
			if (noOfTupleIncompatible < min)
			{
				min = noOfTupleIncompatible;
				returnValue = i;
			}
		}
	}

	return returnValue;
}

//noOfDiffSenValue in eqTuple1-eqTuple2
int injector_noOfDiffSenValue(equivalenceTuple *eqTuple1, equivalenceTuple *eqTuple2, int noOfSenValue)
{
	int returnValue;

	int *temp1;
	int *temp2;
	int i;

	int *member1;
	int noOfMember1;
	dimension *senValue1;

	int *member2;
	int noOfMember2;
	dimension *senValue2;

	int aValue;

	returnValue = 0;

	temp1 = (int *) malloc(sizeof(int)*noOfSenValue);
	memset(temp1, 0, sizeof(int)*noOfSenValue);

	temp2 = (int *) malloc(sizeof(int)*noOfSenValue);
	memset(temp2, 0, sizeof(int)*noOfSenValue);
	
	
	noOfMember1 = eqTuple1->noOfMember;
	member1 = eqTuple1->member;
	senValue1 = eqTuple1->senValue;

	noOfMember2 = eqTuple2->noOfMember;
	member2 = eqTuple2->member;
	senValue2 = eqTuple2->senValue;

	for (i = 0; i < noOfMember1; i++)
	{
		aValue = senValue1[i].value;

		temp1[aValue]++;
	}

	for (i = 0; i < noOfMember2; i++)
	{
		aValue = senValue2[i].value;

		temp2[aValue]++;
	}

	returnValue = 0;
	for (i = 0; i < noOfSenValue; i++)
	{
		//if ((temp1[i] == 0) && (temp2[i] != 0))
		//{
		//	returnValue++;
		//}
		//else 
		if ((temp1[i] != 0) && (temp2[i] == 0))
		{
			returnValue++;
		}
	}


	free(temp1);
	free(temp2);

	return returnValue;
}

int injector_findEqTupleWithMaxSenValueExpand(equivalenceClass *eqClass, equivalenceTuple *eqTuple, int noOfSenValue)
{
	int returnValue;

	int i;

	int noOfEqTuple;
	equivalenceTuple **eqTupleList;

	equivalenceTuple *oneEqTuple;

	int max;
	int noOfDiffSenValue;

	noOfEqTuple = eqClass->noOfEqClass;
	eqTupleList = eqClass->eqClass;

	max = -1;
	returnValue = -1;
	for (i = 0; i < noOfEqTuple ;i++)
	{
		oneEqTuple = eqTupleList[i];

		noOfDiffSenValue = injector_noOfDiffSenValue(oneEqTuple, eqTuple, noOfSenValue);
		if (noOfDiffSenValue > max)
		{
			max = noOfDiffSenValue;
			returnValue = i;
		}
	}

	return returnValue;
}

void injector_specialN_build_incrementIVS(specialN *oneSenSpecialN, mySet *oneIVSList)
{

	int *count;
	int noOfIVS;
	mySet **IVS;

	int i;

	int isFound;
	int isEqual;

	noOfIVS = oneSenSpecialN->noOfIVS;
	count = oneSenSpecialN->count;
	IVS = oneSenSpecialN->IVS;


	isFound = 0;
	for (i = 0; i < noOfIVS; i++)
	{
		isEqual = mySet_isEqual(IVS[i], oneIVSList);

		if (isEqual == 1)
		{
			// increment the count
			count[i]++;

			isFound = 1;
			break;
		}
	}

	if (isFound == 0)
	{
		// create a new entry
		if (noOfIVS == 0)
		{
			count = (int *) malloc(sizeof(int));
			IVS = (mySet **) malloc(sizeof(mySet *));
		}
		else
		{
			count = (int *) realloc(count, sizeof(int)*(noOfIVS+1));
			IVS = (mySet **) realloc(IVS, sizeof(int)*(noOfIVS+1));
		}
		count[noOfIVS] = 1;
		IVS[noOfIVS] = mySet_clone(oneIVSList);

		noOfIVS++;

		oneSenSpecialN->noOfIVS = noOfIVS;
		oneSenSpecialN->count = count;
		oneSenSpecialN->IVS = IVS;
	}
}

void injector_initializeDataStructure(int **noOfTupleWithSenValueList_output, specialN ***aSpecialNList_output, data *aData, int noOfSenValue, mySet **IVSList)
{
	int *noOfTupleWithSenValueList;
	specialN **aSpecialNList;
	specialN *oneSenSpecialN;

	int noOfTuple;
	tuple *tupleList;
	tuple *oneTuple;

	int aSenValue;

	int i;
	int noOfDim;

	mySet *oneIVSList;
	int tupleID;

	noOfTupleWithSenValueList = (int *) malloc(sizeof(int)*noOfSenValue);
	memset(noOfTupleWithSenValueList, 0, sizeof(int)*noOfSenValue);

	aSpecialNList = (specialN **) malloc(sizeof(specialN *)*noOfSenValue);
	memset(aSpecialNList, 0, sizeof(specialN *)*noOfSenValue);
	for (i = 0; i < noOfSenValue; i++)
	{
		aSpecialNList[i] = (specialN *) malloc(sizeof(specialN));
		memset(aSpecialNList[i], 0, sizeof(specialN));
		aSpecialNList[i]->s = i;
	}

	noOfTuple = aData->noOfTuple;
	tupleList = aData->tupleList;

	noOfDim = aData->noOfDim;

	for (i = 0; i < noOfTuple; i++)
	{
		oneTuple = &(tupleList[i]);

		tupleID = i;

		aSenValue = oneTuple->content[noOfDim-1].value;

		oneIVSList = IVSList[tupleID];

		oneSenSpecialN = aSpecialNList[aSenValue];
		injector_specialN_build_incrementIVS(oneSenSpecialN, oneIVSList);

		noOfTupleWithSenValueList[aSenValue]++;
	}

	(*noOfTupleWithSenValueList_output) = noOfTupleWithSenValueList;
	(*aSpecialNList_output) = aSpecialNList;
}

int injector_MBM_findUnmatch_qid(int indexQID, int *matchQID, int *matchSen, int *member, dimension *senID, int *isChooseQID, int *isChooseSen, int noOfMember, mySet **IVSList, int length)
{
	int returnValue;

	int j;

	int tupleID;
	int aSenValue;

	int isExist;
	int isFound;

	if (length > noOfMember - 1)
	{
		return 0;
	}

	if ((isChooseQID[indexQID] == 1) && (matchQID[indexQID] == -1))
	{
		tupleID = member[indexQID];

		length++;
		for (j = 0; j < noOfMember; j++)
		{
			aSenValue = senID[j].value;

			isExist = mySet_isExist(IVSList[tupleID], aSenValue);
			if (isExist == 0)
			{
				isFound = injector_MBM_findUnmatch_sen(indexQID, j, matchQID, matchSen, member, senID, isChooseQID, isChooseSen, noOfMember, IVSList, length);
				if (isFound == 1)
				{
					returnValue = 1;
					break;
				}
				else
				{
					returnValue = injector_MBM_findMatch_sen(j, matchQID, matchSen, member, senID, isChooseQID, isChooseSen, noOfMember, IVSList, length);
					if (returnValue == 1)
					{
						matchQID[indexQID] = j;
						matchSen[j] = indexQID;
						break;
					}
				}
			}
		}
	}

	return returnValue;
}

int injector_MBM_findUnmatch_sen(int indexQID, int indexSen, int *matchQID, int *matchSen, int *member, dimension *senID, int *isChooseQID, int *isChooseSen, int noOfMember, mySet **IVSList, int length)
{
	int returnValue;

	returnValue = 0;
	// isSatisfied for QID
	if ((isChooseSen[indexSen] == 1) && (matchSen[indexSen] == -1))
	{
		// isSatisfied for sen compatible
		returnValue = 1;

		matchSen[indexSen] = indexQID;
		matchQID[indexQID] = indexSen;
		
	}

	return returnValue;
}

int injector_MBM_findMatch_sen(int indexSen, int *matchQID, int *matchSen, int *member, dimension *senID, int *isChooseQID, int *isChooseSen, int noOfMember, mySet **IVSList, int length)
{
	int indexQID;

	int returnValue;

	int j;

	int tupleID;
	int aSenValue;

	int isExist, isFound;

	if (length > noOfMember-1)
	{
		return 0;
	}

	returnValue = 0;
	if ((isChooseSen[indexSen] == 1) && (matchSen[indexSen] != -1))
	{
		indexQID = matchSen[indexSen];

		tupleID = member[indexQID];

		length++;

		for (j = 0; j < noOfMember; j++)
		{
			aSenValue = senID[j].value;

			isExist = mySet_isExist(IVSList[tupleID], aSenValue);
			if (isExist == 0)
			{
				isFound = injector_MBM_findUnmatch_sen(indexQID, j, matchQID, matchSen, member, senID, isChooseQID, isChooseSen, noOfMember, IVSList, length);
				if (isFound == 1)
				{
					returnValue = 1;
					break;
				}
				else
				{
					returnValue = injector_MBM_findMatch_sen(j, matchQID, matchSen, member, senID, isChooseQID, isChooseSen, noOfMember, IVSList, length);
					if (returnValue == 1)
					{
						matchQID[indexQID] = j;
						matchSen[j] = indexQID;
						break;
					}
				}
			}
		}
	}

	return returnValue;
}

int injector_MBM_findUnmatch_fromAnyUnmatchNode(int *matchQID, int *matchSen, int *member, dimension *senID, int *isChooseQID, int *isChooseSen, int noOfMember, mySet **IVSList)
{
	int returnValue;

	int i;

	int isMatch;

	int length;

	// start from unmatched
	for (i = 0; i < noOfMember; i++)
	{
		length = 1;
		isMatch = injector_MBM_findUnmatch_qid(i, matchQID, matchSen, member, senID, isChooseQID, isChooseSen, noOfMember, IVSList, length);
		if (isMatch == 1)
		{
			returnValue = 1;

			
			break;
		}
	}

	return returnValue;
}

int injector_MBM_obtainGreatestNoOfMatch(int *member, dimension *senID, int *isChooseQID, int *isChooseSen, int noOfMember, mySet **IVSList)
{
	int returnValue;

	int *matchQID;
	int *matchSen;

	int i;

	int isExist;

	matchQID = (int *) malloc(sizeof(int)*noOfMember);
	matchSen = (int *) malloc(sizeof(int)*noOfMember);
	for (i = 0; i < noOfMember; i++)
	{
		matchQID[i] = -1;
		matchSen[i] = -1;
	}

	isExist = 1;
	while (isExist == 1)
	{
		isExist = injector_MBM_findUnmatch_fromAnyUnmatchNode(matchQID, matchSen, member, senID, isChooseQID, isChooseSen, noOfMember, IVSList);
	}

	// count
	returnValue = 0;
	for (i = 0; i < noOfMember; i++)
	{
		if (matchQID[i] != -1)
		{
			returnValue++;
		}
	}

	free(matchQID);
	free(matchSen);

	return returnValue;
}

int injector_checkWhetherEqTupleSatisfyMatchingLDiversity(equivalenceTuple *eqTuple, int l, mySet **IVSList)
{
	int returnValue;

	int noOfValid;
	int i, j;

	int noOfMember;

	int noOfMatch;

	int *isChooseQID;
	int *isChooseSen;

	int *member;
	dimension *senValue;

	noOfMember = eqTuple->noOfMember;

	member = eqTuple->member;
	senValue = eqTuple->senValue;

	isChooseQID = (int *) malloc(sizeof(int)*noOfMember);
	isChooseSen = (int *) malloc(sizeof(int)*noOfMember);
	for (i = 0; i < noOfMember; i++)
	{
		isChooseQID[i] = 1;
		isChooseSen[i] = 1;
	}

	returnValue = 1;
	for (i = 0; i < noOfMember; i++)
	{
		isChooseQID[i] = 0;
		noOfValid = 0;
		for (j = 0; j < noOfMember; j++)
		{
			isChooseSen[j] = 0;

			noOfMatch = injector_MBM_obtainGreatestNoOfMatch(member, senValue, isChooseQID, isChooseSen, noOfMember, IVSList);
			if (noOfMatch == noOfMember - 1)
			{
				noOfValid++;
			}

			isChooseSen[j] = 1;
		}

		if (noOfValid < l)
		{
			returnValue = 0;
			break;
		}
		isChooseQID[i] = 1;
	}

	free(isChooseQID);
	free(isChooseSen);

	return returnValue;
}

equivalenceClass *injector_transformDataToEqClass(mySet *aSet, data *aData, configure *config)
{
	equivalenceClass *returnValue;

	int noOfElt;
	int *elt; 

	int i;

	int id;

	equivalenceTuple *oneEqTuple;

	noOfElt = aSet->noOfElt;
	elt = aSet->elt;

	returnValue = new_equivalenceClass();

	returnValue->noOfEqClass = noOfElt;
	returnValue->eqClass = (equivalenceTuple **) malloc(sizeof(equivalenceTuple *)*noOfElt);

	for (i = 0; i < noOfElt; i++)
	{
		id = elt[i];

		oneEqTuple = equivalenceTuple_new();

		returnValue->eqClass[i] = oneEqTuple;

		equivalenceTuple_insertMember(oneEqTuple, id, aData, config);
	}

	return returnValue;
}

// NULL: unsuccessful
equivalenceClass *injector_bucketization(assocRuleTable *aAssocRuleTable, data *aData, int l, configure *config)
{
	equivalenceClass *returnValue;

	mySet **IVSList;

	int *noOfTupleWithSenValueList;

	specialN **aSpecialNList;

	int noOfSenValue;
	int noOfDim;

	mySet *remainSet;

	int eqTupleIndex;

	tuple *tupleList;

	equivalenceTuple *oneGroup;

	int i;

	mySet *set_postProcessing;

	int tupleIndex;
	int tupleID, tupleID2;
	int aSenValue, aSenValue2;
	int isRepeat;

	equivalenceClass *remainClass;

	equivalenceClass *finalEqClass;

	returnValue = NULL;

	set_postProcessing = mySet_new();

	finalEqClass = new_equivalenceClass();

	noOfDim = aData->noOfDim;
	noOfSenValue = aData->dimSize[noOfDim-1];

	// compute IVS list
	IVSList = injector_findIVSForAllTuple(aAssocRuleTable, aData);

	// initialize the data structure
	injector_initializeDataStructure(&noOfTupleWithSenValueList, &aSpecialNList, aData, noOfSenValue, IVSList);

	tupleList = aData->tupleList;

	// insert into the remainSet
	remainSet = mySet_new();
	remainSet->noOfElt = aData->noOfTuple;
	remainSet->elt = (int *) malloc(sizeof(int)*aData->noOfTuple);
	for (i = 0; i < aData->noOfTuple; i++)
	{
		remainSet->elt[i] = i;
	}

	// loop
	while (remainSet->noOfElt >= l)
	{

		// transform from remainData to eqClass
		remainClass = injector_transformDataToEqClass(remainSet, aData, config);

		// find the tuple ID with max. NIT
		eqTupleIndex = injector_findEqTupleIDWithMaxNIT(remainClass, aData, IVSList, noOfSenValue, noOfTupleWithSenValueList, aSpecialNList);

		// free remainClass
		freeEquivalenceClass(remainClass);
		free(remainClass);

		// form a new group
		oneGroup = equivalenceTuple_new();
		
		tupleID = remainSet->elt[eqTupleIndex];
		aSenValue = tupleList[tupleID].content[noOfDim-1].value;

		if (remainSet->noOfElt%100 == 0)
		{
			printf(" -------- Remaining elts: %d (tupldID = %d)\n", remainSet->noOfElt, tupleID);
		}

		if (remainSet->noOfElt == 206)
		{
			printf("");
		}

		equivalenceTuple_insertMember(oneGroup, tupleID, aData, config);
		mySet_removeIndex(remainSet, eqTupleIndex);

		// decrement
		injector_decrementDataStructure(aSenValue, tupleID, IVSList, noOfSenValue, noOfTupleWithSenValueList, aSpecialNList);

		// to form a group
		isRepeat = 0;
		while (oneGroup->noOfMember < l)
		{
			if ((remainSet->noOfElt == 71) && (oneGroup->noOfMember == 3))
			{
				printf("");
			}
			tupleIndex = injector_findIndexCompatibleWithEqTupleAndWithMinNIT(remainSet, oneGroup, aData, IVSList, config, noOfTupleWithSenValueList, aSpecialNList);

			if (tupleIndex == -1)
			{
				for (i = 0; i < oneGroup->noOfMember; i++)
				{
					tupleID2 = oneGroup->member[i];
					aSenValue2 = tupleList[tupleID2].content[noOfDim-1].value;
					
					injector_incrementDataStructure(aSenValue2, tupleID2, IVSList, noOfSenValue, noOfTupleWithSenValueList, aSpecialNList);

					if (tupleID2 != tupleID)
					{
						mySet_insert(remainSet, tupleID2);
					}
					else
					{
						mySet_insert(set_postProcessing, tupleID2);
					}
				}

				freeEquivalenceTuple(oneGroup);
				free(oneGroup);

				isRepeat = 1;
				break;
			}
			else
			{
				tupleID2 = remainSet->elt[tupleIndex];
				mySet_removeIndex(remainSet, tupleIndex);

				equivalenceTuple_insertMember(oneGroup, tupleID2, aData, config);

				aSenValue2 = tupleList[tupleID2].content[noOfDim-1].value;
				
				injector_decrementDataStructure(aSenValue2, tupleID2, IVSList, noOfSenValue, noOfTupleWithSenValueList, aSpecialNList);
			}
		}

		if (isRepeat == 1)
		{
			// do nothing
		}
		else
		{
			insertEqTupleIntoEqClass(finalEqClass, oneGroup);
		}
	}

	{
		FILE *filePtr;

		filePtr = (FILE *) fopen("injector.txt", "w");
		if (filePtr == NULL)
		{
			printf("Injector.txt cannot be opened!\n");
			exit(0);
		}
		fprintf(filePtr, "%d\n", set_postProcessing->noOfElt);
		fclose(filePtr);
	}
/*
	// Processing the remaining tuples in set_postProcessing
	printf(" Postprocessing (noOfElt:%d)....\n", set_postProcessing->noOfElt);
	for (i = 0; i < set_postProcessing->noOfElt; i++)
	{
		if (i%100 == 0)
		{
			printf(" -------- Post-processing elt no.: %d out of %d\n", i, set_postProcessing->noOfElt);
		}

		oneElt = set_postProcessing->elt[i];

		eqTupleIndex = injector_findEqTupleIndexWithMinNoOfTupleIncompatibleWithTuple(config, oneElt, finalEqClass, noOfDim, aData, IVSList);

		if (eqTupleIndex != -1)
		{
			oneGroup = finalEqClass->eqClass[eqTupleIndex];
			equivalenceTuple_insertMember(oneGroup, oneElt, aData, config);

			isSatisfy = injector_checkWhetherEqTupleSatisfyMatchingLDiversity(oneGroup, l, IVSList);
			while (isSatisfy == 0)
			{
				maxEqTupleIndex = injector_findEqTupleWithMaxSenValueExpand(finalEqClass, oneGroup, noOfSenValue);
				if (maxEqTupleIndex == eqTupleIndex)
				{
					//returnValue = -1;
					returnValue = NULL;
					break;
				}

				

				// merge two eqTuple (maxEqTupleIndex and eqTupleIndex)
				equivalenceClass_mergeTwoEqTuple_withRemove_forNormal(finalEqClass, eqTupleIndex, maxEqTupleIndex);

				if (maxEqTupleIndex < eqTupleIndex)
				{
					eqTupleIndex--;
				}
				//oneGroup = finalEqClass->eqClass[eqTupleIndex];
			}
		}
		else
		{
			// do nothing
		}
	}
*/
	mySet_free(set_postProcessing);
	for (i = 0; i < aData->noOfTuple; i++)
	{
		mySet_free(IVSList[i]);
	}
	free(IVSList);

	returnValue = finalEqClass;
	return returnValue;
}


assocRule *assocRule_new(int noOfX, int *dimNo, dimension *X, int Y)
{
	assocRule *returnValue;

	returnValue = (assocRule *) malloc(sizeof(assocRule));
	memset(returnValue, 0, sizeof(assocRule));

	returnValue->noOfX = noOfX;

	returnValue->dimNo = (int *) malloc(sizeof(int)*noOfX);
	returnValue->X = (dimension *) malloc(sizeof(dimension)*noOfX);
	returnValue->Y = Y;

	memcpy(returnValue->dimNo, dimNo, sizeof(int)*noOfX);
	memcpy(returnValue->X, X, sizeof(dimension)*noOfX);

	return returnValue;
}

assocRuleTable *assocRuleTable_new()
{
	assocRuleTable *returnValue;

	returnValue = (assocRuleTable *) malloc(sizeof(assocRuleTable));
	memset(returnValue, 0, sizeof(assocRuleTable));

	return returnValue;
}

void assocRuleTable_insert(assocRuleTable *aAssocRuleTable, assocRule *aRule)
{
	int noOfRule;
	assocRule **ruleList;

	noOfRule = aAssocRuleTable->noOfRule;
	ruleList = aAssocRuleTable->ruleList;

	if (noOfRule == 0)
	{
		ruleList = (assocRule **) malloc(sizeof(assocRule *));
	}
	else
	{
		ruleList = (assocRule **) realloc(ruleList, sizeof(assocRule *)*(noOfRule+1));
	}
	ruleList[noOfRule] = aRule;
	noOfRule++;

	aAssocRuleTable->noOfRule = noOfRule;
	aAssocRuleTable->ruleList = ruleList;

}

int injector_checkWhetherSatisfyNegAssocRule(signature *aSignature, int senIndex, int totalNoOfTuple, double *support_Y, double minConf)
{
	int returnValue;

	int oneSenCount;

	double one_support_Y;
	int noOfMember;

	noOfMember = aSignature->noOfMember;
	oneSenCount = aSignature->senCount[senIndex];
	one_support_Y = support_Y[senIndex];

	if ((noOfMember >= one_support_Y*totalNoOfTuple) && (oneSenCount <= (1-minConf)*noOfMember))
	{
		returnValue = 1;
	}
	else
	{
		returnValue = 0;
	}
	

	return returnValue;
}

void injector_findSenDistFromSignature(signature *aSignature, data *aData)
{
	int noOfTuple;
	tuple *tupleList;

	int noOfQID;
	dimension *QIDSet;
	int *dimNo;

	int noOfTotalDim;

	tuple *aTuple;

	int i;
	int isSatisfy;
	int aSenValue;

	noOfTotalDim = aData->noOfDim;

	noOfTuple = aData->noOfTuple;
	tupleList = aData->tupleList;

	noOfQID = aSignature->noOfAttr;
	QIDSet = aSignature->content;
	dimNo = aSignature->dimNo;
	for (i = 0; i < noOfTuple; i++)
	{
		aTuple = &(tupleList[i]);
		
		isSatisfy = injector_tupleIsSatisfy(noOfQID, QIDSet, dimNo, aTuple);
		if (isSatisfy == 1)
		{
			aSenValue = aTuple->content[noOfTotalDim-1].value;
			
			aSignature->senCount[aSenValue]++;
		}
	}
}

void injector_findSenDistFromSignatureGroup(signatureGroup *aSignatureGroup, data *aData)
{
	//void injector_findSenDistFromSignature(signature *aSignature, data *aData)
	int noOfSig;
	signature **sigList;

	signature *aSignature;

	int i;

	noOfSig = aSignatureGroup->noOfSig;
	sigList = aSignatureGroup->sigList;

	for (i = 0; i < noOfSig; i++)
	{
		aSignature = sigList[i];
		injector_findSenDistFromSignature(aSignature, aData);
	}
}

void injector_findSenDistFromSignatureGroupGroup(signatureGroupGroup *aSignatureGroupGroup, data *aData)
{
	int noOfGroup;
	signatureGroup **sigGroupList;
	signatureGroup *aSignatureGroup;

	int i;

	noOfGroup = aSignatureGroupGroup->noOfGroup;
	sigGroupList = aSignatureGroupGroup->groupList;

	for (i = 0; i < noOfGroup; i++)
	{
		aSignatureGroup = sigGroupList[i];
		injector_findSenDistFromSignatureGroup(aSignatureGroup, aData);
	}
}

// find min.
double injector_findMinSupport_forFreqItemset(double *support_Y, int noOfSenValue)
{
	int i;
	double returnValue;

	returnValue = support_Y[0];
	for (i = 1; i < noOfSenValue; i++)
	{
		if (returnValue > support_Y[i])
		{
			returnValue = support_Y[i];
		}
	}

	return returnValue;
}

double *injector_findSupportY(data *aData, int noOfSenValue, int noOfTuple, double minExp)
{
	double *returnValue;


	int i;

	tuple *tupleList;

	tuple *aTuple;

	int noOfDim;
	int aSenValue;

	noOfDim = aData->noOfDim;

	tupleList = aData->tupleList;

	returnValue = (double *) malloc(sizeof(double)*noOfSenValue);
	memset(returnValue, 0, sizeof(double)*noOfSenValue);

	for (i = 0; i < noOfTuple; i++)
	{
		aTuple = &(tupleList[i]);
		aSenValue = aTuple->content[noOfDim-1].value;

		returnValue[aSenValue]++;
	}

	for (i = 0; i < noOfSenValue; i++)
	{
		if (returnValue[i] == 0)
		{
			returnValue[i] = 1;
		}
		else
		{
			returnValue[i] = 1.0/noOfTuple*log(1 - minExp)/log(1 - returnValue[i]/((double) noOfTuple));
		}
	}

	return returnValue;
}

void injector_findNegAssocRule_fromSignature(assocRuleTable *aAssocRuleTable, int noOfSenValue, signature *aSignature, int totalNoOfTuple, double *support_Y, double minConf)
{
	int i;
	int senIndex;

	int isSatisfy;

	assocRule *aAssocRule;

	for (i = 0; i < noOfSenValue; i++)
	{
		senIndex = i;
		isSatisfy = injector_checkWhetherSatisfyNegAssocRule(aSignature, senIndex, totalNoOfTuple, support_Y, minConf);

		if (isSatisfy == 1)
		{
			// insert
			aAssocRule = assocRule_new(aSignature->noOfAttr, aSignature->dimNo, aSignature->content, senIndex);

			assocRuleTable_insert(aAssocRuleTable, aAssocRule);
		}
	}
}

void injector_findNegAssocRule_fromSignatureGroup(assocRuleTable *aAssocRuleTable, int noOfSenValue, signatureGroup *aSignatureGroup, int totalNoOfTuple, double *support_Y, double minConf)
{
	signature *aSignature;

	int noOfSig;
	signature **sigList;

	int i;

	noOfSig = aSignatureGroup->noOfSig;
	sigList = aSignatureGroup->sigList;

	for (i = 0; i < noOfSig; i++)
	{
		aSignature = sigList[i];
		injector_findNegAssocRule_fromSignature(aAssocRuleTable, noOfSenValue, aSignature, totalNoOfTuple, support_Y, minConf);
	}
}

void injector_findNegAssocRule_fromSignatureGroupGroup(assocRuleTable *aAssocRuleTable, int noOfSenValue, signatureGroupGroup *aSignatureGroupGroup, int totalNoOfTuple, double *support_Y, double minConf)
{
	signatureGroup *aSignatureGroup;

	int noOfGroup;
	signatureGroup **groupList;

	int i;

	noOfGroup = aSignatureGroupGroup->noOfGroup;
	groupList = aSignatureGroupGroup->groupList;

	for (i = 0; i < noOfGroup; i++)
	{
		aSignatureGroup = groupList[i];
		injector_findNegAssocRule_fromSignatureGroup(aAssocRuleTable, noOfSenValue, aSignatureGroup, totalNoOfTuple, support_Y, minConf);
	}
}

void injector_initializeSenCount_forSignature(signature *aSig, int noOfSenValue)
{
	aSig->senCount = (int *) malloc(sizeof(int)*noOfSenValue);
	memset(aSig->senCount, 0, sizeof(int)*noOfSenValue);
}

void injector_initializeSenCount_forSignatureGroup(signatureGroup *aSigGroup, int noOfSenValue)
{
	int noOfSig;
	signature **sigList;

	signature *aSig;

	int i;

	noOfSig = aSigGroup->noOfSig;
	sigList = aSigGroup->sigList;

	for (i = 0; i < noOfSig; i++)
	{
		aSig = sigList[i];

		injector_initializeSenCount_forSignature(aSig, noOfSenValue);
	}
}

void injector_initializeSenCount_forSignatureGroupGroup(signatureGroupGroup *aSigGroupGroup, int noOfSenValue)
{
	int noOfGroup;
	signatureGroup **groupList;

	signatureGroup *aSigGroup;

	int i;

	noOfGroup = aSigGroupGroup->noOfGroup;
	groupList = aSigGroupGroup->groupList;

	for (i = 0; i < noOfGroup; i++)
	{
		aSigGroup = groupList[i];

		injector_initializeSenCount_forSignatureGroup(aSigGroup, noOfSenValue);
	}
}

assocRuleTable *injector_findNegAssocRule(configure *config, data *aData)
{
	assocRuleTable *aAssocRuleTable;

	int noOfTuple;
	int noOfDim;
	int noOfSenValue;

	double *support_Y;
	double supportThreshold;

	int size;

	double minConf;

	signatureGroupGroup *newSigGroupGroup;

	signatureGroupGroup *aSigGroupGroup;

	

	minConf = config->minConf;

	noOfTuple = aData->noOfTuple;
	noOfDim = aData->noOfDim;
	noOfSenValue = aData->dimSize[noOfDim-1];

	support_Y = injector_findSupportY(aData, noOfSenValue, noOfTuple, config->minExp);
	supportThreshold = injector_findMinSupport_forFreqItemset(support_Y, noOfSenValue);

	printf(" Support Threshold = %.10f\n", supportThreshold);

	aSigGroupGroup = sig_createRealSignatureForMultipleAttr(aData, (int) supportThreshold);

	aAssocRuleTable = assocRuleTable_new();

	size = 1;
	while (aSigGroupGroup->noOfGroup > 0)
	{
		printf("---- Processing signature of size %d...\n", size);
	
		// do sth here...
		injector_initializeSenCount_forSignatureGroupGroup(aSigGroupGroup, noOfSenValue);
		injector_findSenDistFromSignatureGroupGroup(aSigGroupGroup, aData);
		injector_findNegAssocRule_fromSignatureGroupGroup(aAssocRuleTable, noOfSenValue, aSigGroupGroup, noOfTuple, support_Y, minConf);
		
		//
		newSigGroupGroup = sig_createSignatureGroupFromLowerLevel(aSigGroupGroup, (int) supportThreshold);

		sigGroupGroup_free(aSigGroupGroup);

		aSigGroupGroup = newSigGroupGroup;

		size++;

		// for efficiency purpose
		if (size >= 2)
		{
			break;
		}
	}

	sigGroupGroup_free(aSigGroupGroup);

	free(support_Y);

	return aAssocRuleTable;
}


void injector_assocRule_print(assocRule *aRule)
{
	int i;
	
	printf(" no. of X = %d\n", aRule->noOfX);
	printf(" X : ");
	for (i = 0; i < aRule->noOfX; i++)
	{
		printf("%d (dimNo:%d) ", aRule->X[i].value, aRule->dimNo[i]);
	}
	printf("\n");
	printf(" Y : %d\n", aRule->Y);
}

void injector_assocRuleTable_print(assocRuleTable *aAssocRuleTable)
{
	int noOfRule;
	assocRule **ruleList;

	assocRule *aRule;

	int i;

	noOfRule = aAssocRuleTable->noOfRule;
	ruleList = aAssocRuleTable->ruleList;

	for (i = 0; i < noOfRule; i++)
	{
		aRule = ruleList[i];

		printf("--------\n");
		printf("Rule %d\n", i);
		printf("--------\n");
		injector_assocRule_print(aRule);
	}
}

equivalenceClass *injector_main(data *aData, int l, configure *config)
{
	assocRuleTable *aAssocRuleTable;
	equivalenceClass *finalEqClass;

	printf("~~~~~~~ Mining negative association rules....\n");
	aAssocRuleTable = injector_findNegAssocRule(config, aData);

	printf("Priting assoc. Rule table (No. of Rules = %d)...\n", aAssocRuleTable->noOfRule);
	//injector_assocRuleTable_print(aAssocRuleTable);

	printf("~~~~~~~ Doing bucketization for injector...\n");
	finalEqClass = injector_bucketization(aAssocRuleTable, aData, l, config);

	return finalEqClass;
}


void injector_singleExecution(data *aData, int l, configure *config, generalizedAllTable *genAllTable)
{
	equivalenceClass *proEqClass;

	int noOfSenValue;
	int noOfDim;

	int m;

	int noOfViolate;

	tuple *tupleList;

	signatureGroupGroupGroup *allSig;

	int ***allSigMapping;

	FILE *avgRelError_filePtr;
	
	allCondition *aAllCondition;
	int noOfQID;

	twoTable *finalTable;

	double avgRelError;

	double avgEqTupleSize, avgSenEqTupleSize;

	mechAllGroup *finalMechAllGroup;

	FILE *debugFilePtr;
	char *debugFilename = "debug.txt";
	int senIndex;

	//int dummy;
	double infoLoss, distortion;
	int noOfSenChange;

	double globalDistDiff;
	double globalDistDiffSen;

	int noOfSenIndex;

	int threshold;

	int noOfEqTupleSenValue;

	char *avgRelErrorFilename = "error.txt";
	char *timeFilename = "time.txt";

	int queryNoOfDim = config->queryNoOfDim;
	double s = config->s;
	int noOfCondition = config->noOfCondition;

#ifndef _WIN32
	FILE *timeFilePtr;

	float  userTime, sysTime;
	float userTime_dist, sysTime_dist;
	struct rusage myTime1, myTime2, myTime_distEnd;

	getrusage(RUSAGE_SELF,&myTime1);
#endif

		noOfDim = aData->noOfDim;
	noOfSenValue = aData->dimSize[noOfDim-1];
	noOfQID = noOfDim-1;
	m = config->d;

	tupleList = aData->tupleList;


	// signature
	threshold = calViolation_calculateSampleSize(config->error, config->sigma);
	//threshold = 1;
	printf(" Sample threshold = %d\n", threshold);
	

	// assume only one sensitive index
	noOfSenIndex = 2;



	// generate all signature
	printf("Generating signatures...\n");
	allSig = sig_findAllSignature(aData, threshold);

	sig_printStatistics(stdout, allSig);

	// find distribution
	printf("Finding the global distributions...\n");
	allSigMapping = sig_findGlobalDistributionFromOrgTable(config, allSig, aData, noOfSenIndex);

#ifndef _WIN32

	getrusage(RUSAGE_SELF, &myTime_distEnd);	

	userTime =
     		((float) (myTime_distEnd.ru_utime.tv_sec  - myTime1.ru_utime.tv_sec)) +
       		((float) (myTime_distEnd.ru_utime.tv_usec - myTime1.ru_utime.tv_usec)) * 1e-6;
	sysTime =
       		((float) (myTime_distEnd.ru_stime.tv_sec  - myTime1.ru_stime.tv_sec)) +
       		((float) (myTime_distEnd.ru_stime.tv_usec - myTime1.ru_stime.tv_usec)) * 1e-6;


	userTime_dist = userTime;
	sysTime_dist = sysTime;
#endif

	proEqClass = injector_main(aData, l, config);

	//
	debugFilePtr = (FILE *) fopen(debugFilename, "w");
	if (debugFilePtr == NULL)
	{
		printf("File %s cannot be opened!\n", debugFilename);
		exit(0);
	}

	//
	//printf("Final equivalence class...\n");
	//printEquivalenceClass(debugFilePtr, proEqClass, noOfDim);

	//
	equivalenceClass_generateGeneralizedValue(proEqClass, aData, genAllTable, noOfDim);
	createSenValueInEqClass(proEqClass, aData, noOfSenValue);
	eq_createSenIndexForAllEqTuple(proEqClass, config, noOfSenValue, noOfSenIndex);


	// transform
	printf("Transform an equivalence class to mechGroup...\n");
	createSenValueInEqClass(proEqClass, aData, noOfSenValue);
	copySenValueToNewSenValueInEqClass(proEqClass, noOfSenValue);
	finalMechAllGroup = transformFromEqClassToMechAllGroup(proEqClass, noOfSenValue);
#ifndef _WIN32
	getrusage(RUSAGE_SELF, &myTime2);	
#endif
	//
	//allCondition *predicate_generateAllCondition(int noOfCondition, int queryNoOfDim, int noOfQID, double s, int *dimSize)
	printf("Generating predicates...\n");
	aAllCondition = predicate_generateAllCondition(noOfCondition, queryNoOfDim, noOfQID, s, aData->dimSize);

	// generate two tables
	printf("Generating two tables...\n");
	finalTable = generateTwoTableFromMechGroup(finalMechAllGroup, aData);

	// printing
	//printTwoTable(debugFilePtr, finalTable);

	// calculate the estimated count and actual count
	avgRelError = calculateAvgRelativeError(finalMechAllGroup, aAllCondition, aData, noOfQID);

	printf("Average relative error : %.2f\n", avgRelError);

	

	// calculate the coverage error (information loss and distortion)
	//void calculateQIDInfoLossAndSenDistortion(int noOfQID, int totalNoOfTuple, equivalenceClass *eqClass, generalizedAllTable *genAllTable, double *infoLoss, double *distortion)
	calculateQIDInfoLossAndSenDistortion(noOfQID, config->noOfTuple, proEqClass, genAllTable, &infoLoss, &distortion, &noOfSenChange);


	printf("Information Loss : %.2f\n", infoLoss);
	printf("Distortion       : %.2f\n", distortion);
	printf("No. of Sen Change: %d\n", noOfSenChange);


	avgRelError_filePtr = (FILE *) fopen(avgRelErrorFilename, "w");
	fprintf(avgRelError_filePtr, "%.6f\n", avgRelError);
	fprintf(avgRelError_filePtr, "%.6f\n", infoLoss);
	fprintf(avgRelError_filePtr, "%.6f\n", distortion);
	fprintf(avgRelError_filePtr, "%d\n", noOfSenChange);
	

	// assume one senIndex
	senIndex = 1;
	avgEqTupleSize = ass_findAvgEqTupleSize(proEqClass);
	avgSenEqTupleSize = ass_findAvgSenEqTupleSize(proEqClass, senIndex);
	fprintf(avgRelError_filePtr, "%.6f\n", avgEqTupleSize);
	fprintf(avgRelError_filePtr, "%.6f\n", avgSenEqTupleSize);

	globalDistDiff = ass_findAvgGlobalDistDiff(proEqClass, senIndex, allSig, allSigMapping, m);
	globalDistDiffSen = ass_findAvgGlobalDistDiffSen(proEqClass, senIndex, allSig, allSigMapping, m);
	fprintf(avgRelError_filePtr, "%.10f\n", globalDistDiff);
	fprintf(avgRelError_filePtr, "%.10f\n", globalDistDiffSen);

	//
	noOfEqTupleSenValue = anatomy_findNoOfEqTupleWithSenIndex(proEqClass, senIndex);
	noOfViolate = anatomy_findNoOfViolateEqTuple(proEqClass, senIndex, allSig, allSigMapping, m, aData->noOfTuple);
	fprintf(avgRelError_filePtr, "\n");
	fprintf(avgRelError_filePtr, "%d\n", proEqClass->noOfEqClass);	// total. no. of eq tuples
	fprintf(avgRelError_filePtr, "%d\n", noOfEqTupleSenValue);			// total no. of eq tuples with sensitive values
	fprintf(avgRelError_filePtr, "%d\n", noOfViolate);				// total no. of eq. tuples with privacy breach


	printf("\n");
	printf("Avg. eq. tuple size = %.6f\n", avgEqTupleSize);
	printf("Avg. sen. eq. tuple size = %.6f\n", avgSenEqTupleSize);
	printf("Global dist. diff = %.10f\n", globalDistDiff);
	printf("Global dist. diff (sen) = %.10f\n", globalDistDiffSen);
	printf("Total no. of eq. classes = %d\n", proEqClass->noOfEqClass);
	printf("Total no. of eq. classes with sen. values = %d\n", noOfEqTupleSenValue);
	printf("Total no. of eq. classes with privacy breach = %d\n", noOfViolate);
	printf("\n");


	fclose(avgRelError_filePtr);


#ifndef _WIN32
	userTime =
     		((float) (myTime2.ru_utime.tv_sec  - myTime1.ru_utime.tv_sec)) +
       		((float) (myTime2.ru_utime.tv_usec - myTime1.ru_utime.tv_usec)) * 1e-6;
	sysTime =
       		((float) (myTime2.ru_stime.tv_sec  - myTime1.ru_stime.tv_sec)) +
       		((float) (myTime2.ru_stime.tv_usec - myTime1.ru_stime.tv_usec)) * 1e-6;

	printf("User time : %f seconds\n",userTime);
	printf("System time : %f seconds\n\n",sysTime);

	fprintf(debugFilePtr, "Execution Time\n");
	fprintf(debugFilePtr, "--------------\n");
	fprintf(debugFilePtr, "User time : %f seconds\n",userTime);
    fprintf(debugFilePtr, "System time : %f seconds\n\n",sysTime);


	timeFilePtr = (FILE *) fopen(timeFilename, "w");
	fprintf(timeFilePtr, "%.6f\n",userTime);
    fprintf(timeFilePtr, "%.6f\n\n",sysTime);
	fprintf(timeFilePtr, "%.6f\n",userTime_dist);
    fprintf(timeFilePtr, "%.6f\n\n",sysTime_dist);
	fprintf(timeFilePtr, "%.6f\n",userTime-userTime_dist);
    fprintf(timeFilePtr, "%.6f\n\n",sysTime-sysTime_dist);
	fclose(timeFilePtr);
#endif

	fclose(debugFilePtr);
}
