
#include "mechDistortion.h"
#include "anatomy.h"
#include "ass.h"

dimensionSet *dimensionSet_new()
{
	dimensionSet *returnValue;

	returnValue = (dimensionSet *) malloc(sizeof(dimensionSet));
	memset(returnValue, 0, sizeof(dimensionSet));

	return returnValue;
}


void dimenionSet_free(dimensionSet *aDimensionSet)
{
	if (aDimensionSet->noOfElt > 0)
	{
		free(aDimensionSet->elt);
	}
	free(aDimensionSet);
}

void dimensionSet_insert(dimensionSet *aDimensionSet, dimension **aDimension)
{
	int noOfElt;

	noOfElt = aDimensionSet->noOfElt;
	if (noOfElt == 0)
	{
		aDimensionSet->elt = (dimension **) malloc(sizeof(dimension *));
	}
	else
	{
		aDimensionSet->elt = (dimension **) realloc(aDimensionSet->elt, sizeof(dimension *)*(noOfElt+1));
	}
	aDimensionSet->elt[noOfElt] = (*aDimension);
	aDimensionSet->noOfElt++;
}

//isContain = dimensionSet_isContain(sensitiveSet, &(senValueInEqTuple[i]));
int dimensionSet_isContain(dimensionSet *aDimensionSet, dimension *aDimension)
{
	int returnValue;
	int i;
	dimension *anotherDimension;

	returnValue = 0;

	for (i = 0; i < aDimensionSet->noOfElt; i++)
	{
		anotherDimension = aDimensionSet->elt[i];
		if ((anotherDimension->value == aDimension->value) && (anotherDimension->level == aDimension->level))
		{
			returnValue = 1;
			break;
		}
	}
	return returnValue;
}

dimensionSetSet *dimensionSetSet_new()
{
	dimensionSetSet *returnValue;

	returnValue = (dimensionSetSet *) malloc(sizeof(dimensionSetSet));
	memset(returnValue, 0, sizeof(dimensionSetSet));

	return returnValue;
}

void dimensionSetSet_insert(dimensionSetSet *aDimensionSetSet, dimensionSet **aDimensionSet)
{
	int noOfElt;

	noOfElt = aDimensionSetSet->noOfElt;
	if (noOfElt == 0)
	{
		aDimensionSetSet->elt = (dimensionSet **) malloc(sizeof(dimensionSet *));
	}
	else
	{
		aDimensionSetSet->elt = (dimensionSet **) realloc(aDimensionSetSet->elt, sizeof(dimensionSet *)*(noOfElt+1));
	}
	aDimensionSetSet->elt[noOfElt] = (*aDimensionSet);
	aDimensionSetSet->noOfElt++;
}

dimensionSetSet *readSensitiveSet(configure *conf)
{
	dimensionSetSet *returnValue;
	dimensionSet *aDimensionSet;
	dimension *aDimension;
	FILE *filePtr;

	int i, j;

	int noOfSet;
	int noOfElt;

	int aNo;

	filePtr = (FILE *) fopen(conf->sensitiveSetFilename, "r");
	if (filePtr == NULL)
	{
		printf("File %s cannot be opened!\n", conf->sensitiveSetFilename);
		exit(0);
	}

	returnValue = dimensionSetSet_new();

	fscanf(filePtr, "%d\n", &(noOfSet));
	fscanf(filePtr, "\n");

	for (i = 0; i < noOfSet; i++)
	{
		aDimensionSet = dimensionSet_new();

		fscanf(filePtr, "%d\n", &(noOfElt));
		for (j = 0; j < noOfElt; j++)
		{
			fscanf(filePtr, "%d\n", &aNo);
			aDimension = (dimension *) malloc(sizeof(dimension));
			aDimension->value = aNo;
			// assume ground value
			aDimension->level = 0;
			dimensionSet_insert(aDimensionSet, &aDimension);
		}
		fscanf(filePtr, "\n");

		dimensionSetSet_insert(returnValue, &aDimensionSet);
	}

	fclose(filePtr);

	return returnValue;
}

void createSenValueInEqTuple(equivalenceTuple *eqTuple, data *dataSet, int noOfValueInSenAttr)
{
	int noOfMember;
	int i;
	int dataID;
	dimension *senValue;
	tuple *tupleList;

	int senAttrNo;

	int groundValue;

	// assume the last dimension
	senAttrNo = dataSet->noOfDim-1;

	noOfMember = eqTuple->noOfMember;

	tupleList = dataSet->tupleList;

	eqTuple->senValue = (dimension *) malloc(sizeof(dimension)*noOfMember);
	eqTuple->noOfEachSen = (int *) malloc(sizeof(int)*noOfValueInSenAttr);
	memset(eqTuple->noOfEachSen, 0, sizeof(int)*noOfValueInSenAttr);

	for (i = 0; i < noOfMember; i++)
	{
		dataID = eqTuple->member[i];

		senValue = &(tupleList[dataID].content[senAttrNo]);

		memcpy(&(eqTuple->senValue[i]), senValue, sizeof(dimension));

		// assume ground value
		groundValue = senValue->value;
		eqTuple->noOfEachSen[groundValue]++;
	}

}

void createSenValueInEqClass(equivalenceClass *eqClass, data *dataSet, int noOfValueInSenAttr)
{
	int i;
	int noOfEqTuple;
	equivalenceTuple **eqList;
	equivalenceTuple *eqTuple;

	noOfEqTuple = eqClass->noOfEqClass;
	eqList = eqClass->eqClass;
	for (i = 0; i < noOfEqTuple; i++)
	{
		eqTuple = eqList[i];
		createSenValueInEqTuple(eqTuple, dataSet, noOfValueInSenAttr);
	}
}

void copySenValueToNewSenValueInEqTuple(equivalenceTuple *eqTuple, int noOfValueInSenAttr)
{
	int noOfMember;

	noOfMember = eqTuple->noOfMember;

	eqTuple->newSenValue = (dimension *) malloc(sizeof(dimension)*noOfMember);
	memcpy(eqTuple->newSenValue, eqTuple->senValue, sizeof(dimension)*noOfMember);

	eqTuple->noOfEachNewSen = (int *) malloc(sizeof(int)*noOfValueInSenAttr);
	memcpy(eqTuple->noOfEachNewSen, eqTuple->noOfEachSen, sizeof(int)*noOfValueInSenAttr);
}

void copySenValueToNewSenValueInEqClass(equivalenceClass *eqClass, int noOfValueInSenAttr)
{
	int noOfEqClass;
	int i;
	equivalenceTuple **eqList;

	noOfEqClass = eqClass->noOfEqClass;
	eqList = eqClass->eqClass;

	for (i = 0; i < noOfEqClass; i++)
	{
		copySenValueToNewSenValueInEqTuple(eqList[i], noOfValueInSenAttr);
	}
}

// return relative id in eqTuple
// -1: cannot find
int findTupleInEqTupleWithNewSenValue(equivalenceTuple *eqTuple, dimension *senValue)
{
	int returnValue;
	int noOfMember;
	int i;
	dimension *senValueInEqTuple;

	returnValue = -1;

	noOfMember = eqTuple->noOfMember;
	senValueInEqTuple = eqTuple->newSenValue;

	for (i = 0; i < noOfMember; i++)
	{
		// assume same level
		if (senValue->value == senValueInEqTuple[i].value)
		{
			returnValue = i;
			break;
		}
	}

	return returnValue;
}

// return relative id in eqTuple
int findTupleInEqTupleNotEqualNewSenValue(equivalenceTuple *eqTuple, dimensionSet *sensitiveSet)
{
	int returnValue;
	int noOfMember;
	int i;
	dimension *senValueInEqTuple;

	int isContain;

	returnValue = -1;

	noOfMember = eqTuple->noOfMember;
	senValueInEqTuple = eqTuple->newSenValue;

	for (i = 0; i < noOfMember; i++)
	{
		isContain = dimensionSet_isContain(sensitiveSet, &(senValueInEqTuple[i]));
		if (isContain == 0)
		{
			returnValue = i;
			break;
		}
	}

	return returnValue;
}

// change to value = -1 if nonSenStartNo > nonSenEndNo
void changeNewSenValueToNonSenInTupleInEqTuple(equivalenceTuple *eqTuple, int relativeID, int nonSenStartNo, int nonSenEndNo)
{
	dimension *senValueInEqTuple;
	int value, level;
	int *noOfEachSen;

	int orgValue;

	senValueInEqTuple = eqTuple->newSenValue;
	noOfEachSen = eqTuple->noOfEachNewSen;

	// value
	if (nonSenStartNo <= nonSenEndNo)
	{
		value = nonSenStartNo + rand()%(nonSenEndNo - nonSenStartNo + 1);

		orgValue = senValueInEqTuple[relativeID].value;
		noOfEachSen[orgValue]--;
		noOfEachSen[value]++;
	}
	else
	{
		value = -1;
	}
	
	// level 
	// assume the ground level
	level = 0;

	senValueInEqTuple[relativeID].value = value;
	senValueInEqTuple[relativeID].level = level;
}

void changeNewSenValueToSenInTupleInEqTuple(equivalenceTuple *eqTuple, int relativeID, dimension *senValue)
{
	dimension *senValueInEqTuple;
	int *noOfEachSen;

	int orgValue, value;
	
	senValueInEqTuple = eqTuple->newSenValue;
	noOfEachSen = eqTuple->noOfEachNewSen;

	orgValue = senValueInEqTuple[relativeID].value;
	value = senValue->value;

	memcpy(&(senValueInEqTuple[relativeID]), senValue, sizeof(dimension));

	noOfEachSen[orgValue]--;
	noOfEachSen[value]++;
}


void calculateQIDInfoLossAndSenDistortionInEqTuple(int noOfQID, equivalenceTuple *eqTuple, generalizedAllTable *genAllTable, double *infoLoss, double *distortion, int *noOfSenChange)
{
	int noOfMember;
	dimension *newSenValue;
	dimension *orgSenValue;

	dimension *aDim;
	dimension *dim1;
	dimension *dim2;

	int i;

	int coverage;
	int parentCoverage;

	generalizedDimTable *genDimTable;

	double weightA;

	weightA = 1.0/noOfQID;

	noOfMember = eqTuple->noOfMember;
	
	(*infoLoss) = 0;
	// QID
	for (i = 0; i < noOfQID; i++)
	{
		aDim = &(eqTuple->content[i]);
		genDimTable = &(genAllTable->genDimTable[i]);
		coverage = mech_findCoverage_dim(aDim, genDimTable);

		(*infoLoss) += weightA*(coverage-1.0)/(genDimTable->coverage-1.0);
	}

	(*infoLoss) *= noOfMember;

	// sen
	newSenValue = eqTuple->newSenValue;
	orgSenValue = eqTuple->senValue;

	genDimTable = &(genAllTable->genDimTable[noOfQID]);

	(*distortion) = 0;
	(*noOfSenChange) = 0;
	for (i = 0; i < noOfMember; i++)
	{
		dim1 = &(newSenValue[i]);
		dim2 = &(orgSenValue[i]);

		parentCoverage = mech_findParentCoverage_dim(genDimTable, dim1, dim2);
		(*distortion) += (parentCoverage-1.0)/(genDimTable->coverage-1.0);

		if (parentCoverage != 1)
		{
			(*noOfSenChange)++;
		}
	}

}



void calculateQIDInfoLossAndSenDistortion(int noOfQID, int totalNoOfTuple, equivalenceClass *eqClass, generalizedAllTable *genAllTable, double *infoLoss, double *distortion, int *noOfSenChange)
{
	int noOfEqTuple;
	equivalenceTuple **eqList;
	equivalenceTuple *aEqTuple;

	double oneInfoLoss, oneDistortion;
	int oneNoOfSenChange;

	int i;

	noOfEqTuple = eqClass->noOfEqClass;
	eqList = eqClass->eqClass;

	(*infoLoss) = 0;
	(*distortion) = 0;
	(*noOfSenChange) = 0;
	for (i = 0; i < noOfEqTuple; i++)
	{
		aEqTuple = eqList[i];

if (i == 4)
{
	printf("");
}

		calculateQIDInfoLossAndSenDistortionInEqTuple(noOfQID, aEqTuple, genAllTable, &oneInfoLoss, &oneDistortion, &oneNoOfSenChange);

		(*infoLoss) += oneInfoLoss;
		(*distortion) += oneDistortion;
		(*noOfSenChange) += oneNoOfSenChange;
	}

	(*infoLoss) /= totalNoOfTuple;
	(*distortion) /= totalNoOfTuple;

}


int findRelatedSenInEqTuple(equivalenceTuple *aEqTuple, dimensionSet *sensitiveSet)
{
	int aSenValue;
	int returnValue;

	int j;

	returnValue = 0;
	for (j = 0; j < sensitiveSet->noOfElt; j++)
	{
		// assume ground value
		aSenValue = sensitiveSet->elt[j]->value;
		returnValue += aEqTuple->noOfEachNewSen[aSenValue];
	}

	return returnValue;
}

// returnValue is NULL when noOfL = 0;
int *mech_findL(configure *conf, int *flag, int noOfEqTuple, double *propList, int *noOfL, int noOfViolate)
{
	int *returnValue;
	int noOfDistElt;
	int noOfRealDistElt;
	int count;

	int i;

	// alg 
	// 1: (d-1)*noOfViolate (satisfied)
	// 2: nonZero (satisfied)
	// 3: all (satisfied)
	if (conf->alg == 1)
	{

		noOfDistElt = (conf->d-1)*noOfViolate;

		if (noOfDistElt <= (noOfEqTuple-noOfViolate))
		{
			noOfRealDistElt = noOfDistElt;
		}
		else
		{
			noOfRealDistElt = noOfEqTuple-noOfViolate;
		}

		// because sorted
		count = 0;
		for (i = 0; i < noOfEqTuple; i++)
		{
			if (flag[i] != 1)
			{
				flag[i] = 2;
				count++;

				if (count == noOfRealDistElt)
				{
					break;
				}
			}
		}
	}
	else if (conf->alg == 2)
	{
		noOfRealDistElt = 0;
		for (i = 0; i < noOfEqTuple; i++)
		{
			if ((flag[i] != 1) && (propList[i] != 0.0))
			{
				flag[i] = 2;
				noOfRealDistElt++;
			}
		}
	}
	else if (conf->alg == 3)
	{
		noOfRealDistElt = 0;
		for (i = 0; i < noOfEqTuple; i++)
		{
			if (flag[i] != 1)
			{
				flag[i] = 2;
				noOfRealDistElt++;
			}
		}
	}
	else
	{
		noOfRealDistElt = 0;
	}
	(*noOfL) = noOfRealDistElt;

	if ((*noOfL) > 0)
	{
		returnValue = (int *) malloc(sizeof(int)*noOfRealDistElt);
		count = 0;
		for (i = 0; i < noOfEqTuple; i++)
		{
			if (flag[i] == 2)
			{
				if (count < noOfRealDistElt)
				{
					returnValue[count] = i;
					count++;
				}
			}
		}
	}
	else
	{
		returnValue = NULL;
	}

	return returnValue;
}


double *mech_findDistribution(int *L, int noOfL, double *propList, int noOfEqTuple)
{
	double *returnValue;

	int i;
	int eltNo;

	if (noOfL == 0)
	{
		returnValue = NULL;
	}
	else
	{
		returnValue = (double *) malloc(sizeof(double)*noOfL);

		for (i = 0; i < noOfL; i++)
		{
			eltNo = L[i];

			returnValue[i] = propList[eltNo];
		}
	}

	return returnValue;
}

// returnValue = NULL if noOfViolate = 0
int *mech_findV(configure *conf, int *flag, int noOfEqTuple, double *propList, int *noOfViolate, int *sortedIndex)
{
	int *returnValue;
	int count;
	int i;
	double threshold;

	int eltNo;


	threshold = 1.0/conf->d;
	(*noOfViolate) = 0;
	for (i = 0; i < noOfEqTuple; i++)
	{
		if (propList[i] > threshold)
		{
			flag[i] = 1;
			(*noOfViolate)++;
		}
	}

	if ((*noOfViolate) > 0)
	{
		returnValue = (int *) malloc(sizeof(int)*(*noOfViolate));
		count = 0;
		for (i = 0; i < noOfEqTuple; i++)
		{
			if (flag[i] == 1)
			{
				eltNo = sortedIndex[i];
				returnValue[count] = eltNo;
				count++;
			}
		}
	}
	else
	{
		returnValue = NULL;
	}

	return returnValue;
}

// return a list of d elements (eq id)
// return NULL when there is no equivalence class
void findDistributionAndAllEqClass(configure *conf, equivalenceClass *eqClass, dimensionSet *sensitiveSet, double **distribution, int *noOfDist, int **eqClassEltViolate, int *noOfEqClassViolate)
{
	double *propList;
	int *sortedIndex;

	equivalenceTuple *aEqTuple;
	equivalenceTuple **eqList;
	int totalRelatedSen;

	int noOfEqTuple;

	int i;

	int *flag;

	int noOfViolate;

	int *L;
	int noOfL;

	noOfEqTuple = eqClass->noOfEqClass;
	eqList = eqClass->eqClass;

	if (noOfEqTuple == 0)
	{
		// find V (i.e. a set of ECs which violates m-deassociation)
		(*eqClassEltViolate) = NULL;
		(*noOfEqClassViolate) = 0;
		
		// find distribution from L
		(*distribution) = NULL;
		(*noOfDist) = 0;
	}
	else
	{

		propList = (double *) malloc(sizeof(double)*noOfEqTuple);
		memset(propList, 0, sizeof(double)*noOfEqTuple);

		sortedIndex = (int *) malloc(sizeof(int)*noOfEqTuple);

		for (i = 0; i < noOfEqTuple; i++)
		{
			sortedIndex[i] = i;

			aEqTuple = eqList[i];
			totalRelatedSen = findRelatedSenInEqTuple(aEqTuple, sensitiveSet);

			propList[i] = totalRelatedSen/((double)aEqTuple->noOfMember);
		}

		// sort
		quicksort(sortedIndex, propList, 0, noOfEqTuple-1);

		flag = (int *) malloc(sizeof(int)*noOfEqTuple);
		// 1: violate
		// 0: normal
		// 2: choose for distribution

		// find V (i.e. a set of ECs which violates m-deassociation)
		(*eqClassEltViolate) = mech_findV(conf, flag, noOfEqTuple, propList, &noOfViolate, sortedIndex);
		(*noOfEqClassViolate) = noOfViolate;


		// find L for distribution
		L = mech_findL(conf, flag, noOfEqTuple, propList, &noOfL, noOfViolate);
		
		// find distribution from L
		(*distribution) = mech_findDistribution(L, noOfL, propList, noOfEqTuple);
		(*noOfDist) = noOfL;


		// for debug
		fprintf(g_debugFilePtr, "%d\n", noOfViolate);
		fprintf(g_debugFilePtr, "%d\n", noOfL);
		fprintf(g_debugFilePtr, "equivalenceClass\n");
		for (i = 0; i < noOfViolate; i++)
		{
			fprintf(g_debugFilePtr, "%.6f\n", propList[i]);
		}
		fprintf(g_debugFilePtr, "distribution\n");
		for (i = 0; i < noOfL; i++)
		{
			fprintf(g_debugFilePtr, "%.6f\n", (*distribution)[i]);
		}


		if (noOfL > 0)
		{
			free(L);
		}
		free(sortedIndex);
		free(propList);

		free(flag);
	}
}

// return a list of d elements (eq id)
// return NULL when there is no equivalence class
int *findEqTupleWithDGreatestProportionSenValue(equivalenceClass *eqClass, int d, dimensionSet *sensitiveSet, int *sizeOfResult)
{
	int *returnValue;
	int noOfEqTuple;
	equivalenceTuple **eqList;
	equivalenceTuple *aEqTuple;

	double *propList;
	int *sortedIndex;

	int i;

	int finalD;

	int totalRelatedSen;

	noOfEqTuple = eqClass->noOfEqClass;
	eqList = eqClass->eqClass;

	if (noOfEqTuple >= d)
	{
		finalD = d;
		returnValue = (int *) malloc(sizeof(int)*finalD);

		propList = (double *) malloc(sizeof(double)*noOfEqTuple);
		memset(propList, 0, sizeof(double)*noOfEqTuple);

		sortedIndex = (int *) malloc(sizeof(int)*noOfEqTuple);

		for (i = 0; i < noOfEqTuple; i++)
		{
			sortedIndex[i] = i;

			aEqTuple = eqList[i];
			totalRelatedSen = findRelatedSenInEqTuple(aEqTuple, sensitiveSet);
			propList[i] = totalRelatedSen/((double)aEqTuple->noOfMember);
		}

		// sort
		quicksort(sortedIndex, propList, 0, noOfEqTuple-1);


		// find d largest
		for (i = 0; i < finalD; i++)
		{
			returnValue[i] = sortedIndex[i];
		}

		free(propList);
		free(sortedIndex);
	}
	else
	{
		finalD = noOfEqTuple;
		if (finalD == 0)
		{
			returnValue = NULL;
		}
		else
		{
			
			returnValue = (int *) malloc(sizeof(int)*finalD);

			for (i = 0; i < finalD; i++)
			{
				returnValue[i] = i;
			}
		}
	}

	(*sizeOfResult) = finalD;

	return returnValue;
	
}

double calculateWeightedAverageProportionSenValue(equivalenceClass *eqClass, int *eltWithDGreatestProp, int noOfDElt, dimensionSet *sensitiveSet)
{

	double returnValue;
	int i;
	int eqClassID;

	equivalenceTuple *aEqTuple;
	equivalenceTuple **eqList;

	double *prop;
	int totalNoOfMember;

	int totalRelatedSen;

	double weight;

	eqList = eqClass->eqClass;

	prop = (double *) malloc(sizeof(double)*noOfDElt);

	totalNoOfMember = 0;
	for (i = 0; i < noOfDElt; i++)
	{
		eqClassID = eltWithDGreatestProp[i];

		aEqTuple = eqList[eqClassID];

		totalRelatedSen = findRelatedSenInEqTuple(aEqTuple, sensitiveSet);
		prop[i] = totalRelatedSen/((double)aEqTuple->noOfMember);

		//returnValue += prop[i]*aEqTuple->noOfMember;
		totalNoOfMember += aEqTuple->noOfMember;

	}

	returnValue = 0;
	for (i = 0; i < noOfDElt; i++)
	{
		eqClassID = eltWithDGreatestProp[i];

		aEqTuple = eqList[eqClassID];

		weight = aEqTuple->noOfMember/((double)totalNoOfMember);
		returnValue += weight*prop[i];
	}

	free(prop);

	return returnValue;
}


//senValue = findRandSenValueInSensitiveSet(sensitiveSet);
dimension *findRandSenValueInSensitiveSet(dimensionSet *sensitiveSet)
{
	dimension *returnValue;
	int index;

	returnValue = (dimension *) malloc(sizeof(dimension));

	index = rand()%sensitiveSet->noOfElt;
	memcpy(returnValue, sensitiveSet->elt[index], sizeof(dimension));

	return returnValue;
}

void performDistortionForDDeassociation(equivalenceClass *eqClass, dimensionSet *sensitiveSet, configure *conf, int nonSenStartNo, int nonSenEndNo)
{
	int noOfEqTuple;
	equivalenceTuple **eqList;
	equivalenceTuple *aEqTuple;

	int i;//, j;
	double prop;


	double threshold;

	int totalRelatedSen;

	int relativeID;

	//int aSenValue;
	dimension *senValue;

	// assume ground value
	//aSenValue = senValue->value;

	noOfEqTuple = eqClass->noOfEqClass;
	eqList = eqClass->eqClass;

	threshold = 1.0/conf->d;

	for (i = 0; i < noOfEqTuple; i++)
	{
		aEqTuple = eqList[i];

		totalRelatedSen = findRelatedSenInEqTuple(aEqTuple, sensitiveSet);
		prop = totalRelatedSen/((double) aEqTuple->noOfMember);

		while (prop > threshold)
		{
			senValue = findRandSenValueInSensitiveSet(sensitiveSet);

			relativeID = findTupleInEqTupleWithNewSenValue(aEqTuple, senValue);
			if (relativeID != -1)
			{
				changeNewSenValueToNonSenInTupleInEqTuple(aEqTuple, relativeID, nonSenStartNo, nonSenEndNo);

				totalRelatedSen = findRelatedSenInEqTuple(aEqTuple, sensitiveSet);
				prop = totalRelatedSen/((double) aEqTuple->noOfMember);
			}

			free(senValue);
		}
	}

}

double findRandProp(double *distribution, int noOfDist)
{
	int temp;

	temp = rand()%noOfDist;

	return distribution[temp];

}

void performDistortionForDNonDisclosure(equivalenceClass *eqClass, dimensionSet *sensitiveSet, configure *conf, int nonSenStartNo, int nonSenEndNo)
{
	int noOfEqClassViolate;
	int *eqClassEltViolate;

	int noOfDist;
	double *distribution;

	//int aSenValue;
	int totalRelatedSen;

	double randProp;

	int i;
	int eqTupleID;

	equivalenceTuple **eqList;
	equivalenceTuple *aEqTuple;

	double prop;

	int relativeID;

	dimension *senValue;

	// assume ground value
	//aSenValue = senValue->value;

	eqList = eqClass->eqClass;

	//eltWithDGreatestProp = findEqTupleWithDGreatestProportionSenValue(eqClass, conf->d, sensitiveSet, &noOfDElt);

	//avg = calculateWeightedAverageProportionSenValue(eqClass, eltWithDGreatestProp, noOfDElt, sensitiveSet);

	findDistributionAndAllEqClass(conf, eqClass, sensitiveSet, &distribution, &noOfDist, &eqClassEltViolate, &noOfEqClassViolate);
	


	for (i = 0; i < noOfEqClassViolate ; i++)
	{
		eqTupleID = eqClassEltViolate[i];

		aEqTuple = eqList[eqTupleID];

		totalRelatedSen = findRelatedSenInEqTuple(aEqTuple, sensitiveSet);
		prop = totalRelatedSen/((double) aEqTuple->noOfMember);

		if (noOfDist > 0)
		{
			randProp = findRandProp(distribution, noOfDist);
		}
		else
		{
			randProp = 1.0/conf->d;
		}

		if (prop < randProp)
		{
			while (prop < randProp)
			{
				relativeID = findTupleInEqTupleNotEqualNewSenValue(aEqTuple, sensitiveSet);

				senValue = findRandSenValueInSensitiveSet(sensitiveSet);
				changeNewSenValueToSenInTupleInEqTuple(aEqTuple, relativeID, senValue);

				totalRelatedSen = findRelatedSenInEqTuple(aEqTuple, sensitiveSet);
				prop = totalRelatedSen/((double) aEqTuple->noOfMember);

				free(senValue);
			}
		}
		if (prop > randProp)
		{
			while (prop > randProp)
			{
				senValue = findRandSenValueInSensitiveSet(sensitiveSet);
				relativeID = findTupleInEqTupleWithNewSenValue(aEqTuple, senValue);

				if (relativeID != -1)
				{
					changeNewSenValueToNonSenInTupleInEqTuple(aEqTuple, relativeID, nonSenStartNo, nonSenEndNo);

					totalRelatedSen = findRelatedSenInEqTuple(aEqTuple, sensitiveSet);
					prop = totalRelatedSen/((double) aEqTuple->noOfMember);
				}

				free(senValue);
			}
		}
	}

	if (noOfDist > 0)
	{
		//free(distribution);
	}
	if (noOfEqClassViolate > 0)
	{
		//free(eqClassEltViolate);
	}
}


void performDistortionForDNonDisclosureOLD(equivalenceClass *eqClass, dimensionSet *sensitiveSet, configure *conf, int nonSenStartNo, int nonSenEndNo)
{
	int noOfDElt;
	int *eltWithDGreatestProp;

	//int aSenValue;
	int totalRelatedSen;

	double avg;

	int i;
	int eqTupleID;

	equivalenceTuple **eqList;
	equivalenceTuple *aEqTuple;

	double prop;

	int relativeID;

	dimension *senValue;

	// assume ground value
	//aSenValue = senValue->value;

	eqList = eqClass->eqClass;

	eltWithDGreatestProp = findEqTupleWithDGreatestProportionSenValue(eqClass, conf->d, sensitiveSet, &noOfDElt);

	avg = calculateWeightedAverageProportionSenValue(eqClass, eltWithDGreatestProp, noOfDElt, sensitiveSet);

	for (i = 0; i < noOfDElt ; i++)
	{
		eqTupleID = eltWithDGreatestProp[i];

		aEqTuple = eqList[eqTupleID];

		totalRelatedSen = findRelatedSenInEqTuple(aEqTuple, sensitiveSet);
		prop = totalRelatedSen/((double) aEqTuple->noOfMember);

		if (prop < avg)
		{
			while (prop < avg)
			{
				relativeID = findTupleInEqTupleNotEqualNewSenValue(aEqTuple, sensitiveSet);

				senValue = findRandSenValueInSensitiveSet(sensitiveSet);
				changeNewSenValueToSenInTupleInEqTuple(aEqTuple, relativeID, senValue);

				totalRelatedSen = findRelatedSenInEqTuple(aEqTuple, sensitiveSet);
				prop = totalRelatedSen/((double) aEqTuple->noOfMember);

				free(senValue);
			}
		}
		if (prop > avg)
		{
			while (prop > avg)
			{
				senValue = findRandSenValueInSensitiveSet(sensitiveSet);
				relativeID = findTupleInEqTupleWithNewSenValue(aEqTuple, senValue);

				if (relativeID != -1)
				{
					changeNewSenValueToNonSenInTupleInEqTuple(aEqTuple, relativeID, nonSenStartNo, nonSenEndNo);

					totalRelatedSen = findRelatedSenInEqTuple(aEqTuple, sensitiveSet);
					prop = totalRelatedSen/((double) aEqTuple->noOfMember);
				}

				free(senValue);
			}
		}
	}
}


void calculateMaxEqClassAndMaxNoOfSenTuple(equivalenceClass *finalEqClass, int *maxEqClass, int *maxEqClass_noOfSen, int *maxNoOfSenTuple_eqClass, int *maxNoOfSenTuple, int *maxNo_eqClass, int *maxNo_noOfSenTuple)
{
	int i, j;

	int size;

	int noOfSen;

	int noOfEqTuple;
	equivalenceTuple **eqTupleList;
	equivalenceTuple *aEqTuple;

	double max_totalNoOfCombination;
	double totalNoOfCombination;

	noOfEqTuple = finalEqClass->noOfEqClass;
	eqTupleList = finalEqClass->eqClass;


	(*maxEqClass) = 0;
	(*maxEqClass_noOfSen) = 0;
	(*maxNoOfSenTuple_eqClass) = 0;
	(*maxNoOfSenTuple) = 0;
	(*maxNo_eqClass) = 0;
	(*maxNo_noOfSenTuple) = 0;

	max_totalNoOfCombination = 0;

	for (i = 0 ; i < noOfEqTuple; i++)
	{
		aEqTuple = eqTupleList[i];

		size = aEqTuple->noOfMember;

		noOfSen = 0;
		for (j = 0; j < size; j++)
		{
			if (aEqTuple->isSen[j] == 1)
			{
				noOfSen++;
			}
		}

		totalNoOfCombination = calculateCombination(size, noOfSen);

		if (size > (*maxEqClass))
		{
			(*maxEqClass) = size;
			(*maxEqClass_noOfSen) = noOfSen;
		}

		if (noOfSen > (*maxNoOfSenTuple))
		{
			(*maxNoOfSenTuple_eqClass) = size;
			(*maxNoOfSenTuple) = noOfSen;
		}

		if (totalNoOfCombination > max_totalNoOfCombination)
		{
			(*maxNo_eqClass) = size;
			(*maxNo_noOfSenTuple) = noOfSen;
			max_totalNoOfCombination = totalNoOfCombination;
		}
		
	}
}

equivalenceClass *mechDistortionMain(data *dataSet, configure *conf, generalizedAllTable *genAllTable)
{
	FILE *debugFilePtr;

	int isAlphaK;

	int kAnonymity;
	int noOfQID;

	summary *synopsis;
	equivalenceClass *proEqClass;
	equivalenceClass *finalEqClass;
	mechAllGroup *finalMechAllGroup;

	int noOfValueInSenAttr;

	int maxEqClass;
	int maxEqClass_noOfSen;
	int maxNoOfSenTuple_eqClass;
	int maxNoOfSenTuple;
	int maxNo_eqClass;
	int maxNo_noOfSenTuple;

	//int noOfGreatestPossRemovalCount;

	//int *listOfPotentialTrunk;
	//int noOfPotentialTrunk;

	//int prevNoOfFinalEqClass;

	//int noOfRemain;

	//int bestAttrNo;

	//int iterationNo;

	//double distortion;

	//double *dist;

	//Hash *hash;
	//DistStorage *distStorage;
	int noOfSlot;
	float delta;

	float maxValue;

	//int dataID;
	//int processEqID;
	//mechGroup *aMechGroup;
	//int isSen;

	double avgRelError;

	int i;
	//int noOfProEqClass;

	//int randIndex;

	//mechGroup *L;

	//int chosenDataID;
	//int chosenEqID;

	allCondition *aAllCondition;

	twoTable *finalTable;
	
	FILE *avgRelError_filePtr;

	//int iterationNo;

	int senStartNo, senEndNo;

	//dimension *senValue;
	int nonSenStartNo;
	int nonSenEndNo;

	//int dummy;
	double infoLoss, distortion;
	int noOfSenChange;

	int senIndex;
	double avgEqTupleSize;
	double avgSenEqTupleSize;
	double globalDistDiff;
	double globalDistDiffSen;

	int threshold;
	int noOfSenIndex;

	signatureGroupGroupGroup *allSig;

	int ***allSigMapping;

	int noOfEqTupleSenValue;
	int noOfViolate;

	int m;
	int noOfSenValue;

	dimensionSetSet *sensitiveSetSet;
	dimensionSet *sensitiveSet;

	char *avgRelErrorFilename = "error.txt";
	char *timeFilename = "time.txt";

	int queryNoOfDim = conf->queryNoOfDim;
	double s = conf->s;
	int noOfCondition = conf->noOfCondition;

#ifndef _WIN32
	FILE *timeFilePtr;
	float  userTime, sysTime;
	float userTime_topDown, sysTime_topDown;
	struct rusage myTime1, myTime2, myTime_topDownEnd;

	getrusage(RUSAGE_SELF,&myTime1);
#endif

	noOfQID = conf->noOfDim-conf->noOfSenAttr;
	kAnonymity = conf->K;

	noOfSenValue = dataSet->dimSize[conf->noOfDim-1];

	m = conf->d;


	if (conf->noOfTuple > 200)
	{
		noOfSlot = ((conf->noOfTuple*conf->noOfTuple)/200)-1;
	}
	else
	{
		noOfSlot = 5;
	}
	delta = 5.0f;

	maxValue = (float) (conf->noOfDim*2)*conf->noOfTuple/conf->K;

	debugFilePtr = (FILE *) fopen("debug.txt", "w");
	g_debugFilePtr = debugFilePtr;


	// signature
	threshold = calViolation_calculateSampleSize(conf->error, conf->sigma);
	//threshold = 1;
	printf(" Sample threshold = %d\n", threshold);
	

	// assume only one sensitive index
	noOfSenIndex = 2;



	// generate all signature
	printf("Generating signatures...\n");
	allSig = sig_findAllSignature(dataSet, threshold);

	sig_printStatistics(stdout, allSig);

	// find distribution
	printf("Finding the global distributions...\n");
	allSigMapping = sig_findGlobalDistributionFromOrgTable(conf, allSig, dataSet, noOfSenIndex);



	//dimensionSetSet *readSensitiveSet(configure *conf)
	sensitiveSetSet = readSensitiveSet(conf);

	printf("Random Algorithm....\n");

	// preprocessing
	//printf("Preprocessing....\n");
	//prepProcessDataSet(dataSet, conf->K, genAllTable);

	// create summary
	printf("Creating Summary....\n");
	synopsis = createSummary(dataSet);

	// printing
	//fprintf(debugFilePtr, "############## Summary ##################\n");
	//printf("############## Summary ##################\n");
	//printSummary(debugFilePtr, synopsis, conf->noOfDim);

	// check whether it is alphaK
	//isAlphaK = isAlphaKAnonymity(dataSet, synopsis, conf);
	//isAlphaK = 1;

	senStartNo = conf->senStartNo;
	senEndNo = conf->senEndNo;

	// generate equivalence class from the data
	printf("Generating eq. class from Data....\n");
	proEqClass = generateEquivalenceClass(dataSet, synopsis, conf); 
	noOfValueInSenAttr = dataSet->dimSize[noOfQID];
	createSenValueInEqClass(proEqClass, dataSet, noOfValueInSenAttr);
	isAlphaK = isMech(proEqClass, conf, noOfValueInSenAttr);
	isAlphaK = 1;

	// 
	if (isAlphaK == 0)
	{
		printf("This data set cannot satisfy the alpha K anonymity!\n");
		exit(0);
	}
	else
	{
		// continue

		printf("Running top down approach ....\n");
		//equivalenceClass *topDownAlpha_KAlphaAnonymity(data *dataSet, configure *conf, generalizedAllTable *genAllTable, FILE *g_debugFilePtr)
		finalEqClass = topDownAlpha_KAlphaAnonymity(dataSet, conf, genAllTable, debugFilePtr);

#ifndef _WIN32

	getrusage(RUSAGE_SELF, &myTime_topDownEnd);	

	userTime =
     		((float) (myTime_topDownEnd.ru_utime.tv_sec  - myTime1.ru_utime.tv_sec)) +
       		((float) (myTime_topDownEnd.ru_utime.tv_usec - myTime1.ru_utime.tv_usec)) * 1e-6;
	sysTime =
       		((float) (myTime_topDownEnd.ru_stime.tv_sec  - myTime1.ru_stime.tv_sec)) +
       		((float) (myTime_topDownEnd.ru_stime.tv_usec - myTime1.ru_stime.tv_usec)) * 1e-6;


	userTime_topDown = userTime;
	sysTime_topDown = sysTime;
#endif
		// createSenValue
		printf("Creating sen. value in eq. class....\n");
		createSenValueInEqClass(finalEqClass, dataSet, noOfValueInSenAttr);
		copySenValueToNewSenValueInEqClass(finalEqClass, noOfValueInSenAttr);

		// temp
		//conf->alg = 1;

// comment for comparison
		nonSenStartNo = 0;
		nonSenEndNo = conf->senStartNo-1;

		//senValue = (dimension *) malloc(sizeof(dimension));		
		//for (i = senStartNo; i <= senEndNo; i++)
		for (i = 0; i < sensitiveSetSet->noOfElt ; i++)
		{
			sensitiveSet = sensitiveSetSet->elt[i];

			// perform distortion steps of the sensitive attribute on each equivalence class in T*
			// such that each equivalence class in T* satisfies d-deassociation

			//printf("Performing distortion steps for d-deassociation (sen value set: %d)...\n", i);
			//senValue->value = i;
			// assume ground value
			//senValue->level = 0;
			// void performDistortionForDDeassociation(equivalenceClass *eqClass, dimension *senValue, configure *conf, int nonSenStartNo, int nonSenEndNo)
			//performDistortionForDDeassociation(finalEqClass, sensitiveSet, conf, nonSenStartNo, nonSenEndNo);

			// perform distortion steps of the sensitive attribute on equivalence classes with the
			// d greatest proportion of v_s in T* such that each equivalence class in T* satisfies
			// d-nondisclosure
			printf("Performing distortion steps for d-nondisclosure (sen value: %d)...\n", i);
			performDistortionForDNonDisclosure(finalEqClass, sensitiveSet, conf, nonSenStartNo, nonSenEndNo);
		}
		//free(senValue);
//
		printf("Transform an equivalence class to mechGroup...\n");
		finalMechAllGroup = transformFromEqClassToMechAllGroup(finalEqClass, noOfValueInSenAttr);
	}

#ifndef _WIN32
	getrusage(RUSAGE_SELF, &myTime2);	
#endif
	//
	//allCondition *predicate_generateAllCondition(int noOfCondition, int queryNoOfDim, int noOfQID, double s, int *dimSize)
	printf("Generating predicates...\n");
	aAllCondition = predicate_generateAllCondition(noOfCondition, queryNoOfDim, noOfQID, s, dataSet->dimSize);

	// generate two tables
	printf("Generating two tables...\n");
	finalTable = generateTwoTableFromMechGroup(finalMechAllGroup, dataSet);

	// printing
	//printTwoTable(debugFilePtr, finalTable);

	// calculate the estimated count and actual count
	avgRelError = calculateAvgRelativeError(finalMechAllGroup, aAllCondition, dataSet, noOfQID);

	printf("Average relative error : %.2f\n", avgRelError);

	

	// calculate the coverage error (information loss and distortion)
	//void calculateQIDInfoLossAndSenDistortion(int noOfQID, int totalNoOfTuple, equivalenceClass *eqClass, generalizedAllTable *genAllTable, double *infoLoss, double *distortion)
	calculateQIDInfoLossAndSenDistortion(noOfQID, conf->noOfTuple, finalEqClass, genAllTable, &infoLoss, &distortion, &noOfSenChange);


	// calculate max. eq. class size and max. # of sens. tuples
	calculateMaxEqClassAndMaxNoOfSenTuple(finalEqClass, &maxEqClass, &maxEqClass_noOfSen, &maxNoOfSenTuple_eqClass, &maxNoOfSenTuple, &maxNo_eqClass, &maxNo_noOfSenTuple);

	proEqClass = finalEqClass;


	printf("Information Loss : %.2f\n", infoLoss);
	printf("Distortion       : %.2f\n", distortion);
	printf("No. of Sen Change: %d\n", noOfSenChange);


	avgRelError_filePtr = (FILE *) fopen(avgRelErrorFilename, "w");
	fprintf(avgRelError_filePtr, "%.6f\n", avgRelError);
	fprintf(avgRelError_filePtr, "%.6f\n", infoLoss);
	fprintf(avgRelError_filePtr, "%.6f\n", distortion);
	fprintf(avgRelError_filePtr, "%d\n", noOfSenChange);
	

	//
	noOfSenIndex = 2;
	eq_createSenIndexForAllEqTuple(proEqClass, conf, noOfSenValue, noOfSenIndex);

	// assume one senIndex
	senIndex = 1;
	avgEqTupleSize = ass_findAvgEqTupleSize(proEqClass);
	avgSenEqTupleSize = ass_findAvgSenEqTupleSize(proEqClass, senIndex);
	fprintf(avgRelError_filePtr, "%.6f\n", avgEqTupleSize);
	fprintf(avgRelError_filePtr, "%.6f\n", avgSenEqTupleSize);

	globalDistDiff = ass_findAvgGlobalDistDiff(proEqClass, senIndex, allSig, allSigMapping, m);
	globalDistDiffSen = ass_findAvgGlobalDistDiffSen_considerNewSen(proEqClass, senIndex, allSig, allSigMapping, m, conf);
	fprintf(avgRelError_filePtr, "%.10f\n", globalDistDiff);
	fprintf(avgRelError_filePtr, "%.10f\n", globalDistDiffSen);

	//
	noOfEqTupleSenValue = anatomy_findNoOfEqTupleWithSenIndex_considerNewSen(proEqClass, senIndex, conf);
	noOfViolate = anatomy_findNoOfViolateEqTuple(proEqClass, senIndex, allSig, allSigMapping, m, dataSet->noOfTuple);
	fprintf(avgRelError_filePtr, "\n");
	fprintf(avgRelError_filePtr, "%d\n", proEqClass->noOfEqClass);	// total. no. of eq tuples
	fprintf(avgRelError_filePtr, "%d\n", noOfEqTupleSenValue);			// total no. of eq tuples with sensitive values
	fprintf(avgRelError_filePtr, "%d\n", noOfViolate);				// total no. of eq. tuples with privacy breach

	//** new
	fprintf(avgRelError_filePtr, "\n");
	fprintf(avgRelError_filePtr, "\n");
	fprintf(avgRelError_filePtr, "\n");
	fprintf(avgRelError_filePtr, "%d\n", maxEqClass);
	fprintf(avgRelError_filePtr, "%d\n", maxEqClass_noOfSen);
	fprintf(avgRelError_filePtr, "%d\n", maxNoOfSenTuple_eqClass);
	fprintf(avgRelError_filePtr, "%d\n", maxNoOfSenTuple);
	fprintf(avgRelError_filePtr, "%d\n", maxNo_eqClass);
	fprintf(avgRelError_filePtr, "%d\n", maxNo_noOfSenTuple);
	fprintf(avgRelError_filePtr, "%f\n", (double) calculateCombination(maxEqClass, maxEqClass_noOfSen));
	fprintf(avgRelError_filePtr, "%f\n", (double) calculateCombination(maxNoOfSenTuple_eqClass, maxNoOfSenTuple));
	fprintf(avgRelError_filePtr, "%f\n", (double) calculateCombination(maxNo_eqClass, maxNo_noOfSenTuple));

	printf("\n");
	printf("Avg. eq. tuple size = %.6f\n", avgEqTupleSize);
	printf("Avg. sen. eq. tuple size = %.6f\n", avgSenEqTupleSize);
	printf("Global dist. diff = %.10f\n", globalDistDiff);
	printf("Global dist. diff (sen) = %.10f\n", globalDistDiffSen);
	printf("Total no. of eq. classes = %d\n", proEqClass->noOfEqClass);
	printf("Total no. of eq. classes with sen. values = %d\n", noOfEqTupleSenValue);
	printf("Total no. of eq. classes with privacy breach = %d\n", noOfViolate);
	printf("\n");


	fclose(avgRelError_filePtr);



#ifndef _WIN32
	userTime =
     		((float) (myTime2.ru_utime.tv_sec  - myTime1.ru_utime.tv_sec)) +
       		((float) (myTime2.ru_utime.tv_usec - myTime1.ru_utime.tv_usec)) * 1e-6;
	sysTime =
       		((float) (myTime2.ru_stime.tv_sec  - myTime1.ru_stime.tv_sec)) +
       		((float) (myTime2.ru_stime.tv_usec - myTime1.ru_stime.tv_usec)) * 1e-6;

	printf("User time : %f seconds\n",userTime);
	printf("System time : %f seconds\n\n",sysTime);

	fprintf(debugFilePtr, "Execution Time\n");
	fprintf(debugFilePtr, "--------------\n");
	fprintf(debugFilePtr, "User time : %f seconds\n",userTime);
    fprintf(debugFilePtr, "System time : %f seconds\n\n",sysTime);


	timeFilePtr = (FILE *) fopen(timeFilename, "w");
	fprintf(timeFilePtr, "%.6f\n",userTime);
    fprintf(timeFilePtr, "%.6f\n\n",sysTime);
	fprintf(timeFilePtr, "%.6f\n",userTime_topDown);
    fprintf(timeFilePtr, "%.6f\n\n",sysTime_topDown);
	fprintf(timeFilePtr, "%.6f\n",userTime-userTime_topDown);
    fprintf(timeFilePtr, "%.6f\n\n",sysTime-sysTime_topDown);
	fclose(timeFilePtr);
#endif

	fclose(debugFilePtr);

	return finalEqClass;
}

