

#include "newton.h"

term ***newton_createMatrixExpression(term **termList, int noOfTerm)
{
	term ***returnValue;

	int i, j;

	int diffIndexNo;

	returnValue = (term ***) malloc(sizeof(term **)*noOfTerm);
	for (i = 0; i < noOfTerm; i++)
	{
		printf("  For the %d-th expression (of %d terms)...\n", i, noOfTerm);
		returnValue[i] = (term **) malloc(sizeof(term *)*noOfTerm);

		//if (i == 10)
		//{
		//	printf("");
		//}
		//term_print(termList[i]);
		//printf("\n");

		for (j = 0; j < noOfTerm; j++)
		{
			//printf("     w.r.t. x%d...\n", j);

//if ((i == 10) && (j == 0))
//{
//	printf("");
//}

			diffIndexNo = j;
			returnValue[i][j] = term_differentiate(termList[i], diffIndexNo);

//if ((i == 7) && (j == 7))
//{
//printf("--------------------\n");
//term_print(termList[i]);
//printf("--------------------\n");
//term_print(returnValue[i][j]);
//printf("\n");
//}
		}
	}

	return returnValue;
}

// one dimensional vector for matrix
MAT_TYPE *newton_createSubstituteMatrix(term ***matrixExpression, int noOfTerm, float *aVector)
{
	MAT_TYPE *returnValue;

	int i, j;
	int index;

	term *aTerm;

	returnValue = (MAT_TYPE *) malloc(sizeof(MAT_TYPE)*noOfTerm*noOfTerm);

	index = 0;
	for (i = 0; i < noOfTerm; i++)
	{
		for (j = 0; j < noOfTerm; j++)
		{
			aTerm = matrixExpression[i][j];
			returnValue[index] = (MAT_TYPE) term_calculateValue(aTerm, aVector);
			index++;
		}
	}

	return returnValue;
}

MAT_TYPE *newton_createSubstituteF(term **termList, int noOfTerm, float *aVector)
{
	MAT_TYPE *returnValue;

	int i;

	term *aTerm;

	returnValue = (MAT_TYPE *) malloc(sizeof(MAT_TYPE)*noOfTerm);

	for (i = 0; i < noOfTerm; i++)
	{
		aTerm = termList[i];
		returnValue[i] = (MAT_TYPE) term_calculateValue(aTerm, aVector);
	}

	return returnValue;
}

int newton_isContainValid(float *aVector, int noOfTerm)
{
	int returnValue;

	float posInf;
	float negInf;

	float aNo;

	int i;

	int isEqualNan;
	char aStr[20];

	posInf = (float) 1e40;
	negInf = (float) -1e40;

	returnValue = 1;
	for (i = 0; i < noOfTerm; i++)
	{
		aNo = aVector[i];

		
		sprintf(aStr, "%.6f", aNo);

		//printf("-------------%.6f %.6f (%s)\n", aNo, posInf, aStr);

		isEqualNan = strcmp(aStr, "nan");

		if ((aNo == posInf) || (aNo == negInf) || (isEqualNan == 0))
		{
			returnValue = 0;
			break;
		}
	}

	return returnValue;
}

int isContainNegative(float *aVector, int noOfTerm)
{
	int returnValue;

	int i;

	returnValue = 0;
	for (i = 0; i < noOfTerm; i++)
	{
		if (aVector[i] < 0)
		{
			returnValue = 1;
			break;
		}
	}

	return returnValue;
}


int newtow_isConvergence(float *aVector1, float *aVector2, int noOfTerm)
{
	int returnValue;

	int i;

	float error;

	float aValue;

	float ERROR_THRESHOLD = 0.0001f;

	int isContain1;
	int isContain2;

	

	error = 0;
	for (i = 0; i < noOfTerm; i++)
	{
		aValue = aVector1[i] - aVector2[i];
		error += aValue*aValue;
	}

	if (error < ERROR_THRESHOLD)
	{
		isContain1 = isContainNegative(aVector1, noOfTerm);
		isContain2 = isContainNegative(aVector2, noOfTerm);

		if ((isContain1 == 0) && (isContain2 == 0))
		{
			// converge
			returnValue = 1;
		}
		else
		{
			returnValue = 0;
		}
	}
	else
	{
		returnValue = 0;
	}

	return returnValue;
}

void vector_print(float *aVector, int noOfTerm)
{
	int i;

	printf("(");
	for (i = 0; i < noOfTerm; i++)
	{
		printf("%.4f", aVector[i]);
		if (i != noOfTerm-1)
		{
			printf(", ");
		}
	}
	printf(")");
}

float *newton_main(term **termList, int noOfTerm, float *initVector)
{
	float *returnValue;

	term ***matrixExpression;

	float *aVector;

	MAT_TYPE *pbJ;
	MAT_TYPE *pbJInv;

	MAT_TYPE *pbF;

	mR *mJ;
	mR *mJInv;

	mR *mF;
	mR *mV;
	mR *mTempMult;
	mR *mResult;

	int isInverseExist;

	int iterationNo;

	int isConverge;

	int i, j;

	int isContainValid;

	int ITERATION_BOUND = 100;

	returnValue = (float *) malloc(sizeof(float)*noOfTerm);
	
	printf("Creating the matrix expression...\n");
	matrixExpression = newton_createMatrixExpression(termList, noOfTerm);

	aVector = (float *) malloc(sizeof(float)*noOfTerm);
	memcpy(aVector, initVector, sizeof(float)*noOfTerm);

	mJ = (mR *) malloc(sizeof(mR));
	mJInv = (mR *) malloc(sizeof(mR));

	mF = (mR *) malloc(sizeof(mR));
	mV = (mR *) malloc(sizeof(mR));
	mTempMult = (mR *) malloc(sizeof(mR));
	mResult = (mR *) malloc(sizeof(mR));

	mJ->rows = noOfTerm;
	mJ->cols = noOfTerm;

	mJInv->rows = noOfTerm;
	mJInv->cols = noOfTerm;

	mF->rows = noOfTerm;
	mF->cols = 1;

	mV->rows = noOfTerm;
	mV->cols = 1;

	mTempMult->rows = noOfTerm;
	mTempMult->cols = 1;

	mResult->rows = noOfTerm;
	mResult->cols = 1;

	// malloc
	//mF->pb = (MAT_TYPE *) malloc(sizeof(MAT_TYPE)*mF->rows*mF->cols);
	//mV->pb = (MAT_TYPE *) malloc(sizeof(MAT_TYPE)*mV->rows*mV->cols);
	mV->pb = aVector;
	mTempMult->pb = (MAT_TYPE *) malloc(sizeof(MAT_TYPE)*mTempMult->rows*mTempMult->cols);
	mResult->pb = (MAT_TYPE *) malloc(sizeof(MAT_TYPE)*mResult->rows*mResult->cols);

	pbJInv = (MAT_TYPE *) malloc(sizeof(MAT_TYPE)*noOfTerm*noOfTerm);
	mJInv->pb = pbJInv;

	iterationNo = 1;
	isConverge = 0;
	while (isConverge == 0)
	{
		printf("Iteration No. %d....", iterationNo);

		if (iterationNo > ITERATION_BOUND)
		{
			return NULL;
			//break;
			// need to free something
		}

		// calculate the matrix
		//printf("  Calculating the matrix ...\n");
		pbJ = newton_createSubstituteMatrix(matrixExpression, noOfTerm, aVector);
		mJ->pb = pbJ;

		// calculate the inverse
		//printf("  Calculating inverse ...\n");
		isInverseExist = inverse_mR(mJ, mJInv);

		if (isInverseExist == 1)
		{

			// calculate the F
			//printf("  Calculating F...\n");
			pbF = newton_createSubstituteF(termList, noOfTerm, aVector);
			mF->pb = pbF;



			// calculate the multiplication
			//printf("  Calculating multiplication...\n");
			mul_mR(mJInv, mF, mTempMult);

			// calculate the subtraction
			//printf("  Calculating subtraction...\n");
			sub_mR(mV, mTempMult, mResult);

			free(pbF);

			isContainValid = newton_isContainValid(mResult->pb, noOfTerm);
			if (isContainValid == 1)
			{
				//
				isConverge = newtow_isConvergence(aVector, mResult->pb, noOfTerm);

				//
				memcpy(aVector, mResult->pb, sizeof(MAT_TYPE)*noOfTerm);
			}
			else
			{
				isConverge = 1;
			}

			printf("    ");
			vector_print(aVector, noOfTerm);
			printf("\n");

			iterationNo++;
		}
		else
		{
			isConverge = 1;
		}

	

		free(pbJ);

		//{
		//	int aNo;
		//	scanf("%d\n", &aNo);
		//}
	}

	memcpy(returnValue, aVector, sizeof(MAT_TYPE)*noOfTerm);

	free(mJ);
	free(mJInv);

	free(pbJInv);
	//free(mV->pb);
	free(mTempMult->pb);
	free(mResult->pb);

	free(mF);
	free(mV);
	free(mTempMult);
	free(mResult);
	for (i = 0; i < noOfTerm; i++)
	{
		for (j = 0; j < noOfTerm; j++)
		{
			term_free(matrixExpression[i][j]);
		}
		free(matrixExpression[i]);
	}
	free(matrixExpression);

	free(aVector);

	return returnValue;
}

// default: no. of terms = 4
term **test_newtonSetInitial()
{
	int noOfTerm = 4;

	term **termList;

	term *aTerm;

	term *aTerm1;
	term *aTerm2;
	term *aTerm3;

	literal *aLiteral11;
	literal *aLiteral12;

	literal *aLiteral3;

	term *aTerm21;
	term *aTerm22;

	term *aTerm221;
	term *aTerm222;

	int noOfTerm22;

	term **termList22;


	
	int noOfLiteral21;
	int noOfLiteral221, noOfLiteral222;

	literal **aLiteral21List;
	literal **aLiteral221List;
	literal **aLiteral222List;

	int noOfSmallTerm;

	int i, j;

	
	termList = (term **) malloc(sizeof(term *)*noOfTerm);
	for (i = 0; i < noOfTerm; i++)
	{
		aTerm = term_new();
		termList[i] = aTerm;
		
		switch (i)
		{
		case 0:
			noOfSmallTerm = 3;
			aTerm->noOfOperand = noOfSmallTerm;
			
			aTerm->operandList = (term **) malloc(sizeof(term *)*noOfSmallTerm);
			aTerm->operationList = (int *) malloc(sizeof(int)*noOfSmallTerm);
			
			for (j = 0; j < noOfSmallTerm; j++)
			{
				aTerm->operationList[j] = OP_MINUS;
			}
			
			// term 1
			aTerm1 = term_new();
			aLiteral11 = literal_createLiteralFromConstant(3);;
			aLiteral12 = literal_createLiteralFromVariable(0, 1);;
			
			aTerm1->noOfOperand = 2;
			aTerm1->operandList = (term **) malloc(sizeof(term *)*aTerm1->noOfOperand);
			aTerm1->operationList = (int *) malloc(sizeof(int)*aTerm1->noOfOperand);
			
			aTerm1->operandList[0] = term_createTermFromLiteral(aLiteral11);
			aTerm1->operandList[1] = term_createTermFromLiteral(aLiteral12);
			
			aTerm1->operationList[0] = OP_MULT;
			aTerm1->operationList[1] = OP_MULT;
			
			
			// term 2
			aTerm2 = term_new();
			
			aTerm2->noOfOperand = 2;
			aTerm2->operandList = (term **) malloc(sizeof(term *)*aTerm2->noOfOperand);
			aTerm2->operationList = (int *) malloc(sizeof(int)*aTerm2->noOfOperand);
			
			
			
			aTerm2->operationList[0] = OP_DIVIDE;
			aTerm2->operationList[1] = OP_DIVIDE;
			
			// term 21
			noOfLiteral21 = 2;
			aLiteral21List = (literal **) malloc(sizeof(literal *)*noOfLiteral21);
			aLiteral21List[0] = literal_createLiteralFromVariable(0, 1);
			aLiteral21List[1] = literal_createLiteralFromVariable(3, 1);
			
			
			aTerm21 = term_mergeLiteral(noOfLiteral21, aLiteral21List, OP_MULT);

			free(aLiteral21List);
			
			// term 22
			// term 221
			noOfLiteral221 = 2;
			aLiteral221List = (literal **) malloc(sizeof(literal *)*noOfLiteral221);
			aLiteral221List[0] = literal_createLiteralFromVariable(0, 1);
			aLiteral221List[1] = literal_createLiteralFromVariable(3, 1);
			
			
			aTerm221 = term_mergeLiteral(noOfLiteral221, aLiteral221List, OP_MULT);

			free(aLiteral221List);
			
			// term 222
			noOfLiteral222 = 2;
			aLiteral222List = (literal **) malloc(sizeof(literal *)*noOfLiteral222);
			aLiteral222List[0] = literal_createLiteralFromVariable(1, 1);
			aLiteral222List[1] = literal_createLiteralFromVariable(2, 1);
			
			
			aTerm222 = term_mergeLiteral(noOfLiteral222, aLiteral222List, OP_MULT);

			free(aLiteral222List);
			
			noOfTerm22 = 2;
			termList22 = (term **) malloc(sizeof(term *)*noOfTerm22);
			termList22[0] = aTerm221;
			termList22[1] = aTerm222;
			aTerm22 = term_mergeTerm(noOfTerm22, termList22, OP_SUM);
			free(termList22);
			
			//			
			
			aTerm2->operandList[0] = aTerm21;
			aTerm2->operandList[1] = aTerm22;
			
			
			// term 3
			aLiteral3 = literal_createLiteralFromConstant(1);
			aTerm3 = term_createTermFromLiteral(aLiteral3);
			
			
			// combine
			aTerm->operandList[0] = aTerm1;
			aTerm->operandList[1] = aTerm2;
			aTerm->operandList[2] = aTerm3;
			
			break;
		case 1:

			noOfSmallTerm = 3;
			aTerm->noOfOperand = noOfSmallTerm;
			
			aTerm->operandList = (term **) malloc(sizeof(term *)*noOfSmallTerm);
			aTerm->operationList = (int *) malloc(sizeof(int)*noOfSmallTerm);
			
			for (j = 0; j < noOfSmallTerm; j++)
			{
				aTerm->operationList[j] = OP_MINUS;
			}
			
			// term 1
			aTerm1 = term_new();
			aLiteral11 = literal_createLiteralFromConstant(3);;
			aLiteral12 = literal_createLiteralFromVariable(1, 1);;
			
			aTerm1->noOfOperand = 2;
			aTerm1->operandList = (term **) malloc(sizeof(term *)*aTerm1->noOfOperand);
			aTerm1->operationList = (int *) malloc(sizeof(int)*aTerm1->noOfOperand);
			
			aTerm1->operandList[0] = term_createTermFromLiteral(aLiteral11);
			aTerm1->operandList[1] = term_createTermFromLiteral(aLiteral12);
			
			aTerm1->operationList[0] = OP_MULT;
			aTerm1->operationList[1] = OP_MULT;
			
			
			// term 2
			aTerm2 = term_new();
			
			aTerm2->noOfOperand = 2;
			aTerm2->operandList = (term **) malloc(sizeof(term *)*aTerm2->noOfOperand);
			aTerm2->operationList = (int *) malloc(sizeof(int)*aTerm2->noOfOperand);
			
			
			
			aTerm2->operationList[0] = OP_DIVIDE;
			aTerm2->operationList[1] = OP_DIVIDE;
			
			// term 21
			noOfLiteral21 = 2;
			aLiteral21List = (literal **) malloc(sizeof(literal *)*noOfLiteral21);
			aLiteral21List[0] = literal_createLiteralFromVariable(1, 1);
			aLiteral21List[1] = literal_createLiteralFromVariable(2, 1);
			
			
			aTerm21 = term_mergeLiteral(noOfLiteral21, aLiteral21List, OP_MULT);

			free(aLiteral21List);
			
			// term 22
			// term 221
			noOfLiteral221 = 2;
			aLiteral221List = (literal **) malloc(sizeof(literal *)*noOfLiteral221);
			aLiteral221List[0] = literal_createLiteralFromVariable(0, 1);
			aLiteral221List[1] = literal_createLiteralFromVariable(3, 1);
			
			
			aTerm221 = term_mergeLiteral(noOfLiteral221, aLiteral221List, OP_MULT);

			free(aLiteral221List);
			
			// term 222
			noOfLiteral222 = 2;
			aLiteral222List = (literal **) malloc(sizeof(literal *)*noOfLiteral222);
			aLiteral222List[0] = literal_createLiteralFromVariable(1, 1);
			aLiteral222List[1] = literal_createLiteralFromVariable(2, 1);
			
			
			aTerm222 = term_mergeLiteral(noOfLiteral222, aLiteral222List, OP_MULT);

			free(aLiteral222List);
			
			noOfTerm22 = 2;
			termList22 = (term **) malloc(sizeof(term *)*noOfTerm22);
			termList22[0] = aTerm221;
			termList22[1] = aTerm222;
			aTerm22 = term_mergeTerm(noOfTerm22, termList22, OP_SUM);
			free(termList22);
			
			//			
			
			aTerm2->operandList[0] = aTerm21;
			aTerm2->operandList[1] = aTerm22;
			
			
			// term 3
			aLiteral3 = literal_createLiteralFromConstant(1);
			aTerm3 = term_createTermFromLiteral(aLiteral3);
			
			
			// combine
			aTerm->operandList[0] = aTerm1;
			aTerm->operandList[1] = aTerm2;
			aTerm->operandList[2] = aTerm3;
			break;
		case 2:

			noOfSmallTerm = 3;
			aTerm->noOfOperand = noOfSmallTerm;
			
			aTerm->operandList = (term **) malloc(sizeof(term *)*noOfSmallTerm);
			aTerm->operationList = (int *) malloc(sizeof(int)*noOfSmallTerm);
			
			for (j = 0; j < noOfSmallTerm; j++)
			{
				aTerm->operationList[j] = OP_MINUS;
			}
			
			// term 1
			aTerm1 = term_new();
			aLiteral11 = literal_createLiteralFromConstant(3);;
			aLiteral12 = literal_createLiteralFromVariable(2, 1);;
			
			aTerm1->noOfOperand = 2;
			aTerm1->operandList = (term **) malloc(sizeof(term *)*aTerm1->noOfOperand);
			aTerm1->operationList = (int *) malloc(sizeof(int)*aTerm1->noOfOperand);
			
			aTerm1->operandList[0] = term_createTermFromLiteral(aLiteral11);
			aTerm1->operandList[1] = term_createTermFromLiteral(aLiteral12);
			
			aTerm1->operationList[0] = OP_MULT;
			aTerm1->operationList[1] = OP_MULT;
			
			
			// term 2
			aTerm2 = term_new();
			
			aTerm2->noOfOperand = 2;
			aTerm2->operandList = (term **) malloc(sizeof(term *)*aTerm2->noOfOperand);
			aTerm2->operationList = (int *) malloc(sizeof(int)*aTerm2->noOfOperand);
			
			
			
			aTerm2->operationList[0] = OP_DIVIDE;
			aTerm2->operationList[1] = OP_DIVIDE;
			
			// term 21
			noOfLiteral21 = 2;
			aLiteral21List = (literal **) malloc(sizeof(literal *)*noOfLiteral21);
			aLiteral21List[0] = literal_createLiteralFromVariable(1, 1);
			aLiteral21List[1] = literal_createLiteralFromVariable(2, 1);
			
			
			aTerm21 = term_mergeLiteral(noOfLiteral21, aLiteral21List, OP_MULT);

			free(aLiteral21List);
			
			// term 22
			// term 221
			noOfLiteral221 = 2;
			aLiteral221List = (literal **) malloc(sizeof(literal *)*noOfLiteral221);
			aLiteral221List[0] = literal_createLiteralFromVariable(0, 1);
			aLiteral221List[1] = literal_createLiteralFromVariable(3, 1);
			
			
			aTerm221 = term_mergeLiteral(noOfLiteral221, aLiteral221List, OP_MULT);

			free(aLiteral221List);
			
			// term 222
			noOfLiteral222 = 2;
			aLiteral222List = (literal **) malloc(sizeof(literal *)*noOfLiteral222);
			aLiteral222List[0] = literal_createLiteralFromVariable(1, 1);
			aLiteral222List[1] = literal_createLiteralFromVariable(2, 1);
			
			
			aTerm222 = term_mergeLiteral(noOfLiteral222, aLiteral222List, OP_MULT);

			free(aLiteral222List);
			
			noOfTerm22 = 2;
			termList22 = (term **) malloc(sizeof(term *)*noOfTerm22);
			termList22[0] = aTerm221;
			termList22[1] = aTerm222;
			aTerm22 = term_mergeTerm(noOfTerm22, termList22, OP_SUM);
			free(termList22);
			
			//			
			
			aTerm2->operandList[0] = aTerm21;
			aTerm2->operandList[1] = aTerm22;
			
			
			// term 3
			aLiteral3 = literal_createLiteralFromConstant(0);
			aTerm3 = term_createTermFromLiteral(aLiteral3);
			
			
			// combine
			aTerm->operandList[0] = aTerm1;
			aTerm->operandList[1] = aTerm2;
			aTerm->operandList[2] = aTerm3;
			break;
		case 3:
			noOfSmallTerm = 3;
			aTerm->noOfOperand = noOfSmallTerm;
			
			aTerm->operandList = (term **) malloc(sizeof(term *)*noOfSmallTerm);
			aTerm->operationList = (int *) malloc(sizeof(int)*noOfSmallTerm);
			
			for (j = 0; j < noOfSmallTerm; j++)
			{
				aTerm->operationList[j] = OP_MINUS;
			}
			
			// term 1
			aTerm1 = term_new();
			aLiteral11 = literal_createLiteralFromConstant(3);;
			aLiteral12 = literal_createLiteralFromVariable(3, 1);;
			
			aTerm1->noOfOperand = 2;
			aTerm1->operandList = (term **) malloc(sizeof(term *)*aTerm1->noOfOperand);
			aTerm1->operationList = (int *) malloc(sizeof(int)*aTerm1->noOfOperand);
			
			aTerm1->operandList[0] = term_createTermFromLiteral(aLiteral11);
			aTerm1->operandList[1] = term_createTermFromLiteral(aLiteral12);
			
			aTerm1->operationList[0] = OP_MULT;
			aTerm1->operationList[1] = OP_MULT;
			
			
			// term 2
			aTerm2 = term_new();
			
			aTerm2->noOfOperand = 2;
			aTerm2->operandList = (term **) malloc(sizeof(term *)*aTerm2->noOfOperand);
			aTerm2->operationList = (int *) malloc(sizeof(int)*aTerm2->noOfOperand);
			
			
			
			aTerm2->operationList[0] = OP_DIVIDE;
			aTerm2->operationList[1] = OP_DIVIDE;
			
			// term 21
			noOfLiteral21 = 2;
			aLiteral21List = (literal **) malloc(sizeof(literal *)*noOfLiteral21);
			aLiteral21List[0] = literal_createLiteralFromVariable(0, 1);
			aLiteral21List[1] = literal_createLiteralFromVariable(3, 1);
			
			
			aTerm21 = term_mergeLiteral(noOfLiteral21, aLiteral21List, OP_MULT);

			free(aLiteral21List);
			
			// term 22
			// term 221
			noOfLiteral221 = 2;
			aLiteral221List = (literal **) malloc(sizeof(literal *)*noOfLiteral221);
			aLiteral221List[0] = literal_createLiteralFromVariable(0, 1);
			aLiteral221List[1] = literal_createLiteralFromVariable(3, 1);
			
			
			aTerm221 = term_mergeLiteral(noOfLiteral221, aLiteral221List, OP_MULT);

			free(aLiteral221List);
			
			// term 222
			noOfLiteral222 = 2;
			aLiteral222List = (literal **) malloc(sizeof(literal *)*noOfLiteral222);
			aLiteral222List[0] = literal_createLiteralFromVariable(1, 1);
			aLiteral222List[1] = literal_createLiteralFromVariable(2, 1);
			
			
			aTerm222 = term_mergeLiteral(noOfLiteral222, aLiteral222List, OP_MULT);

			free(aLiteral222List);
			
			noOfTerm22 = 2;
			termList22 = (term **) malloc(sizeof(term *)*noOfTerm22);
			termList22[0] = aTerm221;
			termList22[1] = aTerm222;
			aTerm22 = term_mergeTerm(noOfTerm22, termList22, OP_SUM);
			free(termList22);
			
			//			
			
			aTerm2->operandList[0] = aTerm21;
			aTerm2->operandList[1] = aTerm22;
			
			
			// term 3
			aLiteral3 = literal_createLiteralFromConstant(2);
			aTerm3 = term_createTermFromLiteral(aLiteral3);
			
			
			// combine
			aTerm->operandList[0] = aTerm1;
			aTerm->operandList[1] = aTerm2;
			aTerm->operandList[2] = aTerm3;
			break;
		}
	}

	return termList;
}

void test_newton()
{
	int noOfTerm = 4;
	term **termList;

	float *initVector;
	float *finalVector;

	int i;

	initVector = (float *) malloc(sizeof(float)*noOfTerm);
	finalVector = (float *) malloc(sizeof(float)*noOfTerm);

	termList = test_newtonSetInitial();

	initVector[0] = 0.5f;
	initVector[1] = 0.5f;
	initVector[2] = 0.167f;
	initVector[3] = 0.833f;

	// print
	// pritn the four terms
	for (i = 0; i < noOfTerm; i++)
	{
		printf("Term %d\n", i);
		printf("------\n");
		term_print(termList[i]);
		printf("\n");
	}

	// Initial vector
	printf("Initial vector ...\n");
	printf("(");
	for (i = 0; i < noOfTerm; i++)
	{
		printf("%.4f ", initVector[i]);
	}
	printf(")\n");

	printf("Executing newton's method...\n");
	finalVector = newton_main(termList, noOfTerm, initVector);

	// print
	printf("Final vector ...\n");
	printf("(");
	for (i = 0; i < noOfTerm; i++)
	{
		printf("%.4f ", finalVector[i]);
	}
	printf(")\n");

	free(initVector);
	free(finalVector);
}


int main_newton()
{

	test_newton();

	return 0;

}
