
#include "term.h"


literal *literal_new()
{
	literal *returnValue;

	returnValue = (literal *) malloc(sizeof(literal));
	memset(returnValue, 0, sizeof(literal));

	returnValue->exponent = 1;

	return returnValue;
}

literal *literal_createLiteralFromConstant(LIT_TYPE aNo)
{
	literal *returnValue;

	returnValue = literal_new();

	returnValue->isConstant = 1;
	returnValue->value = aNo;

	return returnValue;
}

literal *literal_createLiteralFromVariable(int varNo, int exponent)
{
	literal *returnValue;

	returnValue = literal_new();

	returnValue->isConstant = 0;
	returnValue->value = (float) varNo;
	returnValue->exponent = exponent;

	return returnValue;
}

/*
literal **literal_simplifyLiteralList(int noOfLiteral, literal **literalList)
{
	literal **returnValue;
	literal *aLiteral;

	int i;

	for (i = 0; i < noOfLiteral; i++)
	{
		aLiteral = literalList[i];

		if (aLiteral->isConstant == 1)
		{
		}
	}

	returnValue = (literal **) malloc(sizeof(literal *)*);

	return returnValue;
}
*/

term *term_new()
{
	term *returnValue;

	returnValue = (term *) malloc(sizeof(term));
	memset(returnValue, 0, sizeof(term));

	returnValue->exponent = 1;

	return returnValue;
}

term *term_createTermFromLiteral(literal *aLiteral)
{
	term *returnValue;

	returnValue = term_new();

	returnValue->aLiteral = aLiteral;
	returnValue->noOfOperand = 1;

	return returnValue;
}

float literal_returnValueFromLiteral(literal *aLiteral, float *vector)
{
	float returnValue;

	int i;

	int varNo;

	float aValue;

	if (aLiteral->isConstant == 1)
	{
		returnValue = (float) aLiteral->value;
		// don't need to consider exponent
	}
	else
	{
		returnValue = 1;

		varNo = (int) aLiteral->value;

		aValue = vector[varNo];
		
		// don't use "power" which is slow
		// use iterative loop
		for (i = 0; i < aLiteral->exponent; i++)
		{
			returnValue *= aValue;
		}

	}

	return returnValue;
}


float term_combineOperand(float aValue1, float aValue2, int operationNo)
{
	float returnValue;

	switch (operationNo)
	{
	case OP_SUM:
		returnValue = aValue1 + aValue2;
		break;
	case OP_MINUS:
		returnValue = aValue1 - aValue2;
		break;
	case OP_MULT:
		returnValue = aValue1 * aValue2;
		break;
	case OP_DIVIDE:
		returnValue = aValue1 / aValue2;
		break;
	default:
		returnValue = -1;
		break;
	}

	return returnValue;
}

literal *term_getLiteral(term *aTerm)
{
	literal *returnValue;

	if (aTerm->noOfOperand == 1)
	{
		returnValue = aTerm->aLiteral;
	}
	else
	{
		returnValue = NULL;
	}

	return returnValue;
}

float term_calculateValue(term *aTerm, float *vector)
{
	float returnValue;

	//double *stack;
	//int stackIndex;
	//int MAX_STACKSIZE = 10000000;

	int noOfOperand;
	term **operandList;
	int *operationList;
	int i;


	literal *aLiteral1;
	term *aTerm1;
	term *aTerm2;

	float temp;

	float aValue1, aValue2;

	int operationNo;

	returnValue = 0;

	noOfOperand = aTerm->noOfOperand;
	operandList = aTerm->operandList;
	operationList = aTerm->operationList;

	

	if (noOfOperand == 1)
	{
		//aTerm1 = operandList[0];
		//aLiteral1 = aTerm1->aLiteral;
		aLiteral1 = aTerm->aLiteral;

		if (aLiteral1 == NULL)
		{
			returnValue = term_calculateValue(aTerm->operandList[0], vector);
		}
		else
		{
			returnValue = literal_returnValueFromLiteral(aLiteral1, vector);
		}
	}
	else
	{
		//stack = (double *) malloc(sizeof(double)*MAX_STACKSIZE);
		//memset(stack, 0, sizeof(double)*MAX_STACKSIZE);

		//stackIndex = -1;

		for (i = 0; i < noOfOperand-1; i++)
		{
			operationNo = operationList[i];

			if (i == 0)
			{
				aTerm1 = operandList[i];
				aValue1 = term_calculateValue(aTerm1, vector);

				aTerm2 = operandList[i+1];
				aValue2 = term_calculateValue(aTerm2, vector);

				returnValue = term_combineOperand(aValue1, aValue2, operationNo);
			}
			else
			{
				aTerm1 = operandList[i+1];
				aValue1 = term_calculateValue(aTerm1, vector);

				returnValue = term_combineOperand(returnValue, aValue1, operationNo);
			}
		}

		// not use power for efficiency
		temp = 1;
		for (i = 0; i < aTerm->exponent; i++)
		{
			temp *= returnValue;
		}

		returnValue = temp;

		//free(stack);
	}

	return returnValue;
}

void literal_copy(literal *aLiteral1, literal *aLiteral2)
{
	memcpy(aLiteral1, aLiteral2, sizeof(literal));
}


// assume single literal
void term_copySingle(term *aTerm1, term *aTerm2)
{
	literal *aOrgLiteral;
	literal *aNewLiteral;

	aTerm1->noOfOperand = 1;

	aOrgLiteral = aTerm2->aLiteral;

	aNewLiteral = literal_new();
	literal_copy(aNewLiteral, aOrgLiteral);

	aTerm1->aLiteral = aNewLiteral;

}

// re-use term2 content
void term_copy(term *aTerm1, term *aTerm2)
{
	if (aTerm2->noOfOperand == 1)
	{
		// copy literal
		aTerm1->noOfOperand = 1;

		aTerm1->aLiteral = literal_new();

		literal_copy(aTerm1->aLiteral, aTerm2->aLiteral);
	}
	else
	{
		aTerm1->noOfOperand = aTerm2->noOfOperand;

		aTerm1->operandList = (term **) malloc(sizeof(term *)*aTerm1->noOfOperand);
		aTerm1->operationList = (int *) malloc(sizeof(int)*aTerm1->noOfOperand);

		memcpy(aTerm1->operandList, aTerm2->operandList, sizeof(term *)*aTerm1->noOfOperand);
		memcpy(aTerm1->operationList, aTerm2->operationList, sizeof(term *)*aTerm1->noOfOperand);

		aTerm1->exponent = aTerm2->exponent;
	}
}

term *term_clone(term *aTerm)
{
	term *returnValue;

	int i;

	term *aNewTerm;

	returnValue = term_new(); 
	if (aTerm->noOfOperand == 1)
	{
		// copy literal
		returnValue->noOfOperand = 1;

		returnValue->aLiteral = literal_new();

		literal_copy(returnValue->aLiteral, aTerm->aLiteral);
	}
	else
	{
		returnValue->noOfOperand = aTerm->noOfOperand;

		returnValue->operandList = (term **) malloc(sizeof(term *)*returnValue->noOfOperand);
		returnValue->operationList = (int *) malloc(sizeof(int)*returnValue->noOfOperand);

		//memcpy(returnValue->operandList, aTerm->operandList, sizeof(term *)*returnValue->noOfOperand);
		memcpy(returnValue->operationList, aTerm->operationList, sizeof(term *)*returnValue->noOfOperand);

		for (i = 0; i < returnValue->noOfOperand; i++)
		{
			aNewTerm = term_clone(aTerm->operandList[i]);
			returnValue->operandList[i] = aNewTerm;
		}

		returnValue->exponent = aTerm->exponent;
	}

	return returnValue;
}

void term_free(term *aTerm)
{
	int i;

	if (aTerm != NULL)
	{
		if (aTerm->noOfOperand == 0)
		{

		}
		else if (aTerm->noOfOperand == 1)
		{
			free(aTerm->aLiteral);
		}
		else
		{
			free(aTerm->operationList);
			for (i = 0; i < aTerm->noOfOperand; i++)
			{
				term_free(aTerm->operandList[i]);
			}
			free(aTerm->operandList);
		}

		free(aTerm);
	}
}

void literal_print(literal *aLiteral)
{
	if (aLiteral->isConstant == 1)
	{
		printf(" %.4f ", aLiteral->value);
	}
	else
	{
		printf(" x%d", (int) aLiteral->value);
		if (aLiteral->exponent != 1)
		{
			printf("^%d", aLiteral->exponent);
		}
		printf(" ");
	}
}

void op_print(int operationNo)
{
	switch (operationNo)
	{
	case OP_SUM:
		printf(" + ");
		break;
	case OP_MINUS:
		printf(" - ");
		break;
	case OP_MULT:
		printf(" * ");
		break;
	case OP_DIVIDE:
		printf(" / ");
		break;
	default:
		break;
	}
}

void term_print(term *aTerm)
{
	int noOfOperand;
	term **operandList;
	int *operationList;

	term *aTermSingle;

	literal *aLiteral;

	int i;

	noOfOperand = aTerm->noOfOperand;
	operandList = aTerm->operandList;
	operationList = aTerm->operationList;

	if (noOfOperand == 1)
	{
		
		aLiteral = aTerm->aLiteral;
		if (aLiteral == NULL)
		{
			term_print(aTerm->operandList[0]);
		}
		else
		{
			literal_print(aLiteral);
		}
	}
	else
	{

		for (i = 0; i < noOfOperand; i++)
		{
			aTermSingle = operandList[i];

			if (aTermSingle->noOfOperand != 1)
			{
				printf(" ( ");
			}

			term_print(aTermSingle);

			if (aTermSingle->noOfOperand != 1)
			{
				printf(" )");

				if (aTermSingle->exponent != 1)
				{
					printf("^%d", aTermSingle->exponent);
				}

				printf(" ");
			}

			if (i != noOfOperand-1)
			{
				op_print(operationList[i]);
			}
		}
	}
}


term *term_mergeLiteral(int noOfLiteral, literal **literalList, int operationNo)
{
	term *returnValue;

	literal *aLiteral;

	int i;

	returnValue = term_new();

	returnValue->noOfOperand = noOfLiteral;

	returnValue->operandList = (term **) malloc(sizeof(term *)*returnValue->noOfOperand);
	returnValue->operationList = (int *) malloc(sizeof(int)*returnValue->noOfOperand);

	for (i = 0; i < returnValue->noOfOperand; i++)
	{
		aLiteral = literalList[i];

		returnValue->operandList[i] = term_createTermFromLiteral(aLiteral);
		returnValue->operationList[i] = operationNo;
	}

	return returnValue;
}

term *term_mergeTerm(int noOfTerm, term **termList, int operationNo)
{
	term *returnValue;

	term *aTerm;

	int i;

	returnValue = term_new();

	returnValue->noOfOperand = noOfTerm;

	returnValue->operandList = (term **) malloc(sizeof(term *)*returnValue->noOfOperand);
	returnValue->operationList = (int *) malloc(sizeof(int)*returnValue->noOfOperand);

	for (i = 0; i < returnValue->noOfOperand; i++)
	{
		aTerm = termList[i];

		returnValue->operandList[i] = aTerm;
		returnValue->operationList[i] = operationNo;
	}

	return returnValue;
}

